/* Copyright (C) 2003, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*
 * ClassEnvironment.java
 *
 *
 */

package uiuc.Jumbo.Compiler;

import uiuc.Jumbo.Jaemus.*;
import java.util.*;
import java.io.*;
import uiuc.Jumbo.Util.*;

/**
 * An environment for the first traversal of the constructors.
 * Contains variable definitions, free variable lists, and gensym names.
 * @version      1.0, 6 Feb 2003
 * @author       Lars Clausen
 *
 */

public final class ClassEnvironment extends CodeEnvironment {
    Local locals;
    Local free;
    int gensym;
    /** Outer scope, including statement-level scopes */
    final ClassEnvironment outer;

    public ClassEnvironment() {
        outer = null;
    }

    private ClassEnvironment(ClassEnvironment outer) {
        this.fixed = outer.fixed;
        this.locals = outer.locals;
        this.gensym = outer.gensym;
        this.free = outer.free;
        this.outer = outer;
    }

    public ClassEnvironment(ClassEnvironment outer, ClassInfo c) {
        super(outer, c, null);
        this.outer = outer;
        this.gensym = outer.gensym;
    }
    
    public ClassEnvironment startClass(ClassInfo c) {
        return new ClassEnvironment(this, c);
    }

    public ClassEnvironment openScope() {
        ClassEnvironment newenv = new ClassEnvironment(this);
        return newenv;
    }

    /** Close off a scope, preserving free variables */
    /* Not compositional -- changes this. */
    public ClassEnvironment closeScope(ClassEnvironment closing) {
        if (closing.free != null) {
            if (free != null) {
                Local combined = free;
                for (Local newcomer = closing.free; 
                     newcomer != null && newcomer != free;
                     newcomer = newcomer.next) {
                    combined = new Local(newcomer.name, newcomer.type,
                                       newcomer.flags, combined);
                }
                free = combined;
            } else {
                free = closing.free;
            }
        }
        gensym = closing.gensym;
        return this;
    }

    public ClassEnvironment defineVar(String name, Type t, int flags) {
        locals = new Local(name, t, flags, locals);
        return this;
    }

    public ClassEnvironment generateVar(String name, Type t, int flags) {
        locals = new Local(name+"$$"+locals.size(), t, flags, locals);
        return this;
    }

    public Local getNewestVar() {
        return locals;
    }

    /** Look up a variable in this environment */
    private Local lookupVar(String name) {
        for (Local v = locals; v != null; v = v.next) {
            if (v.name == name) {
                return v;
            }
        }
        return null;
    }

    public Local getVar(String name) {
        for (Local v = locals; v != null; v = v.next) {
            if (v.name == name) {
                return v;
            }
        }
        if (outer != null) return outer.getVar(name);
        return null;
    }

    /** Return the ClassEnvironment in which this var is defined */
    private ClassEnvironment varDefined(String name) {
        if (lookupVar(name) != null) return this;
        if (outer != null) return outer.varDefined(name);
        return null;
    }

    public Local findFreeVar(String name) {
        String n = name.intern();
        ClassEnvironment definer = varDefined(n);
        if (definer != null && 
            definer.getCurrentClass() != this.getCurrentClass()) {
            Local v = definer.lookupVar(n);
            if ((v.flags & Flags.FINAL) == 0)
                throw new Error("Can't capture non-final variables: "+v);
            return v;
        }
        return null;
    }

    public ClassEnvironment addFree(Local v) {
        String n = v.name.intern();
        for (Local v1 = free; v1 != null; v1 = v1.next) {
            if (v1.name == n) return this;
        }
        free = new Local(n, v.type, v.flags, free);
        return this;
    }

    public ClassEnvironment resetFree() {
        free = null;
        return this;
    }

    public String genSym(String prefix) {
        return prefix+"$"+gensym;
    }

    public ClassEnvironment updateGenSym() {
        gensym++;
        return this;
    }

    public String toString() {
        String s = "Free: {";
        for (Local v = free; v != null; v = v.next) {
            s += v;
            if (v.next != null) s += ", ";
        }
        return s+"}";
    }
}
