/* Copyright (C) 2003, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*
 * CodeEnvironment.java
 *
 *
 */

package uiuc.Jumbo.Compiler;

import uiuc.Jumbo.Jaemus.*;
import java.util.*;
import java.io.*;
import uiuc.Jumbo.Util.*;

/**
 * An environment for traversal of methods.
 * Keeps a link to constant stuff, and methods to access it.
 * @version      1.0, 7 Feb 2003
 * @author       Lars Clausen
 *
 */

public class CodeEnvironment {
    FixedEnvironment fixed = new FixedEnvironment();

    public CodeEnvironment() {
        addImportPackage("java/lang");
    }

    public CodeEnvironment(CodeEnvironment old) {
        fixed = old.fixed;
    }

    /** Reset the FixedEnvironment */
    public CodeEnvironment(FixedEnvironment old) {
        fixed = new FixedEnvironment(old);
    }

    public CodeEnvironment(CodeEnvironment e, ClassInfo c, Local free) {
        fixed = new FixedEnvironment(e, c, free);
    }

    protected CodeEnvironment(CodeEnvironment old, MethodInfo method) {
        fixed = new FixedEnvironment(old.fixed, method);
    }

    public final void addInitializer(Code initval, boolean isInstance) {
        fixed.addInitializer(initval, isInstance);
    }

    public final MonoList getInstanceInitializers() {
        return fixed.getInstanceInitializers();
    }

    public final MonoList getStaticInitializers() {
        return fixed.getStaticInitializers();
    }

    public final ClassInfo getClass(String classname) {
        ClassInfo ci = ClassTable.get(classname);
        if (ci == null)
            throw new Error("Can't find class "+classname);
        else return ci;
    }

    public final boolean isClass(String classname) {
        try {
            return get(classname) != null;
        } catch (ClassNotFoundException e) {
            return false;
        }
    }

    public final String fullClassName(String name) {
        return fixed.fullClassName(name);
    }

    public final void addClass(ClassInfo c) {
        // Should check for overlap here.
        ClassTable.put(c);
    }

    public ClassInfo get(String classname) throws ClassNotFoundException {
        return fixed.get(classname);
    }

    public final CodeEnvironment startCompilationUnit() {
        fixed = fixed.startCompilationUnit();
        return this;
    }

    public final void addImport(String classname) {
        fixed.addImport(classname);
    }

    public final void addImportPackage(String packagename) {
        fixed.addImportPackage(packagename);
    }

    /** Returns the method with the more specific type.
     * Note: Doesn't handle ambiguous methods (e.g. foo(int, long) vs
     * foo(long, int)).
     * @param m1 A MethodInfo instance, or null.
     * @param m2 A MethodInfo instance, or null.
     */
    private final MethodInfo findBestMethod(MethodInfo m1, MethodInfo m2) {
        if (m1 == null) return m2;
        return m1.findBestMethod(m2);
    }

    private final MethodInfo findMethodCore(ClassInfo c, String name, MethodType type)
        throws MethodNotFoundException {
        MethodInfo best = null;
        for (Iterator i = c.methods.iterator(); i.hasNext(); ) {
            MethodInfo m = (MethodInfo)i.next();
            if (false)
                System.out.println("Testing "+m+" ("+m.getName()+") vs. "+name+" type "+type);
            if (m.getName().equals(name) &&
                type.conforms(m.getType())) {
                best = findBestMethod(m, best);
            }
        }

        if (c.supername != null && !name.equals("<init>")) {
            try {
                best = findBestMethod(best, findMethod(c.supername, name, type));
            } catch (MethodNotFoundException e) {
            }
        } 

        if (c.implementees != null && c.implementees.size() > 0) {
            for (Iterator i = c.implementees.iterator(); i.hasNext(); ) {
                String iface = (String)i.next();
                try {
                    best = findBestMethod(best, findMethod(iface, name, type));
                } catch (MethodNotFoundException e) {
                }
            }
        } else if (c.isInterface()) {
            // An interface with no superinterfaces inherits from Object
            try {
                best = findBestMethod(best, findMethod("java/lang/Object", name, type));
            } catch (MethodNotFoundException e) {
            }
        }

        // FIXME: We should have the original class name in here
        if (best == null) {
            throw new MethodNotFoundException(c.getName(), name, type,
                                              "No such method");
        } else {
            return best;
        }

    }

    public final MethodInfo findMethod(ClassInfo c, String name, MethodType type) 
        throws MethodNotFoundException {
        try {
            return findMethodCore(c, name, type);
        } catch (MethodNotFoundException e) {
            String reason = e.reason;
            throw new MethodNotFoundException(c.getName(), name, type,
                                              reason);
        }
    }

    public final MethodInfo findMethod(String inClass, String name, MethodType type) 
        throws MethodNotFoundException {
        ClassInfo c = null;
        try {
            c = get(inClass);
        } catch (ClassNotFoundException e) {
            throw new MethodNotFoundException(inClass, name, type, "No such class");
        }
        return findMethod(c, name, type);
    }

    public final FieldInfo findField(ClassInfo c, String name) {
        if (c == null) return null;
        for (Iterator i = c.fields.iterator(); i.hasNext(); ) {
            FieldInfo f = (FieldInfo)i.next();
            if (f.getName().equals(name)) {
                return f;
            }
        }
        // FIXME: Must check interface fields as well.
        if (c.supername != null) {
            FieldInfo f = findField(c.supername, name);
            if (f != null) return f;
        }
        if (c.implementees != null) {
            for (Iterator i = c.implementees.iterator(); i.hasNext(); ) {
                FieldInfo f = findField((String)i.next(), name);
                if (f != null) return f;
            }
        }
            
        return null;
    }

    /** Finds a field in the current class */
    public final FieldInfo findField(String name) {
        return findField(getSelfClass(), name);
    }

    public final FieldInfo findField(String inClass, String name) {
        ClassInfo c = null;
        try {
            c = get(inClass);
        } catch (ClassNotFoundException e) {
            return null;
        }
        return findField(c, name);
    }

    public final Type getSelfType() {
        return fixed.getSelfType();
    }
    /** Returns the lexically enclosing class.
     * @deprecated Use getCurrentClass() instead
     */
    public final ClassInfo getSelfClass() {
        return fixed.getCurrentClass();
    }
    public final void setCurrentClass(ClassInfo c) {
        fixed = new FixedEnvironment(this, c, null);
    }
    public final ClassInfo getCurrentClass() {
        return fixed.getCurrentClass();
    }

    public final void setCurrentMethod(MethodInfo m) {
        fixed = new FixedEnvironment(fixed, m);
    }
    public final MethodInfo getCurrentMethod() {
        return fixed.getCurrentMethod();
    }
    public final void setCurrentPackage(String name) {
        name = name.replace('.', '/');
        fixed = new FixedEnvironment(fixed, name+"/");
    }
    public final String getCurrentPackage() {
        return fixed.getCurrentPackage();
    }
    public final CodeEnvironment getOuter() {
        return fixed.getOuter();
    }
    public final int numEnclosingClasses() {
        return fixed.numEnclosingClasses();
    }
    public final Local getFree() {
        return fixed.getFree();
    }
    public final boolean selfIsInner() {
        return fixed.getOuter() != null && fixed.getOuter().getCurrentClass() != null;
    }
    protected final Error error(String message) {
        return new Error(getCurrentClass()+":"+
                         getCurrentMethod()+":"+
                         message);
    }
}    

class FieldNotFoundException extends Exception {
    String inClass;
    String name;
    String reason;
    public FieldNotFoundException(String inClass, String name, String reason) {
        this.inClass = inClass;
        this.name = name;
        this.reason = reason;
    }
    public String toString() {
        return inClass+"."+name+": "+reason;
    }
}

class MethodNotFoundException extends Exception {
    final String inClass;
    final String name;
    final String reason;
    final MethodType type;
    public MethodNotFoundException(String inClass, String name, 
                                   MethodType type, String reason) {
        this.inClass = inClass;
        this.name = name;
        this.type = type;
        this.reason = reason;
    }
    public String toString() {
        return type.toString(inClass+"."+name)+": "+reason;
    }
}

