/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*
 * ConstVal.java
 *
 *
 */

package uiuc.Jumbo.Compiler;

import uiuc.Jumbo.Jaemus.*;

/**
 * Class handling constant values and their calculations
 * @version      1.0, 29 Oct 2001
 * @author       Lars Clausen
 *
*/

public abstract class ConstVal {
    public final static ConstVal make(int val) {
        return new IntConstVal(val);
    }
    public final static ConstVal make(float val) {
        return new FloatConstVal(val);
    }
    public final static ConstVal make(long val) {
        return new LongConstVal(val);
    }
    public final static ConstVal make(double val) {
        return new DoubleConstVal(val);
    }
    /*
    public static ConstVal make(char val) {
        return new CharConstVal(val);
    }
    public static ConstVal make(byte val) {
        return new ByteConstVal(val);
    }
    public static ConstVal make(short val) {
        return new ShortConstVal(val);
    } 
    */
    public final static ConstVal make(boolean val) {
        return new BooleanConstVal(val);
    }
    public final static ConstVal make(String val) {
        return new StringConstVal(val);
    }
    public final static ConstVal makeNull() {
        return new NullConstVal();
    }
    public final static ConstVal makeVoid() {
        return new VoidConstVal();
    }
    public final static ConstVal make(ConstVal v, Type castto) {
        switch (castto.kind) {
        case Type.ARRAY:
        case Type.BOTTOM:
            throw new Error("Can't make constval of type "+castto);
        case Type.OBJECT:
            if (!(castto.class_name.equals("java/lang/String")))
                throw new Error("Can't make constval of type "+castto);
            else if (v.isNull()) return make("null");
            else return v;
        case Type.NULL:
            return new NullConstVal();
        case Type.VOID:
            return new VoidConstVal();
        case Type.BOOLEAN:
            return v;
        case Type.CHAR:
            return new IntConstVal((char)v.intVal());
        case Type.BYTE:
            return new IntConstVal((byte)v.intVal());
        case Type.SHORT:
            return new IntConstVal((short)v.intVal());
        case Type.INT:
            return new IntConstVal(v.intVal());
        case Type.LONG:
            return new LongConstVal(v.longVal());
        case Type.FLOAT:
            return new FloatConstVal(v.floatVal());
        case Type.DOUBLE:
            return new DoubleConstVal(v.doubleVal());
        default:
            throw new Error("Unknown type "+castto);
        }
    }

    public final static ConstVal make(Object o, Type t) {
        if (o instanceof Byte)
            return make(make(((Byte)o).byteValue()), t);
        else if (o instanceof Character)
            return make(make(((Character)o).charValue()), t);
        else if (o instanceof Short)
            return make(make(((Short)o).shortValue()), t);
        else if (o instanceof Integer)
            return make(make(((Integer)o).intValue()), t);
        else if (o instanceof Long)
            return make(make(((Long)o).longValue()), t);
        else if (o instanceof Float)
            return make(make(((Float)o).floatValue()), t);
        else if (o instanceof Double)
            return make(make(((Double)o).doubleValue()), t);
        else if (o instanceof String)
            return make(make((String)o), t);
        else if (o instanceof Boolean)
            return make(make(((Boolean)o).booleanValue()), t);
        else throw new Error("Can't make constval from "+o.getClass()+" to "+t);
    }

    public final static boolean legalConstantType(Type t) {
        return t.isPrimitive() || t.equals(Type.string_type);
    }

    abstract public Object constantObject();
    public ConstVal mathBinOp(int op, ConstVal other) {
        throw new WrongConstTypeError();
    }
    public ConstVal bitwiseBinOp(int op, ConstVal other) {
        throw new WrongConstTypeError();
    }
    public ConstVal compareBinOp(int op, ConstVal other) {
        throw new WrongConstTypeError();
    }

    // These are used internally to make binops easier
    protected ConstVal mathBinOpReversed(int op, ConstVal other) {
        throw new WrongConstTypeError();
    }
    protected ConstVal bitwiseBinOpReversed(int op, ConstVal other) {
        throw new WrongConstTypeError();
    }
    protected ConstVal compareBinOpReversed(int op, ConstVal other) {
        throw new WrongConstTypeError();
    }

    public ConstVal logicalNegOp() { throw new WrongConstTypeError(); }
    public ConstVal bitwiseNegOp() { throw new WrongConstTypeError(); }
    public ConstVal numericalNegOp() { throw new WrongConstTypeError(); }

    public abstract Instr genLoadCode();

    public boolean isInt() {
        return false;
    }
    public boolean isFloat() {
        return false;
    }
    public boolean isLong() {
        return false;
    }
    public boolean isDouble() {
        return false;
    }
    public boolean isBoolean() {
        return false;
    }
    public boolean isString() {
        return false;
    }
    public boolean isNull() {
        return false;
    }
    public boolean isVoid() {
        return false;
    }

    public int intVal() { throw new WrongConstTypeError(Type.int_type); }
    // Returns an int or booleans as the bytecode value (i.e. boolean = 0||1)
    public int bytecodeIntVal() { return intVal(); }
    public float floatVal() { throw new WrongConstTypeError(Type.float_type); }
    public long longVal() { throw new WrongConstTypeError(Type.long_type); }
    public double doubleVal() { throw new WrongConstTypeError(Type.double_type); }
    public boolean booleanVal() { throw new WrongConstTypeError(Type.boolean_type); }
    public String stringVal() { throw new WrongConstTypeError(Type.string_type); }

    public abstract Type getType();

    class WrongConstTypeError extends Error { 
        Type wrong;
        WrongConstTypeError() {}
        WrongConstTypeError(Type t) {
            wrong = t;
        }
        public String toString() {
            if (wrong != null) 
                return "Constant value is "+getType()+", accessed "+wrong;
            else
                return "Constant value of type "+getType()+" wrongly accessed";
        }
    }
    
    public static final class IntConstVal extends ConstVal {
        final int val;
        public IntConstVal(final int val) { this.val = val; }
        public int intVal() { return val; }
        public float floatVal() { return (float)val; }
        public long longVal() { return (long)val; }
        public double doubleVal() { return (double)val; }
        public String stringVal() { return Integer.toString(val); }
        public Object constantObject() { return new Integer(val); }

        public boolean isInt() { return true; }
        public ConstVal mathBinOp(int op, ConstVal other) {
            if (other.isFloat() || other.isLong() || 
                other.isDouble() || other.isString())
                return other.mathBinOpReversed(op, this);
	    else if (other.isBoolean()) throw new WrongConstTypeError(other.getType());
            else switch (op) {
            case Instr.ADD: return make(val+other.intVal());
            case Instr.SUB: return make(val-other.intVal());
            case Instr.MUL: return make(val*other.intVal());
            case Instr.DIV: return make(val/other.intVal());
            case Instr.REM: return make(val%other.intVal());
            default:
                throw new WrongConstTypeError(other.getType());
            }
        }
        public ConstVal bitwiseBinOp(int op, ConstVal other) {
            if (other.isDouble() || other.isFloat() ||
                other.isString() || other.isBoolean()) {
                throw new WrongConstTypeError(other.getType());
            }
            // Note that long RHS of shifts is converted to int
            else if (other.isLong() && op >= Instr.AND) {
                switch (op) {
                case Instr.AND: return make(((long)val) & other.longVal());
                case Instr.OR: return make(((long)val) | other.longVal());
                case Instr.XOR: return make(((long)val) ^ other.longVal());
                default:
                    throw new WrongConstTypeError(other.getType());
                }
            }
            else switch (op) {
            case Instr.SHL: return make(val << other.intVal());
            case Instr.SHR: return make(val >> other.intVal());
            case Instr.USHR: return make(val >>> other.intVal());
            case Instr.AND: return make(val & other.intVal());
            case Instr.OR: return make(val | other.intVal());
            case Instr.XOR: return make(val ^ other.intVal());
            default:
                throw new WrongConstTypeError(other.getType());
            }
        }
        public ConstVal compareBinOp(int op, ConstVal other) {
            if (other.isFloat() || other.isLong() || 
                other.isDouble() || other.isString())
                return other.compareBinOpReversed(op, this);
            else if (other.isBoolean()) throw new WrongConstTypeError(other.getType());
            else switch (op) {
            case Instr.CMP_EQ: return make(val == other.intVal());
            case Instr.CMP_NE: return make(val != other.intVal());
            case Instr.CMP_LT: return make(val < other.intVal());
            case Instr.CMP_GE: return make(val >= other.intVal());
            case Instr.CMP_GT: return make(val > other.intVal());
            case Instr.CMP_LE: return make(val <= other.intVal());
            default:
                throw new WrongConstTypeError(other.getType());
            }
        }
        public ConstVal bitwiseNegOp() {
            return make(~val);
        }
        public ConstVal numericalNegOp() {
            return make(-val);
        }
        public Instr genLoadCode() {
            return Instr.genConst(val);
        }
        public Type getType() { return Type.int_type; }
    }

    public static final class FloatConstVal extends ConstVal {
        float val;
        FloatConstVal(float val) { this.val = val; }
        public int intVal() { return (int)val; }
        public long longVal() { return (long)val; }
        public float floatVal() { return val; }
        public double doubleVal() { return (double)val; }
        public String stringVal() { return Float.toString(val); }
        public Object constantObject() { return new Float(val); }

        public boolean isFloat() { return true; }
        public ConstVal mathBinOp(int op, ConstVal other) {
            if (other.isDouble() || other.isString())
                return other.mathBinOpReversed(op, this);
            else if (other.isBoolean())
                throw new WrongConstTypeError(other.getType());
            else switch (op) {
            case Instr.ADD: return make(val+other.floatVal());
            case Instr.SUB: return make(val-other.floatVal());
            case Instr.MUL: return make(val*other.floatVal());
            case Instr.DIV: return make(val/other.floatVal());
            case Instr.REM: return make(val%other.floatVal());
            default:
                throw new WrongConstTypeError(other.getType());
            }
        }
        public ConstVal compareBinOp(int op, ConstVal other) {
            if (other.isDouble())
                return other.compareBinOpReversed(op, this);
            else if (other.isBoolean() || other.isString())
                throw new WrongConstTypeError(other.getType());
            else switch (op) {
            case Instr.CMP_EQ: return make(val == other.floatVal());
            case Instr.CMP_NE: return make(val != other.floatVal());
            case Instr.CMP_LT: return make(val < other.floatVal());
            case Instr.CMP_GE: return make(val >= other.floatVal());
            case Instr.CMP_GT: return make(val > other.floatVal());
            case Instr.CMP_LE: return make(val <= other.floatVal());
            default:
                throw new WrongConstTypeError(other.getType());
            }
        }
        protected ConstVal mathBinOpReversed(int op, ConstVal other) {
            switch (op) {
            case Instr.ADD: return make(other.floatVal()+val);
            case Instr.SUB: return make(other.floatVal()-val);
            case Instr.MUL: return make(other.floatVal()*val);
            case Instr.DIV: return make(other.floatVal()/val);
            case Instr.REM: return make(other.floatVal()%val);
            default:
                throw new WrongConstTypeError(other.getType());
            }
        }
        protected ConstVal compareBinOpReversed(int op, ConstVal other) {
            switch (op) {
            case Instr.CMP_EQ: return make(other.floatVal() == val);
            case Instr.CMP_NE: return make(other.floatVal() != val);
            case Instr.CMP_LT: return make(other.floatVal() < val);
            case Instr.CMP_GE: return make(other.floatVal() >= val);
            case Instr.CMP_GT: return make(other.floatVal() > val);
            case Instr.CMP_LE: return make(other.floatVal() <= val);
            default:
                throw new WrongConstTypeError(other.getType());
            }
        }
        public ConstVal numericalNegOp() {
            return make(-val);
        }
        public Instr genLoadCode() {
            return Instr.genConst(val);
        }
        public Type getType() { return Type.float_type; }
    }

    public static final class LongConstVal extends ConstVal {
        long val;
        LongConstVal(long val) { this.val = val; }
        public int intVal() { return (int)val; }
        public long longVal() { return val; }
        public float floatVal() { return (float)val; }
        public double doubleVal() { return (double)val; }
        public String stringVal() { return Long.toString(val); }
        public Object constantObject() { return new Long(val); }

        public boolean isLong() { return true; }
        public ConstVal mathBinOp(int op, ConstVal other) {
            if (other.isFloat() || other.isDouble() || other.isString())
                return other.mathBinOpReversed(op, this);
            else if (other.isBoolean())
                throw new WrongConstTypeError(other.getType());
            else switch (op) {
            case Instr.ADD: return make(val+other.longVal());
            case Instr.SUB: return make(val-other.longVal());
            case Instr.MUL: return make(val*other.longVal());
            case Instr.DIV: return make(val/other.longVal());
            case Instr.REM: return make(val%other.longVal());
            default:
                throw new WrongConstTypeError(other.getType());
            }
        }
        public ConstVal bitwiseBinOp(int op, ConstVal other) {
            if (other.isDouble() || other.isFloat() ||
                other.isString() || other.isBoolean()) {
                throw new WrongConstTypeError(other.getType());
            }
            else switch (op) {
            case Instr.SHL: return make(val << other.intVal());
            case Instr.SHR: return make(val >> other.intVal());
            case Instr.USHR: return make(val >>> other.intVal());
            case Instr.AND: return make(val & other.longVal());
            case Instr.OR: return make(val | other.longVal());
            case Instr.XOR: return make(val ^ other.longVal());
            default:
                throw new WrongConstTypeError(other.getType());
            }
        }
        public ConstVal compareBinOp(int op, ConstVal other) {
            if (other.isFloat() || other.isDouble() || other.isString())
                return other.compareBinOpReversed(op, this);
            if (other.isBoolean()) 
                throw new WrongConstTypeError(other.getType());
            switch (op) {
            case Instr.CMP_EQ: return make(val == other.longVal());
            case Instr.CMP_NE: return make(val != other.longVal());
            case Instr.CMP_LT: return make(val < other.longVal());
            case Instr.CMP_GE: return make(val >= other.longVal());
            case Instr.CMP_GT: return make(val > other.longVal());
            case Instr.CMP_LE: return make(val <= other.longVal());
            default:
                throw new WrongConstTypeError(other.getType());
            }
        }
        protected ConstVal mathBinOpReversed(int op, ConstVal other) {
            switch (op) {
            case Instr.ADD: return make(other.longVal()+val);
            case Instr.SUB: return make(other.longVal()-val);
            case Instr.MUL: return make(other.longVal()*val);
            case Instr.DIV: return make(other.longVal()/val);
            case Instr.REM: return make(other.longVal()%val);
            default:
                throw new WrongConstTypeError(other.getType());
            }
        }
        protected ConstVal compareBinOpReversed(int op, ConstVal other) {
            switch (op) {
            case Instr.CMP_EQ: return make(other.longVal() == val);
            case Instr.CMP_NE: return make(other.longVal() != val);
            case Instr.CMP_LT: return make(other.longVal() < val);
            case Instr.CMP_GE: return make(other.longVal() >= val);
            case Instr.CMP_GT: return make(other.longVal() > val);
            case Instr.CMP_LE: return make(other.longVal() <= val);
            default:
                throw new WrongConstTypeError(other.getType());
            }
        }
        // Bitwise reversed is never called
        public ConstVal bitwiseNegOp() {
            return make(~val);
        }
        public ConstVal numericalNegOp() {
            return make(-val);
        }
        public Instr genLoadCode() {
            return Instr.genConst(val);
        }
        public Type getType() { return Type.long_type; }
    }

    public static final class DoubleConstVal extends ConstVal {
        double val;
        DoubleConstVal(double val) { this.val = val; }
        public int intVal() { return (int)val; }
        public long longVal() { return (long)val; }
        public float floatVal() { return (float)val; }
        public double doubleVal() { return val; }
        public String stringVal() { return Double.toString(val); }
        public Object constantObject() { return new Double(val); }

        public boolean isDouble() { return true; }
        public ConstVal mathBinOp(int op, ConstVal other) {
            if (other.isString())
                return other.mathBinOpReversed(op, this);
            else if (other.isBoolean())
                throw new WrongConstTypeError(other.getType());
            else switch (op) {
            case Instr.ADD: return make(val+other.doubleVal());
            case Instr.SUB: return make(val-other.doubleVal());
            case Instr.MUL: return make(val*other.doubleVal());
            case Instr.DIV: return make(val/other.doubleVal());
            case Instr.REM: return make(val%other.doubleVal());
            default:
                throw new WrongConstTypeError(other.getType());
            }
        }
        public ConstVal compareBinOp(int op, ConstVal other) {
            if (other.isBoolean() || other.isString())
                throw new WrongConstTypeError(other.getType());
            switch (op) {
            case Instr.CMP_EQ: return make(val == other.doubleVal());
            case Instr.CMP_NE: return make(val != other.doubleVal());
            case Instr.CMP_LT: return make(val < other.doubleVal());
            case Instr.CMP_GE: return make(val >= other.doubleVal());
            case Instr.CMP_GT: return make(val > other.doubleVal());
            case Instr.CMP_LE: return make(val <= other.doubleVal());
            default:
                throw new WrongConstTypeError(other.getType());
            }
        }
        protected ConstVal mathBinOpReversed(int op, ConstVal other) {
            switch (op) {
            case Instr.ADD: return make(other.doubleVal()+val);
            case Instr.SUB: return make(other.doubleVal()-val);
            case Instr.MUL: return make(other.doubleVal()*val);
            case Instr.DIV: return make(other.doubleVal()/val);
            case Instr.REM: return make(other.doubleVal()%val);
            default:
                throw new WrongConstTypeError(other.getType());
            }
        }
        protected ConstVal compareBinOpReversed(int op, ConstVal other) {
            switch (op) {
            case Instr.CMP_EQ: return make(other.doubleVal() == val);
            case Instr.CMP_NE: return make(other.doubleVal() != val);
            case Instr.CMP_LT: return make(other.doubleVal() < val);
            case Instr.CMP_GE: return make(other.doubleVal() >= val);
            case Instr.CMP_GT: return make(other.doubleVal() > val);
            case Instr.CMP_LE: return make(other.doubleVal() <= val);
            default:
                throw new WrongConstTypeError(other.getType());
            }
        }
        public ConstVal numericalNegOp() {
            return make(-val);
        }
        public Instr genLoadCode() {
            return Instr.genConst(val);
        }
        public Type getType() { return Type.double_type; }
    }
    public static final class BooleanConstVal extends ConstVal {
        boolean val;
        BooleanConstVal(boolean val) { this.val = val; }
        public boolean booleanVal() { return val; }
        public int bytecodeIntVal() { return val?1:0; }
        public String stringVal() { return (val?"true":"false"); }
        public Object constantObject() { return new Boolean(val); }

        public boolean isBoolean() { return true; }
        public ConstVal compareBinOp(int op, ConstVal other) {
            switch (op) {
            case Instr.CMP_EQ: return make(val == other.booleanVal());
            case Instr.CMP_NE: return make(val != other.booleanVal());
            default:
                throw new WrongConstTypeError(other.getType());
            }
        }
        public ConstVal logicalNegOp() {
            return make(!val);
        }
        public Instr genLoadCode() {
            return Instr.genConst(val);
        }
        public Type getType() { return Type.boolean_type; }
    }
    public static final class StringConstVal extends ConstVal {
        String val;
        StringConstVal(String val) { this.val = val; }
        public String stringVal() { return val; }
        public Object constantObject() { return val; }

        public boolean isString() { return true; }
        public ConstVal mathBinOp(int op, ConstVal other) {
            if (op == Instr.ADD) {
                return make(this.val+other.stringVal());
            } else
		throw new WrongConstTypeError(other.getType());
        }
        public ConstVal mathBinOpReversed(int op, ConstVal other) {
            if (op == Instr.ADD) {
                return make(other.stringVal()+this.val);
            } else 
		throw new WrongConstTypeError(other.getType());
        }
        public ConstVal compareBinOp(int op, ConstVal other) {
            if (op == Instr.CMP_EQ)
                return make(val.equals(other.stringVal()));
            else if (op == Instr.CMP_NE)
                return make(!(val.equals(other.stringVal())));
            else throw new WrongConstTypeError(other.getType());
        }
        public ConstVal compareBinOpReversed(int op, ConstVal other) {
            return compareBinOp(op, other);
        }
        public Instr genLoadCode() {
            return Instr.genConst(val);
        }
        public Type getType() { return Type.string_type; }
    }
    public static final class NullConstVal extends ConstVal {
        public Object constantObject() { return null; }
        public String stringVal() { return "null"; }

        public boolean isNull() { return true; }
        public ConstVal mathBinOp(int op, ConstVal other) {
            if (op == Instr.ADD && other.isString()) {
                return make(stringVal()+other.stringVal());
            }
            else throw new WrongConstTypeError(other.getType());
        }
        public ConstVal mathBinOpReversed(int op, ConstVal other) {
            if (op == Instr.ADD && other.isString()) {
                return make(other.stringVal()+stringVal());
            }
            else throw new WrongConstTypeError(other.getType());
        }
        public ConstVal compareBinOp(int op, ConstVal other) {
            if (op == Instr.CMP_EQ)
                return make(other instanceof NullConstVal);
            else if (op == Instr.CMP_NE)
                return make(!(other instanceof NullConstVal));
            else throw new WrongConstTypeError(other.getType());
        }
        public Instr genLoadCode() {
            return Instr.genNull();
        }
        public Type getType() { return Type.null_type; }
    }
    public static final class VoidConstVal extends ConstVal {
        public Object constantObject() { return Void.TYPE; }
        public String stringVal() { return "void"; }

        public boolean isVoid() { return true; }
        public Instr genLoadCode() {
            return Instr.genNil("Void constant");
        }
        public Type getType() { return Type.void_type; }
    }
}

/* End of class ConstVal */
