/* Copyright (C) 2003, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*
 * FixedEnvironment.java
 *
 *
 */

package uiuc.Jumbo.Compiler;

import uiuc.Jumbo.Jaemus.*;
import java.util.*;
import java.io.*;
import uiuc.Jumbo.Util.*;

/**
 * Parts of the environment that are fixed within a method.
 * @version      1.0, 6 Feb 2003
 * @author       Lars Clausen
 *
 */
final class FixedEnvironment {
    final CodeEnvironment outer;
    // This one is sideeffected
    final MonoList imports;
    String currentPackage;
    final Hashtable instanceInitializers;
    final Hashtable classInitializers;
    final Local freeVars;
    final ClassInfo currentClass;
    final MethodInfo currentMethod;

    FixedEnvironment() {
        this.outer = null;
        imports = new ArrayMonoList();
        addImportPackage("java/lang");
        this.currentPackage = "";
        this.currentClass = null;
        this.currentMethod = null;
        instanceInitializers = new Hashtable();
        classInitializers = new Hashtable();
        freeVars = null;
    }

    /** Create a new FixedEnvironment from an old FixedEnvironment, 
     * resetting all local info. */
    FixedEnvironment(FixedEnvironment old) {
        this.outer = null;
        this.currentClass = null;
        this.currentMethod = null;
        this.imports = new ArrayMonoList();
        addImportPackage("java/lang");
        this.currentPackage = "";

        this.instanceInitializers = old.instanceInitializers;
        this.classInitializers = old.classInitializers;
        freeVars = null;
    }

    /** Create a new FixedEnvironment from an old CodeEnvironment,
     *  with a new class */
    FixedEnvironment(CodeEnvironment old, ClassInfo c, Local free) {
        if (old.getCurrentClass() != null) 
            this.outer = old;
        else
            this.outer = null;
        this.currentClass = c;
        this.currentMethod = null;

        this.imports = old.fixed.imports;
        this.currentPackage = old.fixed.currentPackage;
        this.instanceInitializers = old.fixed.instanceInitializers;
        this.classInitializers = old.fixed.classInitializers;
        this.freeVars = free;
    }

    /** Create a new FixedEnvironment from an old one, with a new method */
    FixedEnvironment(FixedEnvironment old, MethodInfo newmethod) {
        this.currentMethod = newmethod;

        this.outer = old.outer;
        this.imports = old.imports;
        this.currentClass = old.currentClass;
        this.currentPackage = old.currentPackage;
        this.instanceInitializers = old.instanceInitializers;
        this.classInitializers = old.classInitializers;
        this.freeVars = old.freeVars;
    }

    /** Create a new FixedEnvironment from an old one, with a new package */
    FixedEnvironment(FixedEnvironment old, String p) {
        this.currentPackage = p;

        this.currentMethod = old.currentMethod;
        this.outer = old.outer;
        this.imports = new ArrayMonoList();
        this.imports.add(p);
        this.imports.addAll(old.imports);
        this.currentClass = old.currentClass;
        this.instanceInitializers = old.instanceInitializers;
        this.classInitializers = old.classInitializers;
        this.freeVars = old.freeVars;
    }

    public void addInitializer(Code initval, boolean isInstance) {
        if (isInstance) {
            MonoList inits = (MonoList)instanceInitializers.get(getSelfClass());
            if (inits == null) inits = new ArrayMonoList();
            inits = inits.add(initval);
            instanceInitializers.put(getSelfClass(), inits);
        } else {
            MonoList inits = (MonoList)classInitializers.get(getSelfClass());
            if (inits == null) inits = new ArrayMonoList();
            inits = inits.add(initval);
            classInitializers.put(getSelfClass(), inits);
        }
    }

    public MonoList getInstanceInitializers() {
        ClassInfo c = getSelfClass();
        if (c == null) return null;
        return (MonoList)instanceInitializers.get(c);
    }

    public MonoList getStaticInitializers() {
        ClassInfo c = getSelfClass();
        if (c == null) return null;
        return (MonoList)classInitializers.get(c);
    }

    /** Try to find a class directly or via imports.
     */
    public ClassInfo get(String classname) throws ClassNotFoundException {
        ClassInfo c = ClassTable.getKnown(classname);
        if (c != null) return c;
        MonoList l = new ArrayMonoList(imports.size()+1);
        // Since we're trying the imports, classname must be
        // a non-FQCN.  Therefore, we must replace . and / with $
        // to find inner classes.
        
        String dollarname = classname.replace('.', '$').replace('/', '$');
        for (Iterator i = imports.iterator(); i.hasNext(); ) {
            String s = (String)i.next();
            if (s.endsWith("/")) {
                String fullname = s+dollarname;
                c = ClassTable.getKnown(fullname);
                if (c != null) return c;
                else l = l.add(fullname);
            } else {
                if (s.endsWith("/"+dollarname)) {
                    c = ClassTable.get(s);
                    if (c != null) return c;
                    else l = l.add(s);
                }
            }
        }
        c = ClassTable.get(classname);
        if (c != null) return c;
        for (Iterator i = l.iterator(); i.hasNext(); ) {
            String s = (String)i.next();
            c = ClassTable.get(s);
            if (c != null) return c;
        }

        throw new ClassNotFoundException("Class "+classname+" not found");
    }

    public FixedEnvironment startCompilationUnit() {
        return new FixedEnvironment(this);
    }

    public void addImport(String classname) {
        imports.add(classname.replace('.', '/'));
    }
  
    public void addImportPackage(String packagename) {
        imports.add(packagename+"/");
    }

    public Type getSelfType() {
        if (currentClass != null)
            return new Type(currentClass.getName());
        else
            return null;
    }

    /** Returns the lexically enclosing class.
     * @deprecated Use getCurrentClass() instead
     */
    public ClassInfo getSelfClass() {
        return getCurrentClass();
    }
    public ClassInfo getCurrentClass() {
        return currentClass;
    }
    
    public MethodInfo getCurrentMethod() {
        return currentMethod;
    }

    public String fullClassName(String name) {
        final CodeEnvironment outer = getOuter();
        if (outer != null && getCurrentClass() != null) {
            return getCurrentClass().getName()+"/"+name;
        } else {
            String s = getCurrentPackage();
            return s+name;
        }
    }
    public String getCurrentPackage() {
        return (currentPackage != null?currentPackage:"");
    }
    public CodeEnvironment getOuter() {
        return outer;
    }
    /** Returns the number of outer classes enclosing this env */
    public int numEnclosingClasses() {
        if (outer == null) return 0;
        return outer.fixed.numEnclosingClasses()+1;
    }

    public Local getFree() {
        return freeVars;
    }
}
