/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*
 * ClassTable.java
 *
 *
 */

package uiuc.Jumbo.Jaemus;

import java.util.*;
import java.io.*;
import uiuc.Jumbo.Util.*;
import java.util.zip.*;

/**
 * A table of classes that are known to uiuc.Jumbo.Jaemus.  
 * This includes classes that are defined using uiuc.Jumbo.Jaemus, as well
 * as classes loaded from class files.
 * All class names must be fully qualified with slashes.
 * This class can be seen as holding all classes, but implements lazy loading.
 * @version      1.0, 23 Oct 2001
 * @author       Lars Clausen
 * */

public class ClassTable implements Cloneable {
    private static final Hashtable classes = new Hashtable();

    public static void put(ClassInfo c) {
        classes.put(c.getName(), c);
    }

    /** Remove a class from the class table.  This should only be used after
     * a created class has been written, so that the resources can be freed.
     * If the class is referenced later, the class file will be consulted.
     */
    public static void remove(ClassInfo c) {
	classes.remove(c.getName());
    }

    /** Check if a class has already been loaded, and return it if so.
     * @param classname The fully qualified class name to look up, in internal
     * bytecode format (i.e. path names separated by '/')
     */
    public static ClassInfo getKnown(String classname) {
        return (ClassInfo)classes.get(classname);
    }

    /** Try to find a class in the class table or java lookup.
     * @param classname The fully qualified class name to look up, in internal
     * bytecode format (i.e. path names separated by '/')
     */
    public static ClassInfo get(String classname) {
        ClassInfo c = (ClassInfo)classes.get(classname);
        if (c == null) {
            String dotted = classname.replace('/', '.');
            try {
                Class cl = Class.forName(dotted);
                if (cl.isInterface())
                    c = new InterfaceInfo(cl);
                else
                    c = new ClassInfo(cl);
            } catch (ClassNotFoundException e) {
                return null;
            }
            put(c);
        }
        return c;
    }

    private static MonoList classpath;

    private static void readClassPath() {
        if (classpath != null) return;
        String cp = System.getProperty("java.class.path", ".");
        MonoList classpathnames = new ArrayMonoList(cp, File.pathSeparatorChar);
        classpath = new ArrayMonoList(classpathnames.size());
        for (Iterator i = classpathnames.iterator(); i.hasNext(); ) {
            String s = (String)i.next();
            s=s.trim();
            if (s.endsWith(".jar")) {
                try {
                    ZipFile zf = new ZipFile(s);
                    classpath.add(zf);
                } catch (Exception e) {
                    System.err.println("Can't find jar file '"+s+"'");
                    continue;
                }
            } else {
                File f = new File(s);
                if (f.isDirectory())
                    classpath.add(f);
                else {
                    System.err.println("Can't find directory `"+s+"'");
                    continue;
                }
            }
        }
    }

    private static InputStream getClassStream(String classname) {
        for (Iterator i = classpath.iterator(); i.hasNext(); ) {
            Object o = i.next();
            try {
                if (o instanceof ZipFile) {
                    ZipFile zf = (ZipFile)o;
                    ZipEntry ze = zf.getEntry(classname+".class");
                    if (ze != null) {
                        return zf.getInputStream(ze);
                    }
                } else {
                    File dir = (File)o;
                    File f = new File(dir.getPath()+File.separator+classname+".class");
                    if (f.exists()) {
                        return new FileInputStream(f);
                    }
                }
            } catch (Exception e) {}
        }
        return null;
    }
}

/* End of class ClassTable*/
