/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*
 * CodeAttribute.java
 *
 *
 */

package uiuc.Jumbo.Jaemus;

import java.io.*;
import java.util.*;

import uiuc.Jumbo.Util.*;

/**
 * An attribute that holds actual bytecode
 * @version      1.0,  3 Oct 2001
 * @author       Lars Clausen
 *
*/

public class CodeAttribute extends AttributeInfo {
    public Instr code;
    public MonoList exceptions = new ArrayMonoList();
    public MonoList attributes = new ArrayMonoList();
    int max_stack = 50;
    int max_local = 50;
    MethodInfo in_method;
    public CodeAttribute() {
        super("Code");
    }

    public CodeAttribute(MethodInfo m) {
        super("Code");
        in_method = m;
    }

    private void assignOffsets() {
        int offset = 0;
        for (Instr instr = code; instr != null; instr = instr.next) {
            instr.setOffset(offset);
            offset += instr.size();
        }
    }

    public int findMaxStack() {
        /* Start with empty stack */
        /* Update stack for each instruction */
        /* At a branch, store alternative paths and the stack at their starts*/
        /* When reaching the end or an already-checked instruction,
           pick the next stored alternative path */
        /* So, every instruction must hold a stack.  It only needs to be a
           copy for the first instruction reached by a branch (unless we
           want to compare two incoming branches to check validity).
        */
        return 100;
    }

    public int findMaxLocal() {
        int max = in_method.type.arguments.length+
            (in_method.isStatic()?0:1);
        for (Instr instr = code; instr != null; instr = instr.next) {
            int index = instr.getLocalIndex();
            if (index+1 > max) max = index+1;
        }
        return max;
    }

    public void prepareForEmit(ConstantPool c) {
        if (code.handlers != null) {
            exceptions = code.handlers;
        }
        super.prepareForEmit(c);
        int offset = 0;
        int max = in_method.type.arguments.length+
            (in_method.isStatic()?0:1);
        for (Instr instr = code; instr != null; instr = instr.next) {
            instr.setOffset(offset);
            if (instr instanceof NilArg) continue;
            instr.prepareForEmit(c);
            offset += instr.size();
            int index = instr.getLocalIndex()+1;
            if (index > max) max = index;
        }
        max_local = max;
        for (Iterator i = exceptions.iterator(); i.hasNext(); ) {
            ExceptionInfo e = (ExceptionInfo)i.next();
            e.prepareForEmit(c);
        }
        for (Iterator i = attributes.iterator(); i.hasNext();) {
            AttributeInfo f = (AttributeInfo)i.next();
            f.prepareForEmit(c);
        }
    }

    public void emitTo(DataOutputStream o) throws IOException {
        size = 2+2+4+code.totalSize()+2+8*exceptions.size()+2+2*attributes.size();
        emitHeader(o);
        o.writeShort(max_stack);
        //        o.writeShort(findMaxLocal());
        o.writeShort(max_local);
        //                code.prettyPrint();
        o.writeInt(code.totalSize());
        for (Instr instr = code; instr != null; instr = instr.next) {
            if (instr instanceof NilArg) continue;
            instr.emitTo(o);
        }

        o.writeShort(exceptions.size());
        for (Iterator i = exceptions.iterator(); i.hasNext(); ) {
            ExceptionInfo e = (ExceptionInfo)i.next();
            e.emitTo(o);
        }

        o.writeShort(attributes.size());
        for (Iterator i = attributes.iterator(); i.hasNext();) {
            AttributeInfo f = (AttributeInfo)i.next();
            f.emitTo(o);
        }
    }

    public boolean equals(Object o) {
        if (!(o instanceof CodeAttribute)) return false;
        CodeAttribute ca = (CodeAttribute)o;
        // Should check attribuets as well */
        if (!code.equals(ca.code)) return false;
        if (exceptions.size() != ca.exceptions.size()) return false;
        for (Iterator i1 = exceptions.iterator(),
                 i2 = ca.exceptions.iterator(); i1.hasNext(); ) {
            if (!i1.next().equals(i2.next())) return false;
        }
        if (attributes.size() != ca.attributes.size()) return false;
        for (Iterator i1 = attributes.iterator(),
                 i2 = ca.attributes.iterator(); i1.hasNext(); ) {
            if (!i1.next().equals(i2.next())) return false;
        }
        return true;
    }
}

/* End of class CodeAttribute*/
