/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*
 * ConstantPool.java
 *
 *
 */

package uiuc.Jumbo.Jaemus;

import java.util.*;
import java.io.*;

/**
 * A container of constants
 * @version      1.0,  2 Oct 2001
 * @author       Lars Clausen
 *
*/

public class ConstantPool {
    private Hashtable constants;
    private Constant[] indexed;

    public ConstantPool() {
        constants = new Hashtable();
    }
    public ConstantPool(DataInputStream di) 
        throws Exception {
        int count = di.readUnsignedShort();
        indexed = new Constant[count];
        for (int i = 1; i < count; i++) {
            try {
                int constanttype = di.readUnsignedByte();
                switch (constanttype) {
                case Constant.CLASS:
                    indexed[i] = new ConstantClass(di);
                    break;
                case Constant.FIELDREF:
                    indexed[i] = new ConstantFieldRef(di);
                    break;
                case Constant.METHODREF:
                    indexed[i] = new ConstantMethodRef(di);
                    break;
                case Constant.STRING:
                    indexed[i] = new ConstantString(di);
                    break;
                case Constant.INTEGER:
                    indexed[i] = new ConstantInteger(di);
                    break;
                case Constant.FLOAT:
                    indexed[i] = new ConstantFloat(di);
                    break;
                case Constant.LONG:
                    indexed[i] = new ConstantLong(di);
                    break;
                case Constant.DOUBLE:
                    indexed[i] = new ConstantDouble(di);
                    break;
                case Constant.INTERFACE:
                    indexed[i] = new ConstantInterfaceMethodRef(di);
                    break;
                case Constant.NAMEANDTYPE:
                    indexed[i] = new ConstantNameAndType(di);
                    break;
                case Constant.UTF8:
                    indexed[i] = new ConstantUTF8(di);
                    break;
                default: throw new Error("Illegal constant number "+constanttype);
                }
            } catch (Throwable e) {
                System.err.println("Exception reading constant "+i+": "+e);
            }
            // These two types take up "two" spots in the table
            if (indexed[i].isDoubleConstant()) {
                if (i == count-1) {
                    throw new Error("Half double-word at end of constant pool.\n");
                }
                i++;
            }
        }
        constants = new Hashtable(count*2, 0.60f);
        try {
            for (int i = 1; i < count; i++) {
                indexed[i].resolve(indexed);
                constants.put(indexed[i], indexed[i]);
                if (indexed[i].isDoubleConstant()) i++;
            }
        } catch (Throwable e) {
            e.printStackTrace();
        }
    }

    public Constant get(int i) {
        return indexed[i];
    }
    public int size() {
        return indexed.length;
    }

    public Constant addClass(String name) {
        Constant cl = new ConstantClass(name);
        return addConstant(cl);
    }
    public Constant addClass(ClassInfo c) {
        Constant cl = c.getConstant();
        if (cl == null) {
            cl = addClass(c.getName());
            c.setConstant(cl);
        }
        return cl;
    }
    public Constant addMethod(String ofClass, String name, MethodType type) {
        Constant mr = new ConstantMethodRef(ofClass, name, type);
        return addConstant(mr);
    }
    public Constant addMethod(MethodInfo m) {
        Constant mr = m.getConstant();
        if (mr == null) {
            mr = addMethod(m.inClass.name, m.getName(), m.type);
            m.setConstant(mr);
        }
        return mr;
    }
    public Constant addInterface(String ofClass, String name, MethodType type) {
        Constant mr = new ConstantInterfaceMethodRef(ofClass, name, type);
        return addConstant(mr);
    }
    public Constant addInterface(MethodInfo m) {
        Constant mr = m.getConstant();
        if (mr == null) {
            mr = addInterface(m.inClass.name, m.getName(), m.type);
            m.setConstant(mr);
        }
        return mr;
    }
    public Constant addField(String ofClass, String name, String type) {
        Constant fr = new ConstantFieldRef(ofClass, name, type);
        return addConstant(fr);
    }
    public Constant addField(String ofClass, String name, Type type) {
        return addField(ofClass, name, type.bytecodeString());
    }
    public Constant addField(FieldInfo f) {
        Constant fr = f.getConstant();
        if (fr == null) {
            fr = addField(f.inClass.name, f.name, f.type.bytecodeString());
            f.setConstant(fr);
        }
        return fr;
    }
    public Constant addNameAndType(String name, String type) {
        Constant nat = new ConstantNameAndType(name, type);
        return addConstant(nat);
    }
    public Constant addString(String s) {
        ConstantString sr = new ConstantString(s);
        return addConstant(sr);
    }
    public Constant addUTF8(String s) {
        ConstantUTF8 sr = new ConstantUTF8(s);
        return addConstant(sr);
    }
    public Constant addInteger(int i) {
        ConstantInteger ic = new ConstantInteger(i);
        return addConstant(ic);
    }
    public Constant addLong(long i) {
        ConstantLong ic = new ConstantLong(i);
        return addConstant(ic);
    }
    public Constant addFloat(float i) {
        ConstantFloat ic = new ConstantFloat(i);
        return addConstant(ic);
    }
    public Constant addDouble(double i) {
        ConstantDouble ic = new ConstantDouble(i);
        return addConstant(ic);
    }

    public Constant addConstant(String s) {
        return addString(s);
    }

    public Constant addConstant(int i) {
        return addInteger(i);
    }
    public Constant addConstant(long i) {
        return addLong(i);
    }
    public Constant addConstant(float i) {
        return addFloat(i);
    }
    public Constant addConstant(double i) {
        return addDouble(i);
    }
    public Constant addConstant(MethodInfo m) {
        return addMethod(m);
    }
    public Constant addConstant(FieldInfo f) {
        return addField(f);
    }
    public Constant addConstant(ClassInfo c) {
        return addClass(c);
    }

    public Constant getClass(String name) { 
        Constant c = new ConstantClass(name);
        return (Constant)constants.get(c);
    }
    public Constant getString(String name) {
        return (Constant)constants.get(new ConstantString(name));
    }

    public Constant addConstant(Constant c) {
        Constant cr = (Constant)constants.get(c);
        if (cr == null) {
            constants.put(c, c);
            makeIndex(c);
            c.onInsertion(this);
            return c;
        }
        return cr;
    }

    List early;
    List late;
    int index;

    public void makeIndex(Constant c) {
        if (early != null) {
            if (c.isLdcConstant()) {
                early.add(c);
                c.index = index++;
                if (c.isDoubleConstant())
                    index++;
            }
            else late.add(c);
        }
    }

    /** Prepare to emit the constant pool. 
     * Note that the 'early' constants (those that can be loaded with LDC) are indexed as they
     * come in, which allows us to figure out instruction indices correctly. */
    public void prepareForEmit() {
        early = new ArrayList();
        late = new ArrayList();
        index = 1;
    
        for (Enumeration i = constants.elements(); i.hasMoreElements(); ) {
            Constant c = (Constant)i.nextElement();
            if (c.isLdcConstant()) {
                early.add(c);
                c.index = index++;
                if (c.isDoubleConstant())
                    index++;
            }
            else late.add(c);
        }
    }

    /** Emit the contents into a class file.
     * To reduce the number of LDC_W instructions, those that can be referenced
     * by LDC are put first in the constant pool.
     */
    public void emitTo(DataOutputStream o) throws IOException {
	Constant spare = null;
        for (Iterator i = late.iterator(); i.hasNext(); ) {
            Constant c = (Constant)i.next();
            c.index = index++;
	    if(c instanceof ConstantUTF8)
		spare = c;
            // Non-ldc constants cannot be double
        }
        o.writeShort(index);
        int in = 0;
        for (Iterator i = early.iterator(); i.hasNext(); ) {
            Constant c = (Constant)i.next();
	    
	    if(c instanceof ConstantMethodRef)
		if(((ConstantMethodRef)c).name.equals("<clinit>") && spare != null){
		    //System.out.println("Skipped emitting <clinit>");
		    c = spare;
		}
	    
            c.emitConstant(o);
            in++;
        }
	//System.out.println("in:"+in);
        for (Iterator i = late.iterator(); i.hasNext(); ) {
            Constant c = (Constant)i.next();

	    if(c instanceof ConstantMethodRef)
		if(((ConstantMethodRef)c).name.equals("<clinit>") && spare != null){
		    //System.out.println("Skipped emitting <clinit>");
		    c = spare;
		}

            c.emitConstant(o);
            in++;
        }
        early = late = null;
    }
}

/* End of class ConstantPool*/
