/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*
 * FieldInfo.java
 *
 *
 */

package uiuc.Jumbo.Jaemus;

import java.util.*;
import java.io.*;
import java.lang.reflect.*;

import uiuc.Jumbo.Util.*;

/**
 * A class describing a field of a java class
 * @version      1.0,  2 Oct 2001
 * @author       Lars Clausen
 *
*/

public class FieldInfo implements ClassMember {
    int flags;
    String name;
    Type type;
    MonoList attributes = new ArrayMonoList();
    ConstantFieldRef fieldconstant;
    ClassInfo inClass;
    Object constval;

    public FieldInfo(ClassInfo inClass, int flags, String name, Type type) {
        this.flags = flags;
        this.name = name;
        this.type = type;
        this.inClass = inClass;
        if (inClass == null) throw new Error("Null class for "+name);
    }

    public FieldInfo(ClassInfo inClass, Field f) {
        this.inClass = inClass;
        this.flags = f.getModifiers();
        this.name = f.getName();
        this.type = Type.convertType(f.getType());
        if (this.isStaticFinal()) {
            try {
                constval = f.get(null);
            } catch (Exception e) {}
        }
        if (inClass == null) throw new Error("Null class for "+name);
    }

    public FieldInfo(DataInputStream di, ConstantPool pool) 
        throws Exception {
        flags = di.readUnsignedShort();
        int cindex = di.readUnsignedShort();
        if (cindex < 1 || cindex >= pool.size()) {
            throw new Error("Illegal field name constant "+cindex);
        } 
        Constant n = pool.get(cindex);
        if (!(n instanceof ConstantUTF8)) {
            throw new Error("Non-UTF8 field name constant "+n);
        } 
        name = n.toString();

        cindex = di.readUnsignedShort();
        if (cindex < 1 || cindex >= pool.size()) {
            throw new Error("Illegal field signature constant "+cindex);
        } 
        Constant s = pool.get(cindex);
        if (!(s instanceof ConstantUTF8)) {
            throw new Error("Non-UTF8 field signature constant "+s);
        }
        type = Type.parseType(s.toString());
        if (type == null) {
            throw new Error("Illegal field signature "+s.toString());
        }
        System.out.println("Field "+name+" in "+inClass);
        int count = di.readUnsignedShort();
        for (int i = 0; i < count; i++) {
            int nameindex = di.readUnsignedShort();
            int length = di.readInt();
            String name = pool.get(nameindex).toString();
            /* We're only interested in the ConstantValue attribute */
            if (name.equals("ConstantValue")) {
                int constantindex = di.readUnsignedShort();
                makeConstant(pool.get(constantindex));
            } else {
                di.skipBytes(length);
            }
        }
    }

    public Constant getConstant() { return fieldconstant; }
    public void setConstant(Constant c) { fieldconstant = (ConstantFieldRef)c; }

    /** Set this field value to be a constant.
     * @param o The constant value (encapsulated for primitive values)
     */
    public void makeConstant(Object o) { constval = o; }
    public boolean isConstant() { return constval != null; }
    public Object constantValueObject() { return constval; }

    public ClassInfo getClassInfo() { return inClass; }
    public String getName() { return name; }
    public Type getType() { return type; }
    public boolean isStatic() { return (flags & Flags.STATIC) != 0; }
    public boolean isFinal() { return (flags & Flags.FINAL) != 0; }
    public boolean isStaticFinal() {
        return Flags.bitsSet(flags, Flags.STATIC|Flags.FINAL); 
    }

    public void prepareForEmit(ConstantPool c) {
        fieldconstant = (ConstantFieldRef)c.addField(this);
        if (isConstant()) {
            attributes.add(new ConstantValueAttribute(constantValueObject()));
        }
        for (Iterator i = attributes.iterator(); i.hasNext();) {
            AttributeInfo f = (AttributeInfo)i.next();
            f.prepareForEmit(c);
        }
    }

    public void getAdded(ClassInfo c) {
        c.addMember(this);
    }

    public void emitTo(DataOutputStream o) throws IOException {
        Constant nameconstant = fieldconstant.nameAndType.name_constant;
        Constant typeconstant = fieldconstant.nameAndType.type_constant;
        o.writeShort(flags);
        nameconstant.emitIndex(o);
        typeconstant.emitIndex(o);

        o.writeShort(attributes.size());
        for (Iterator i = attributes.iterator(); i.hasNext();) {
            AttributeInfo f = (AttributeInfo)i.next();
            f.emitTo(o);
        }
    }

    public String toString() {
        return inClass+"."+name;
    }
}

/* End of class FieldInfo*/
