/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*
 * TestConstantPool.java
 *
 *
 */

package uiuc.Jumbo.Jaemus;

import junit.framework.TestCase;

import uiuc.Jumbo.Jaemus.*;
import java.lang.reflect.*;

// classes imported for testing purposes
import java.awt.Point;
import java.util.Date;

public class TestConstantPool extends TestCase {
    /* Note that testing genConst is pointless because we don't set its 
     * arg to be the result from addConstant.
     */
    private ConstantPool constantPool;

    public TestConstantPool(String name) {
        super(name);
        constantPool = new ConstantPool();
    }

    public void testAddConstantIntegers() {
        int n = 7324234;
        Constant c1 = constantPool.addConstant(n);
        Constant c2 = constantPool.addConstant(n);

        assertEquals(c1, c2);
        assertSame(c1, c2);
    }

    public void testAddConstantLongs() {
        long n = 2234324323234l;
        Constant c1 = constantPool.addConstant(n);
        Constant c2 = constantPool.addConstant(n);

        assertEquals(c1, c2);
        assertSame(c1, c2);
    }

    public void testAddConstantFloats() {
        float n = 1324134.1134134134f;
        Constant c1 = constantPool.addConstant(n);
        Constant c2 = constantPool.addConstant(n);

        assertEquals(c1, c2);
        assertSame(c1, c2);
    }

    public void testAddConstantDoubles() {
        double n = 23423432431414143314143.134134141413414143134134;
        Constant c1 = constantPool.addConstant(n);
        Constant c2 = constantPool.addConstant(n);

        assertEquals(c1, c2);
        assertSame(c1, c2);
    }

    public void testClassAbstractNoSuper() {
        checkClass( Number.class, "java/lang/Number" );
    }

    public void testClassNoSuper() {
        checkClass( Thread.class, "java/lang/Thread" );
    }

    public void testClassWithSuper() {
        checkClass( Float.class, "java/lang/Float" );
    }

    public void testMethodOneArg() {
        checkMethod( Object.class, "java/lang/Object", "equals", 
                     Type.boolean_type, new Type[] {Type.object_type}, 
                     new Class[] {Object.class} );
    }

    public void testMethodVoidNoArg() {
        checkMethod( Object.class, "java/lang/Object", "notify", 
                     Type.void_type, new Type[] {}, 
                     new Class[] {} );
    }

    public void testMethodNoArg() {
        checkMethod( Object.class, "java/lang/Object", "toString", 
                     Type.string_type, new Type[] {}, 
                     new Class[] {} );
    }

    public void testMethodPrimitiveArgs() {
        checkMethod( Integer.class, "java/lang/Integer", "toBinaryString", 
                     Type.string_type, new Type[] {Type.int_type}, 
                     new Class[] { Integer.TYPE } );
    }

    public void testMethodWithMultipleSignatures() {
        checkMethod( String.class, "java/lang/String", "indexOf", 
                     Type.int_type, new Type[] {Type.string_type, 
                                                Type.int_type},
                     new Class[] { String.class, Integer.TYPE } );
    }

    public void testStaticPublicFinalPrimitiveField() {
        checkField( Double.class, "java/lang/Double", 
                    "NEGATIVE_INFINITY", Type.double_type );
    }

    public void testStaticPublicFinalObjectField() {
        checkField( System.class, "java/lang/System", 
                    "out", new Type("java/io/PrintStream") );
    }

    public void testPublicField() {
        checkField( Point.class, "java/awt/Point", "x", Type.int_type );
    }

    public void testSimpleNameAndType() {
        checkNameAndType("foo", Type.int_type);
    }

    public void testArrayNameAndType() {
        checkNameAndType("baz", new Type(new Type(Type.int_type)) );
    }

    private void checkNameAndType(String name, Type type) {
        Constant c1 = constantPool.addNameAndType(name, type.bytecodeString());
        Constant c2 = constantPool.addNameAndType(name, type.bytecodeString());

        assertEquals(c1, c2);
        assertSame(c1, c2);
    }

    private void checkClass( Class c, String name ) {
        Constant c1 = constantPool.addClass(name);

        ClassInfo cli1 = new ClassInfo(c);
        Constant c2 = constantPool.addClass(cli1);

        assertEquals(name + " addClass with ClassInfo equals", c1, c2);
        assertSame(name + " addClass with ClassInfo same", c1, c2);

        Constant c3 = constantPool.addConstant(cli1);
        assertEquals(name + " addConstant with ClassInfo equals", c1, c3);
        assertSame(name + " addConstant with ClassInfo same", c1, c3);
    }

    private void checkField(Class c, String className, 
                            String fieldName, Type fieldType) {
        String messagePrefix = className + "." + fieldName + ": " + 
                                fieldType.bytecodeString();

        Constant c1 = constantPool.addField(className, fieldName, fieldType);
        Constant c2 = constantPool.addField(className, fieldName, 
                                            fieldType.bytecodeString());

        assertEquals(messagePrefix + " addField with type bytestring equals", 
                     c1, c2);
        assertSame(messagePrefix + " addField with type bytestring same", 
                    c1, c2);

        Constant c3;

        try {
            ClassInfo cli = new ClassInfo(c);
            FieldInfo fi = new FieldInfo(cli, c.getDeclaredField(fieldName));
            c3 = constantPool.addField(fi);
        }
        catch(NoSuchFieldException e) {
            fail(messagePrefix + " checkField: field does not exist!");
            return;
        }

        assertEquals(messagePrefix + " addField with FieldInfo equals", 
                        c1, c3);
        assertSame(messagePrefix + " addField with FieldInfo same", 
                        c1, c3);

    }

    private void checkMethod(Class c, String className, 
                             String methodName, Type return_type, 
                             Type[] type_params, Class[] class_params) {

        MethodType type = new MethodType(return_type, type_params);
        String messagePrefix = className + "." + methodName + ":" + 
                                type.bytecodeString();

        Constant c1 = constantPool.addMethod(className, methodName, type);
        Constant c2 = constantPool.addMethod(className, methodName,
                                             type.bytecodeString());

        assertEquals(messagePrefix + " addMethod with type bytestring equals", 
                        c1, c2);
        assertSame(messagePrefix + " addMethod with type bytestring same", 
                        c1, c2);

        Constant c3;

        try {
            ClassInfo cli = new ClassInfo(c);
            MethodInfo mi = new MethodInfo(cli, 
                                c.getDeclaredMethod(methodName, class_params));
            c3 = constantPool.addMethod(mi);
        }
        catch(NoSuchMethodException e) {
            fail(messagePrefix + " checkMethod: method does not exist!");
            return;
        }

        assertEquals(messagePrefix + " addMethod with MethodInfo equals", 
                        c1, c3);
        assertSame(messagePrefix + " addMethod with MethodInfo same", 
                        c1, c3);
    }
}
