/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*
 * Type.java
 *
 *
 */

package uiuc.Jumbo.Jaemus;

import java.util.*;

/**
 * Class that describes the various types available in Java.
 * @version      1.0,  5 Oct 2001
 * @author       Lars Clausen
 *
*/

public final class Type implements java.io.Serializable{
    // This ordering is fixed because it's used to generate bytecode value
    /** A type for class indicators 
     * This is used to indicate the type of a QualifiedNameOrClass expression
     * that indicates a class.  It doesn't correspond to a stack object, but
     * allows type checks and member lookups. */
    public static final byte CLASS   = -2;
    /** A type that represents the unreachable code */
    public static final byte BOTTOM  = -1;
    /** A type for NULL */
    public static final byte NULL    = 0;
    public static final byte ARRAY   = 1;
    public static final byte OBJECT  = 2;
    public static final byte VOID    = 3;
    public static final byte BOOLEAN = 4;
    public static final byte CHAR    = 5;
    public static final byte FLOAT   = 6;
    public static final byte DOUBLE  = 7;
    public static final byte BYTE    = 8;
    public static final byte SHORT   = 9;
    public static final byte INT     = 10;
    public static final byte LONG    = 11;
    /** This determines the first letter of a bytecode type (the overall type)
     *  Possible values are:
     *  BYTE   - 8-bit unsigned integer
     *  CHAR   - A character
     *  SHORT  - 16-bit signed integer
     *  INT    - 32-bit signed integer
     *  LONG   - 64-bit signed integer
     *  FLOAT  - 32-bit IEEE real number
     *  DOUBLE - 64-bit IEEE real number
     *  ARRAY  - Any array type
     *  OBJECT - Any object type
     *  VOID   - The void type, used only in method returns
     */
    public final int kind;
    /** This is the type of the elements of an array */
    public final Type array_contents;
    /** This is the name of an object type */
    public final String class_name;

    public static final Type bottom = new Type(BOTTOM);
    public static final Type byte_type = new Type(BYTE);
    public static final Type char_type = new Type(CHAR);
    public static final Type short_type = new Type(SHORT);
    public static final Type int_type = new Type(INT);
    public static final Type long_type = new Type(LONG);
    public static final Type float_type = new Type(FLOAT);
    public static final Type double_type = new Type(DOUBLE);
    public static final Type void_type = new Type(VOID);
    public static final Type boolean_type = new Type(BOOLEAN);

    public static final Type null_type = new Type(NULL);
    public static final Type string_type = new Type("java/lang/String");
    public static final Type stringbuffer_type = new Type("java/lang/StringBuffer");
    public static final Type object_type = new Type("java/lang/Object");
    public static final Type throwable_type = new Type("java/lang/Throwable");
    public static final Type class_type = new Type("java/lang/Class");
    
    public Type(byte kind) {
        this.kind = kind;
        this.array_contents = null;
        this.class_name = null;
    }

    public Type(Type element_type) {
        this.kind = ARRAY;
        this.array_contents = element_type;
        this.class_name = null;
    }

    public Type(String class_name) {
        this.kind = OBJECT;
        this.array_contents = null;
        this.class_name = class_name;
    }

    /** Make a new type for either an object or a class indicator */
    public Type(String class_name, boolean is_class_name) {
        this.kind = (is_class_name?CLASS:OBJECT);
        this.array_contents = null;
        this.class_name = class_name;
    }

    public Type(ClassInfo cl) {
        this.kind = OBJECT;
        this.array_contents = null;
        this.class_name = cl.name;
    }

    /** Create a type from a java style class object */
    public static Type convertType(Class cl) {
        if (cl.isPrimitive()) {
            if (cl == Boolean.TYPE) return boolean_type;
            if (cl == Byte.TYPE) return byte_type;
            if (cl == Character.TYPE) return char_type;
            if (cl == Short.TYPE) return short_type;
            if (cl == Integer.TYPE) return int_type;
            if (cl == Long.TYPE) return long_type;
            if (cl == Float.TYPE) return float_type;
            if (cl == Double.TYPE) return double_type;
            if (cl == Void.TYPE) return void_type;
            throw new Error("Unknown primitive type "+cl);
        }
        if (cl.isArray()) {
            return new Type(convertType(cl.getComponentType()));
        }
        return new Type(cl.getName().replace('.', '/'));
    }

    public boolean equals(Object other) {
        if (!(other instanceof Type)) return false;
        Type t = (Type)other;
        if (kind != t.kind) return false;
        if (kind == OBJECT || kind == CLASS)
            return class_name.equals(t.class_name);
        if (kind == ARRAY) return array_contents.equals(t.array_contents);
        return true;
    }

    public boolean hasClassName() {
        if (kind == OBJECT || kind == CLASS) return true;
        else if (kind == ARRAY) return array_contents.hasClassName();
        else return false;
    }

    public String getClassName() {
        if (kind == OBJECT || kind == CLASS) return class_name;
        else if (kind == ARRAY) return array_contents.getClassName();
        else throw new Error("Bad boy!  Called getClassName() without checking with hasClassName() first!  (On "+this+")");
    }

    public Type replaceClassName(String newname) {
        if (kind == OBJECT || kind == CLASS) return new Type(newname);
        else if (kind == ARRAY) return new Type(array_contents.replaceClassName(newname));
        else return this;
    }

    /** Returns true if this type is an object of the specified class. 
     * Doesn't return true for subclasses.
     */
    public boolean isObjectType(String class_name) {
        if (kind == NULL) return true;
        else if (kind != OBJECT && kind != CLASS) return false;
        else return this.class_name.equals(class_name);
    }

    /** The name of this type as used for field lookup
     */
    public String className() {
        if (kind == ARRAY) return "java/lang/Object";
        else if (kind == OBJECT || kind == CLASS) return class_name;
        else if (kind == NULL) return "java/lang/Object";
        else return bytecodeString();
    }

    /** The name of this type as used for casting
     */
    public String castName() {
        if (kind == OBJECT || kind == CLASS) return class_name;
        else if (kind == NULL) return "java/lang/Object";
        else return bytecodeString();
    }

    public boolean isArray() {
        return kind == ARRAY;
    }

    public int getDimensions() {
        if (kind == ARRAY) return array_contents.getDimensions()+1;
        else return 0;
    }

    public Type getElementType() {
        if (kind != ARRAY) throw new Error("Taking element type of non-array");
        else return array_contents;
    }

    public boolean isPrimitive() {
        return kind > VOID;
    }

    public boolean isVoid() {
        return kind == VOID;
    }

    public boolean isReference() {
        return kind < VOID;
    }

    public boolean isTwoWordType() {
        return kind == LONG || kind == DOUBLE;
    }

    public boolean isNumeric() {
        return kind >= CHAR;
    }

    public boolean isInteger() {
        return kind == CHAR || kind >= BYTE && kind <= INT;
    }

    public boolean isDecimal() {
        return kind == CHAR || kind >= BYTE;
    }

    public int wordSize() {
        return (isTwoWordType()?2:(kind==VOID?0:1));
    }

    public int bytecodeKind() {
        switch (kind) {
        case NULL:
        case ARRAY:
        case OBJECT:
            return OBJECT;
        case VOID:
        case CLASS:
            return VOID;
        case BOOLEAN:
        case CHAR:
        case BYTE:
        case SHORT:
        case INT:
            return INT;
        case LONG:
        case FLOAT:
        case DOUBLE:
        case BOTTOM:
            return kind;
        default:
            throw new Error("Unknown kind "+kind);
        }
    }

    /** A Least Upper Bound on primitive values.
        These conversions are always possible, though sometimes you
        lose precision.
    */
    public Type lub(Type other) {
        if (kind == DOUBLE) return double_type;
        if (other.kind == DOUBLE) return double_type;
        if (kind == FLOAT) return float_type;
        if (other.kind == FLOAT) return float_type;
        if (kind == LONG) return long_type;
        if (other.kind == LONG) return long_type;
        return this;
    }

    /** A more general LUB, also handles references types.
     * May throw an error if there's a mismatch (primitive vs. reference)
     */
    public Type commonAncestor(Type other) {
        if (conformsMethod(other)) return other;
        if (other.conformsMethod(this)) return this;
        if (!isPrimitive() &&
            !other.isPrimitive())
            return object_type;
        throw new Error("Can't LUB "+this+" and "+other);
    }

    /** A LUB of bottom values.
     */
    public Type bottomLub(Type other) {
        if (this.kind == BOTTOM && other.kind == BOTTOM) return bottom;
        if (this.kind == BOTTOM) return other;
        return this;
    }

  /** Returns true if this type conforms (for array assignment) 
   * to the given type
   */
  public boolean conformsArray(Type other) {
    return conformsAssign(other);
  }

    /** Returns true if this type conforms (for assignment) to the given type.
     */
    public boolean conformsAssign(Type other) {
        // Assignment-conform is like method-conform, except it can also
        // do narrowing
        if (conformsMethod(other)) return true;
        else if (kind == INT &&
                 (other.kind == BYTE ||
                  other.kind == SHORT ||
                  other.kind == CHAR)) return true;
        else return false;
    }

    /** Returns true if this type conforms (for arguments) to the given type.
     */
    public boolean conformsMethod(Type other) {
        if (!isReference()) {
            if (kind == other.kind) return true;
            switch (kind) {
            case BOOLEAN:
                return false; // Only boolean conforms to boolean
            case BYTE:
                return other.kind == SHORT ||
                    other.kind == INT ||
                    other.kind == LONG ||
                    other.kind == FLOAT ||
                    other.kind == DOUBLE;
            case SHORT: 
            case CHAR:
                return other.kind == INT ||
                    other.kind == LONG ||
                    other.kind == FLOAT ||
                    other.kind == DOUBLE;
            case INT:
                return other.kind == LONG ||
                    other.kind == FLOAT ||
                    other.kind == DOUBLE;
            case LONG:
                return other.kind == FLOAT ||
                    other.kind == DOUBLE;
            case FLOAT:
                return other.kind == DOUBLE;
            case DOUBLE:
                return false; // Only double conforms to double
            case VOID:
                throw new Error("Trying to conform void type to "+other);
            default:
                throw new Error("Unknown non-reference type "+this+" for conform to "+other);
            }
        }
        if (kind == NULL) return other.isReference();
        if (kind == ARRAY) {
            if (other.kind != ARRAY) 
                return other.isObjectType("java/lang/Object");
            return array_contents.conformsMethod(other.array_contents);
        }
        if (kind == OBJECT || kind == CLASS) {
          return subtypeOf(other);
        }
        throw new Error("Unknown kind "+kind+" for conform to "+other);
    }

  /** Returns true if this type (an OBJECT or CLASS type) is a subclass
   * of the given type.
   */
  public boolean subtypeOf(Type other) {
    // Would you ever have to conform to NULL?
    if (other.kind != OBJECT && other.kind != CLASS)
      return false;
    if (class_name.equals(other.class_name)) return true;
    ClassInfo ci = ClassTable.get(class_name);
    if (ci == null) {
        throw new Error("Failed to find class "+class_name+" during conforms test");
    }
    return ci.subclassOf(ClassTable.get(other.class_name));
  }

    public String bytecodeString() {
        switch (kind) {
        case BYTE: return "B";
        case CHAR: return "C";
        case SHORT: return "S";
        case INT : return "I";
        case LONG: return "J";
        case FLOAT: return "F";  
        case DOUBLE: return "D"; 
        case ARRAY: return "["+array_contents.bytecodeString();  
        case OBJECT:
        case CLASS:
            return "L"+class_name+";"; 
        case VOID: return "V";
        case BOOLEAN: return "Z";
        case NULL: return "Ljava/lang/Object;";
        default: throw new Error("Unknown kind "+kind);
        }
    }

    public String toString() {
        switch (kind) {
        case BOTTOM: return "BOTTOM";
        case BYTE: return "byte";
        case CHAR: return "char";
        case SHORT: return "short";
        case INT : return "int";
        case LONG: return "long";
        case FLOAT: return "float";  
        case DOUBLE: return "double"; 
        case ARRAY: return array_contents.toString()+"[]";
        case OBJECT: return class_name; 
        case CLASS: return class_name+"*"; 
        case VOID: return "void";
        case BOOLEAN: return "boolean";
        case NULL: return "null";
        default: throw new Error("Unknown kind "+kind);
        }
    }

    public String toSource() {
        switch (kind) {
        case BYTE: return "byte";
        case CHAR: return "char";
        case SHORT: return "short";
        case INT : return "int";
        case LONG: return "long";
        case FLOAT: return "float";  
        case DOUBLE: return "double"; 
        case ARRAY: return array_contents.toSource()+"[]";
        case OBJECT: case CLASS:
            return class_name;
        case VOID: return "void";
        case BOOLEAN: return "boolean";
        default: throw new Error("Type not convertible to source: "+this);
        }
    }

    /** Unit tests */
    public static void unitTest() {
        Test.assertEquals(new Type("java/lang/object").bytecodeString(),
                          "Ljava/lang/object;");
        Test.assertEquals(new Type(new Type(int_type)).bytecodeString(),
                          "[[I");
    }

    public static Type parseType(String t) {
        if (t == null) return null;
        if (t.endsWith("[]"))
            return new Type(parseType(t.substring(0, t.length()-2)));
        if (t.equals("void")) return void_type;
        if (t.equals("boolean")) return boolean_type;
        if (t.equals("byte")) return byte_type;
        if (t.equals("char")) return char_type;
        if (t.equals("short")) return short_type;
        if (t.equals("int")) return int_type;
        if (t.equals("long")) return long_type;
        if (t.equals("float")) return float_type;
        if (t.equals("double")) return double_type;
        return new Type(t);
    }
    
    public static Type parseBytecodeType(String s, int off) {
        if (s == null) return null;
        int dims = 0;
        Type t;
        while (s.charAt(off+dims) == '[') dims++;
        off += dims;
        if (s.charAt(off) == 'L')
            t = new Type(s.substring(off+1, s.indexOf(';', off+1)));
        else 
            switch (s.charAt(off)) {
            case 'B': t = byte_type; break;
            case 'C': t = char_type; break;
            case 'S': t = short_type; break;
            case 'I': t = int_type; break;
            case 'J': t = long_type; break;
            case 'F': t = float_type; break;
            case 'D': t = double_type; break;
            case 'V': t = void_type; break;
            case 'Z': t = boolean_type; break;
            default: return null;
            }
        while (dims > 0) {
            t = new Type(t);
            dims--;
        }
        return t;
    }
    public static int bytecodeTypeLength(String t, int off) {
        while (t.charAt(off) == '[') off++;
        if (t.charAt(off) == 'L') return t.indexOf(';', off)+1;
        else return off+1;
    }
    public final static Hashtable typeclasses = new Hashtable();
    static {
        typeclasses.put("void", "Void");
        typeclasses.put("byte", "Byte");
        typeclasses.put("short", "Short");
        typeclasses.put("char", "Character");
        typeclasses.put("int", "Integer");
        typeclasses.put("long", "Long");
        typeclasses.put("float", "Float");
        typeclasses.put("double", "Double");
    }

}

/* End of class Type*/
