/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uiuc.Jumbo.Ast;

import uiuc.Jumbo.Util.*;
import java.util.*;

/**
 * Represents an anonymous class instance creation.
 *
 * @author      Lars Clausen
 * @version     %I%, %G%
 * @since       JDK 1.3
 */

public class AnonymousClassCreation extends ClassInstanceCreation 
    implements ClassDefiner, InnerClass {
    /**
     * @param c  name of the class being instanced
     * @param a  optional argument list
     */
    public AnonymousClassCreation( Identifier c, ArgumentList a, ClassBody b ) { 
	super(c, a);
	setBody(b);
    }

    protected boolean replaceUC(Node old_child, Node new_child) {
	if (old_child == body) {
	    setBody((ClassBody)new_child);
	    return true;
	}
	return super.replaceUC(old_child, new_child);
    }

    private void setBody(ClassBody b) {
	body = b;
	setParent(b);
    }

    public IR toIR() {
	MonoList list = new ArrayMonoList()
	    .add(IR.make(class_name.toString()));
	if(arguments == null)
	    list.add(IR.make(new ArrayMonoList()));
	else
	    list.add(arguments.toIR());
	list.add(body.toIR());
	return IR.make("anonymousClass", list);
    }

    public void toSource() {
	print("new ");
	class_name.toSource();
	print("(");
	if(arguments != null)
	    printSourceList(arguments.getList(), ", ");
	print(") ");
	ocurly();
	linebreak();
	body.toSource();
	ccurly();
    }

    /** Find the matching field declaration in the give list of members. */
    protected FieldDefiner getField(String field, List members) {
        for (Iterator i = members.iterator(); i.hasNext(); ) {
            Object o = i.next();
            if (o instanceof FieldDeclaration) {
                FieldDeclaration fd = (FieldDeclaration)o;
                Variable v = fd.findField(field);
                if (v != null) return v;
            }
        }
        return null;
    }

    /** Returns the variable defining the given field, if any.
     */
    public FieldDefiner findField(String field, ClassEnv loader) {
        FieldDefiner f = getField(field, body.members);
        if (f != null) return f;
        try {
            return getSuperClass(loader).findField(field, loader);
        } catch (Error e) {
            try {
                return getEnclosingClass().findField(field, loader);
	    } catch (Error e1) {
		throw new Error("Can't find AST for "+this+"."+field);
	    }
        }
    }

    /** Returns the type of the given field, if any.
     */
    public Type findFieldType(String field, ClassEnv loader) {
        FieldDefiner f = findField(field, loader);
        if (f != null) return f.getType();
	return null;
    }

    /** Find the matching method declaration in the given list of members. */
    protected MethodDeclaration getMethod(String name, List args, List members,
                                          ClassEnv loader) {
        for (Iterator i = members.iterator(); i.hasNext(); ) {
            Object o = i.next();
            if (o instanceof MethodDeclaration) {
                MethodDeclaration md = (MethodDeclaration)o;
                if (md.conformsTo(name, args, loader))
                    return md;
            }
        }
        return null;
    }

    protected ClassDefiner getSuperClass(ClassEnv loader) {
        String supername = getSuperName();
        if (supername == null) return null;
        return loader.getClass(supername);
    }


    /** Find a method based on name and types.
     * @param name The name of the method.  Constructors are named "<init>".
     * @param types A list of Type values representing the types of the 
     *              arguments.
     * @return A method with this name and types.  Null if no method was found.
     */
    public MethodDefiner findMethod(String name, List args,
                                        ClassEnv loader) {
        if (args == null) args = new ArrayList(0);
        MethodDefiner md = getMethod(name, args, body.members, loader);
        if (md != null || name.equals("<init>")) return md;
        try {
	    md = getSuperClass(loader).findMethod(name, args, loader);
	    if (md != null) return md;
        } catch (Error e) {
	}
	try {
	    md = getEnclosingClass().findMethod(name, args, loader);
	    if (md != null) return md;
	} catch (Error e1) {
        }
	throw new Error("Can't find AST for "+this+"."+name+"("+args+")");
    }

    /** Find the return type of a method based on name and types.
     * @param name The name of the method.  Constructors are named "<init>".
     * @param types A list of Type values representing the types of the 
     *              arguments.
     * @return The type of the matching method, or null if this class
     * doesn't define such a method.
     */
    public Type findMethodType(String name, List args, ClassEnv loader) {
	MethodDefiner md = findMethod(name, args, loader);
	if (md != null) return md.getType();
	return null;
    }

    /** Returns an identifier for the name of the super class */
    public String getSuperName() {
        if (class_name == null) return "java.lang.Object";
        return class_name.toString();
    }

    /** Returns the name of this class. */
    public String getName() {
        throw new Error("Anon classes have no name");
    }

    /** Returns the fully qualified class name of this class. */
    public String getFQCN() {
        throw new Error("Anon classes have no name");
    }

    /** Returns a list of names of interfaces */
    public List getInterfaces() {
        return new ArrayList(0);
    }

    /** Returns true if this class is final.  It always is. */
    public boolean isFinal() {
        return true;
    }

    public boolean isInterface() { return false; }

    public String toString() {
        return "new "+class_name.toString()+"() { <"+(body != null?body.members.size():0)+" members> }";
    }

    /**
     * Contents of the class body
     */
    protected ClassBody body;
}
