/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uiuc.Jumbo.Ast;

import java.util.*;

import uiuc.Jumbo.Util.*;
import uiuc.Jumbo.Compiler.*;

/**
 * Represents a class declaration (named, anonymous, inner, etc).
 *
 * @author      Ava Arachne Jarvis
 * @version     %I%, %G%
 * @since       JDK 1.3
 */

public class ClassDeclaration extends TypeDeclaration implements ClassDefiner {
    /**
     * @param mods    Optional list of modifier flags
     * @param cname   Class name
     * @param sname   Optional superclass name
     * @param ifaces  Optional list of interface names
     * @param cbody   Class body (another class)
     */
    public ClassDeclaration( ModifierList mods,
            Identifier cname,
            Identifier sname,
            List ifaces,
            ClassBody cbody) 
    { 
        super(mods, cname);

        setSuperName(sname);
        interfaces = ifaces;
        setClassBody(cbody);
    }

    public boolean replaceUC(Node old_child, Node new_child) {
        if (old_child == super_name) {
            setSuperName((Identifier)new_child);
            return true;
        }
        else if (old_child == class_body) {
            setClassBody((ClassBody)new_child);
            return true;
        }
        else {
            return replaceInList(interfaces, old_child, new_child);
        }
    }

    private void setSuperName(Identifier sname) {
        super_name = sname;
        setParent(super_name);
    }

    private void setClassBody(ClassBody cbody) {
        class_body = cbody;
        setParent(class_body);
    }

    /** Outputs source code for this class definition.
    */
    public void toSource() {
        if (modifiers != null)
            modifiers.toSource();
        print("class ");
        name.toSource();
        if (super_name != null) {
            print( " extends ");
            super_name.toSource();
        }
        if (interfaces != null && interfaces.size() != 0) {
            print( " implements ");
            printSourceList(interfaces, ", ");
        }
        print(" ");
        ocurly();
        linebreak();
        class_body.toSource();
        ccurly();
    }

    public IR toIR() {
	
        IRCode ir = IR.make("defineClass", new ArrayMonoList());
        ir.addArg(modifiers != null?modifiers.toIR():IR.make(0));
        ir.addArg(name.toIR());
        ir.addArg(super_name != null?super_name.toIR():IR.make(""));
        ir.addArg(toIRList(interfaces));
        return ir.addArg(class_body.toIR());
    }

    public boolean needsSemicolon() { return false; }

    /** Returns the variable defining the given field, if any.
    */
    public FieldDefiner findField(String field, ClassEnv loader) {
        FieldDefiner f = getField(field, class_body.members);
        if (f != null) return f;
        String supername = getSuperName();
        if (supername == null) return null;
        try {
            return loader.getClass(supername).findField(field, loader);
        } catch (Error e) {
            throw new Error("Can't find field "+this+"."+field);
        }
    }

    /** Returns the type of the given field, if any.
     */
    public Type findFieldType(String field, ClassEnv loader) {
	FieldDefiner f = findField(field, loader);
	if (f == null) return null;
	return f.getType();
    }

    /** Find a method based on name and types.
     * @param name The name of the method.  Constructors are named "<init>".
     * @param types A list of Type values representing the types of the 
     *              arguments.
     * @return A method with this name and types.  Null if no method was found.
     */
    public MethodDefiner findMethod(String name, List args, ClassEnv loader) {
        if (args == null) args = new ArrayList(0);
        MethodDeclaration md = getMethod(name, args, class_body.members, loader);
        if (md != null || name.equals("<init>")) return md;
        String supername = getSuperName();
        if (supername == null) return null;
        try {
            return loader.getClass(supername).findMethod(name, args, loader);
        } catch (Error e) {
            /* We get an exception once we hit Object, as that is 
             * a wrapped object and thus won't give use an AST node. 
             */
            return null;
        }
    }

    /** Find the return type of a method based on name and types.
     * @param name The name of the method.  Constructors are named "<init>".
     * @param types A list of Type values representing the types of the 
     *              arguments.
     * @return The type of the matching method, or null if this class
     * doesn't define such a method.
     */
    public Type findMethodType(String name, List args, ClassEnv loader) {
        if (args == null) args = new ArrayList(0);
        MethodDeclaration md = getMethod(name, args, class_body.members, loader);
        if (md != null) return md.type;
	if (name.equals("<init>")) throw new Error("Can't find method "+this+"."+name);
        String supername = getSuperName();
        if (supername == null) return null;
        try {
            return loader.getClass(supername).findMethodType(name, args, loader);
        } catch (Error e) {
            throw new Error("Can't find method "+this+"."+name);
        }
    }

    /** Returns an identifier for the name of the superclass. */
    public String getSuperName() {
        if (super_name == null) return "java.lang.Object";
        return super_name.toString();
    }

    /** Returns a list of names of interfaces */
    public List getInterfaces() {
        if (interfaces == null) return new ArrayList(0);
        List l = new ArrayList(interfaces.size());
        for (Iterator i = interfaces.iterator(); i.hasNext(); ) {
            Identifier id = (Identifier)i.next();
            l.add(id.toString());
        }
        return l;
    }

    /** Returns a string representation of this class declaration */
    public String toString() {
        return name.toString();
    }

    public boolean equals(Object o) {
	if (o instanceof ClassDeclaration) {
	    ClassDeclaration ocd = (ClassDeclaration)o;
	    return name.equals(ocd.name);
	}
	return false;
    }

    public boolean isInterface() { return false; } 

    /**
     * Name of superclass
     */
    protected Identifier super_name;
    /**
     * List of implemented interfaces
     */
    protected List interfaces;
    /**
     * Body of the class
     */
    protected ClassBody class_body;
}


