/* Copyright (c) 2002, University of Illinois.  All rights reserved.
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/*
 * ClassWrapper.java
 *
 */

package uiuc.Jumbo.Ast;

import uiuc.Jumbo.Util.*;

import java.lang.reflect.*;
import java.util.*;

/**
 * A ClassDefiner wrapper around the Class object (for sourceless classes)
 * @version      1.0, 22 Nov 2002
 * @author       Lars Clausen
 *
*/

public class ClassWrapper implements ClassDefiner {
    public ClassWrapper(Class c) {
        this.thisclass = c;
    }

    /** Find a representation of this field, if possible.
     */
    public FieldDefiner findField(String name, ClassEnv loader) {
        try {
            Field f = thisclass.getDeclaredField(name);
            return new FieldWrapper(f);
        } catch (NoSuchFieldException e) {
            Field[] fs = thisclass.getDeclaredFields();
            for (int i = 0; i < fs.length; i++) {
                if (name.equals(fs[i].getName()))
                    return new FieldWrapper(fs[i]);
            }

            Class sc = thisclass.getSuperclass();
            if (sc != null) {
                try {
                    return new ClassWrapper(sc).findField(name, loader);
                } catch (Error err) {
                    throw new Error("Can't find type for "+this+"."+name);
                }
            }
            throw new Error("Can't find type for "+this+"."+name);
        }
    }

    /** Return the type of the named field in this class.
     */
    public Type findFieldType(String name, ClassEnv loader) {
	FieldDefiner f = findField(name, loader);
	return f.getType();
    }

    /** Check if the method could be called with these arguments.
     * @param name
     * @param args
     * @param m A method 
     * @return true if the parameters of `m' are the same or superclasses
     * of the corresponding types in `args', and m's name is `name', 
     * false otherwise.
     *
     * This annoyingly extraneuous function is necessary because Sun
     * doesn't seem to think that doing method lookup the way compilers
     * do it is worth implementing.  Not even three years after the
     * original request.
     */
    private boolean methodConforms(String name, Class[] args, Method m) {
        if (!name.equals(m.getName())) return false;
        Class[] params = m.getParameterTypes();
        if (args.length != params.length) return false;
        if (false)
            System.out.println("Comparing "+m+" with "+name+args);
        for (int i = 0; i < args.length; i++) {
            if (params[i].isPrimitive()) {
                if (params[i] == args[i]) continue;
                if (params[i] == Integer.TYPE) {
                    if (args[i] == Byte.TYPE ||
                        args[i] == Character.TYPE ||
                        args[i] == Short.TYPE)
                        continue;                       
                }
                if (params[i] == Short.TYPE) {
                    if (args[i] == Byte.TYPE)
                        continue;                       
                }
            } else {
                if (args[i] == null) continue;
                if (params[i].isAssignableFrom(args[i]))
                    continue;
            }
            /* Here should be some fiddling with integer promotion.  Feck! */
            return false;            
        }
        return true;
    }

    private Method findConformingMethod(String name, Class[] args) {
        try {
            Method m = thisclass.getDeclaredMethod(name, args);
            return m;
        } catch (NoSuchMethodException e) {
            Method[] ms = thisclass.getDeclaredMethods();
            for (int i = 0; i < ms.length; i++) {
                if (methodConforms(name, args, ms[i]))
                    return ms[i];
            }
        }
        return null;
    }

    private Method findInterfaceMethod(String name, Class[] args) {
        Method m = findConformingMethod(name, args);
        if (m != null || name.equals("<init>")) return m;
        Class[] is = thisclass.getInterfaces();
        for (int i = 0; i < is.length; i++) {
            m = new ClassWrapper(is[i])
                .findMethodInternal(name, args);
            if (m != null) return m;
        }
        return null;
    }

    private Method findMethodInternal(String name, Class[] args) {
        // Interfaces may have methods defined in superinterfaces
        if (thisclass.isInterface())
            return findInterfaceMethod(name, args);
        // But an actual class must have the method defined in a
        // non-interface superclass to correctly implement an interface

        Method m = findConformingMethod(name, args);
        if (m != null || name.equals("<init>")) return m;
        Class sc = thisclass.getSuperclass();
        if (sc != null) 
            return new ClassWrapper(sc).
                findMethodInternal(name, args);
        return null;
    }

    private Class[] argsToClassList(List args, ClassEnv loader) {
	Class[] argarray = new Class[args.size()];
        int index = 0;
        for (Iterator i = args.iterator(); i.hasNext(); ) {
            Type t = (Type)i.next();
            if (!t.isNull()) {
                argarray[index] = t.toClass(loader);
	    }
	    index++;
        }
	return argarray;
    }

    /** Find an AST node representing this method, if possible.
     */
    public MethodDefiner findMethod(String name, List args, ClassEnv loader) {
        try {
            Method m = findMethodInternal(name, argsToClassList(args, loader));
            if (m != null) return new MethodWrapper(m);
        } catch (Error e) {
            throw new Error("Error finding type for "+this+"."+name+"("+args+"): "+e);
        }
        throw new Error("Can't find "+this+"."+name+"("+args+")");

    }
    /** Return the return type of the indicated method in this class.
     */
    public Type findMethodType(String name, List args, ClassEnv loader) {
        try {
            Method m = findMethodInternal(name, argsToClassList(args, loader));
            if (m != null) return Type.fromClass(m.getReturnType());
        } catch (Error e) {
	    e.printStackTrace();
            throw new Error("Error finding type for "+this+"."+name+"("+args+"): "+e);
        }
        throw new Error("Can't find type for "+this+"."+name+"("+args+")");
    }
    private static String nonInnerName(String name) {
	return name.replace('$', '.');
    }
    /** The full name of the superclass, with dollarsigns removed. */
    private String noninnersupername;
    /** Return the FQCN of the super class, or NULL if this class is
     * java.lang.Object.
     */
    public String getSuperName() {
	if (noninnersupername == null) {
	    Class sc = thisclass.getSuperclass();
	    if (sc != null) {
		noninnersupername = sc.getName();
		// Don't want these spurious $'s in there
		noninnersupername = nonInnerName(noninnersupername);
	    }
	}
        return noninnersupername;

    }
    /** The full name of this class, with dollarsigns removed.
     * Dollarsigns are introduced for inner classes. */
    private String noninnername;
    /** Return the name of this class.
     */
    public String getName() {
	if (noninnername == null) {
	    noninnername = thisclass.getName();
            // Don't want these spurious $'s in there
            noninnername = nonInnerName(noninnername);
	}
        return noninnername;
    }
    /** Returns the fully qualified class name of this class.
     * For ClassWrapper, the name is always fully qualified.
     */
    public String getFQCN() {
        return getName();
    }

    private List interfacenames;
    /** Returns a list of names of interfaces.  This list may be
     * cached, so don't mess with it. */
    public List getInterfaces() {
	if (interfacenames == null) {
	    Class[] interfaces = thisclass.getInterfaces();
	    List l = new ArrayList(interfaces.length);
	    for (int i = 0; i < interfaces.length; i++) {
		l.add(nonInnerName(interfaces[i].getName()));
	    }
	    interfacenames = l;
	}
        return interfacenames;
    }

    /** Return true if this is a final class.
     */
    public boolean isFinal() {
        int mods = thisclass.getModifiers();
        return Modifier.isFinal(mods);
    }

    public boolean isInterface() { return thisclass.isInterface(); }

    /** Get the contents of this field, if constant.
     */
    public Literal getFieldContents(String fieldname) {
        try {
            Field f = thisclass.getField(fieldname);
            int mod = f.getModifiers();
            if ((mod & Modifier.STATIC) == 0 ||
                (mod & Modifier.FINAL) == 0)
                return null;
            Object o = f.get(null);
            if (o != null) return Literal.make(o);
            return null;
        } catch (NoSuchFieldException e) {
            return null;
        } catch (IllegalAccessException e) {
            return null;
        }
    }

    public String toString() {
        return "wrapped "+thisclass;
    }

    Class thisclass;
}

/* End of class ClassWrapper*/
