/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uiuc.Jumbo.Ast;

import java.util.*;
import java.util.Iterator;
import uiuc.Jumbo.Util.*;

/**
 * Represents identifiers of function names, variables, etc.
 * When printed as a constructor, includes quotes around the name.
 * Includes identifiers such as 'a.b.c'.
 *
 * @author      Ava Arachne Jarvis
 * @version     %I%, %G%
 * @since       JDK 1.3
 */

public class Identifier extends Node  {
    /**
     * @param sname  identifier string object
     */
    public Identifier( IdentifierString sname) {
        this();
	if (sname == null) Thread.dumpStack();
        appendField(sname);
    }

    /**
     * @param iname  identifier string
     */
    public Identifier( String iname ) {
        this();
	if (iname == null) Thread.dumpStack();
        appendField(new IdentifierString(iname));
    }

    /**
     * @param iname  identifier
     */
    public Identifier( Identifier i ) {
        this(i.identifier_name);
	if (i == null) Thread.dumpStack();
    }

    /**
     * @param names List of String or IdentifierString elements
     */
    public Identifier( List names ) {
        Iterator i = names.iterator();
        {
            Object o = i.next();
            if (o instanceof String) setIdentifier( new IdentifierString((String)o));
            if (o instanceof IdentifierString) setIdentifier( (IdentifierString)o);
        }
        while (i.hasNext()) {
            Object o = i.next();
            if (o instanceof String) appendField((String)o);
            if (o instanceof IdentifierString) appendField( (IdentifierString)o);           
        }
    }

    /**
     * Constructor to allow other constructors to do their job.
     * We can't create new objects before the super constructor is
     * called, but the super constructor requires an IdentifierString
     */
    protected Identifier() {
        identifier_name = new LinkedList();
    }

    /**
     * Set the identifier name after construction
     * This should <em>only</em> be called from within constructors
     * after calling Identifier().
     */
    protected void setIdentifier( IdentifierString iname ) { 
        identifier_name = new LinkedList();
        identifier_name.add( iname );
    }

    /**
     * Adds another field to the end of the list.
     *
     * @param fname  name of the field to be added
     */
    public void appendField( IdentifierString fname ) {
        identifier_name.add( fname );
    }

    /**
     * Adds another field to the end of the list.
     *
     * @param fname  name of the field to be added
     */
    public void appendField( String fname ) {
        identifier_name.add( new IdentifierString( fname ) );
    }

    /**
     * Adds another field to the front of the list.
     *
     * @param fname  name of the field to be added
     */
    public void prependField( IdentifierString fname ) {
        identifier_name.add( 0, fname );
    }

    /**
     * Adds another field to the front of the list.
     *
     * @param fname  name of the field to be added
     */
    public void prependField( String fname ) {
        identifier_name.add( 0, new IdentifierString( fname ) );
    }

    /**
     * Prepends another Identifier to this one.
     *
     * @param ident  Identifier to be added
     */
    public void prependIdentifier( Identifier ident ) {
        identifier_name.addAll( 0, ident.identifier_name );
    }

    /**
     * Appends another Identifier to this one.
     *
     * @param ident  Identifier to be added
     */
    public void appendIdentifier( Identifier ident ) {
        identifier_name.addAll( ident.identifier_name );
    }

    /**
     * Removes the last field name and returns it as another 
     * Identifier object.
     *
     * @return last field name of the identifier.
     */
    public Identifier splitOffLastField() {
        IdentifierString last_field = (IdentifierString)identifier_name.remove(identifier_name.size()-1);

        return new Identifier(last_field);
    }

    /** Returns the nth field of the identifier.
     * @param n The field number to return, starting at 0
     * @return The indexed field.
     */
    public IdentifierString getField(int n) {
        return (IdentifierString)identifier_name.get(n);
    }

    /** Returns the number of fields in this identifier.
     * @return The number of fields in a dotted identifier.  Always >= 1.
     */
    public int size() {
        return identifier_name.size();
    }

    /**
     * Indicates whether this is a simple identifier.
     *
     * @return true if this is a simple identifier, false otherwise.
     */
    public boolean isSimple() {
        return (identifier_name.size() == 1);
    }

    /**
     * Indicates whether this is an identifier with no unknown parts.
     *
     * @return true if this identifier contains only known parts.
     */
    public boolean isFullyKnown() {
        for (Iterator i = identifier_name.iterator(); i.hasNext(); ) {
            if (i.next() instanceof AntiquoteIdentifierString)
                return false;
        }
        return true;
    }

    /**
     * Returns the identifier as either list of fields.
     */
    public String toString() {
        String result = "";

        Iterator iter = identifier_name.iterator();

        result += iter.next();

        while( iter.hasNext() ) {
            result += "."+iter.next();
        }

        return result;
    }

    public String getString() {
        String res = "";
        for (Iterator i = identifier_name.iterator(); i.hasNext(); ) {
            res += ((IdentifierString)i.next()).getString();
            if (i.hasNext()) res += ".";
        }
        return res;
    }

    public IR toIR() {
        if (isSimple()) {
            return ((IdentifierString)identifier_name.get(0)).toIR();
        } else {
            MonoList list = new ArrayMonoList();
            for (Iterator i = identifier_name.iterator(); i.hasNext(); ) {
                list.add(((IdentifierString)i.next()).toIR());
            }
            return IR.make(list);
        }
    }

    public IR toIRString() {
        if (isSimple()) {
            return ((IdentifierString)identifier_name.get(0)).toIR();
        } else {
            MonoList l = new ArrayMonoList();
            if (identifier_name != null)
                for (Iterator i = identifier_name.iterator();
                     i.hasNext(); ) {
                    l.add(((IdentifierString)i.next()).toIR());
                }
            return IR.make(l);
        }
    }

    public void toSource() {
	if (specialOutputFormatter != null) 
	    specialOutputFormatter.output(this);
	else
	    print(getString());
    }

    public boolean equals(Object o) {
        if (!(o instanceof Identifier)) return false;
        return ((Identifier)o).identifier_name.equals(identifier_name);
    }

    public int hashCode() {
        return identifier_name.hashCode();
    }

    /**
     * Identifier string, broken up into fields (formerly separated by '.').
     * Order goes from leftmost to rightmost: a.b.c == (a, b, c).
     */
    protected List identifier_name;
}
