/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uiuc.Jumbo.Ast;

import java.util.*;
import java.util.Iterator;

import uiuc.Jumbo.Util.*;

/**
 * Represents an interface declaration.
 *
 * @author      Ava Arachne Jarvis
 * @version     %I%, %G%
 * @since       JDK 1.3
 */

public class InterfaceDeclaration extends TypeDeclaration
    implements ClassDefiner {
    /**
     * @param mods    Optional list of modifier flags
     * @param iname   Interface name
     * @param ifaces  Optional list of interface names extended
     * @param ibody   Interface body (another class)
     */
    public InterfaceDeclaration( ModifierList mods, 
            Identifier iname, 
            LinkedList ifaces, 
            InterfaceBody ibody) 
    { 
        super(mods, iname);

        interfaces = ifaces;
        setInterfaceBody(ibody);
    }

    protected boolean replaceUC(Node old_child, Node new_child) {
        if (super.replaceUC(old_child, new_child)) {
            return true;
        }
        else if (old_child == interface_body) {
            setInterfaceBody((InterfaceBody)new_child);
            return true;
        }
        else {
            return replaceInList(interfaces, old_child, new_child);
        }
    }

    private void setInterfaceBody(InterfaceBody ibody) {
        interface_body = ibody;
        setParent(interface_body);
    }

    public void toSource() {
        if (modifiers != null)
            modifiers.toSource();
        print("interface ");
        name.toSource();
        if (interfaces != null && interfaces.size() != 0) {
            print( " extends ");
            printSourceList(interfaces, ", ");
        }
        print(" ");
        ocurly();
        linebreak();
        interface_body.toSource();
        ccurly();

    }

    public IR toIR() {
        return IR.make("defineInterface", new ArrayMonoList()
                       .add(modifiers!=null?modifiers.toIR():IR.make(0))
                       .add(name.toIR())
                       .add(toIRStringList(interfaces))
                       .add(interface_body.toIR()));
    }

    public boolean needsSemicolon() { return false; }

    /** Returns the variable defining the given field, if any.
     */
    public FieldDefiner findField(String field, ClassEnv loader) {
        FieldDefiner f = getField(field, interface_body.members);
        if (f != null) return f;
        List interfaces = getInterfaces();
        for (Iterator i = interfaces.iterator(); i.hasNext(); ) {
            String iname = (String)i.next();
            f = loader.getClass(iname).findField(field, loader);
            if (f != null) return f;
        }
        return null;
    }

    /** Returns the type of the given field, if any.
     */
    public Type findFieldType(String field, ClassEnv loader) {
	FieldDefiner f = findField(field, loader);
	if (f == null) return null;
	return f.getType();
    }

    /** Find a method based on name and types.
     * @param name The name of the method.  Constructors are named "<init>".
     * @param types A list of Type values representing the types of the 
     *              arguments.
     * @return A method with this name and types.  Null if no method was found.
     */
    public MethodDefiner findMethod(String name, List args, ClassEnv loader) {
        if (args == null) args = new ArrayList(0);
        MethodDefiner md = 
	    getMethod(name, args, interface_body.members, loader);
        if (md != null || name.equals("<init>")) return md;
        List interfaces = getInterfaces();
        for (Iterator i = interfaces.iterator(); i.hasNext(); ) {
            String iname = (String)i.next();
            md = loader.getClass(iname).findMethod(name, args, loader);
            if (md != null) return md;
        }
        return null;
    }

    /** Find the return type of a method based on name and types.
     * @param name The name of the method.  Constructors are named "<init>".
     * @param types A list of Type values representing the types of the 
     *              arguments.
     * @return The type of the matching method, or null if this class
     * doesn't define such a method.
     */
    public Type findMethodType(String name, List args, ClassEnv loader) {
        MethodDefiner md = findMethod(name, args, loader);
	if (md != null) return md.getType();
	return null;
    }

    /** Return the FQCN of the super class.
     */
    public String getSuperName() {
        throw new Error("Can't find a single super name of an interface.  Please think.");
    }

    /** Returns a list of names of interfaces */
    public List getInterfaces() {
        if (interfaces == null) return new ArrayList(0);
        List l = new ArrayList(interfaces.size());
        for (Iterator i = interfaces.iterator(); i.hasNext(); ) {
            Identifier id = (Identifier)i.next();
            l.add(id.toString());
        }
        return l;
    }

    public boolean isInterface() { return true; }

    public String toString() {
        return "interface "+name+" { <"+interface_body.members.size()+" members> }";
    }

    /**
     * List of extended interfaces
     */
    protected LinkedList interfaces;
    /**
     * Body of the interface
     */
    protected InterfaceBody interface_body;
}
