/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uiuc.Jumbo.Ast;

import java.util.*;
import java.util.Iterator;
import uiuc.Jumbo.Util.*;

/**
 * Represents a method invocation. 
 *
 * @author      Ava Arachne Jarvis
 * @version     %I%, %G%
 * @since       JDK 1.3
 */

public class MethodInvocation extends Expression {
    /**
     * @param t     target of the method invocation
     * @param n     name of method being invoked
     * @param args  optional argument list
     */
    public MethodInvocation( Expression t, Identifier n, ArgumentList args ) { 
        this("invoke", t, n, args);
    }

    /**
     * @param c     name of the constructor.
     * @param t     target of the method invocation
     * @param n     name of method being invoked
     * @param args  optional argument list
     */
    protected MethodInvocation( String c, Expression t, 
            Identifier n, ArgumentList args ) {
        super(c);
        target = t;
        setParent(target);
        name = n;
        setParent(name);
        arguments = args;
        setParent(arguments);
    }

    /**
     * @param v     variable that is the target of the method invocation
     * @param n     name of method being invoked
     * @param args  optional argument list
     */
    public MethodInvocation( String s, Identifier n, ArgumentList args ) { 
        this("invoke", new VarExpression(new Identifier(s)), n, args);
    }

    protected boolean replaceUC(Node old_child, Node new_child) {
        if (old_child == target) {
            setTarget(new_child);
            return true;
        }
        else if (old_child == name) {
            setName((Identifier)new_child);
            return true;
        }
        else {
            return replace(old_child, new_child);
        }
    }

    protected boolean replaceUC(List old_child, List new_child) {
        if (old_child == arguments) {
	    if(new_child instanceof ArgumentList)
		setArguments((ArgumentList)new_child);
	    else
		setArguments(new ArgumentList(new_child));
            return true;
        }
        return false;
    }

    private void setTarget(Node t) {
        target = (Expression)t;
        setParent(target);
    }

    protected void setName(Identifier n) {
        name = n;
        setParent(name);
    }

    private void setArguments(ArgumentList l) {
        arguments = l;
        setParent(arguments);
    }

    public IR toIR() {
        IRCode ir = IR.make(constructor, new ArrayMonoList());
        if (target != null) {
            ir.addArg(target.toIR());
        }
        else {
            ir.addArg(IR.make(default_string));
        }
        ir.addArg(name.toIR());
	if(arguments == null)
	    ir.addArg(IR.make(new ArrayMonoList()));
	else
	    ir.addArg(arguments.toIR());
        return ir;
    }

    /**
     * Prints out this method invocation as source.
     */
    public void toSource() {
        if (target != null) {
            target.toSource();
            print(".");
        }
        name.toSource();
	if(arguments != null)
	    printSourceArgumentList(arguments.getList());
    }

    /** Returns a string representation of this method invocation. */
    public String toString() {
        String s = "";
        if (target != null)
            s = target.toString()+".";
        s += name.toString()+"(";
        if (arguments != null)
            for (Iterator i = arguments.getList().iterator(); i.hasNext(); ) {
                s += i.next().toString();
                if (i.hasNext()) s += ", ";
            }
        s += ")";
        return s;
    }

    public List getArgTypes(ClassEnv loader) {
        List argtypes = new ArrayList();
        if (arguments != null)
            for (Iterator i = arguments.getList().iterator(); i.hasNext(); ) {
                Expression e = ((Expression)i.next());
                Type t = e.getType(loader);
                argtypes.add(t);
            }
	return argtypes;
    }

    public Type getType(ClassEnv loader) {
        ClassDefiner cd;
	if (target == null ||
	    (target instanceof ThisExpression &&
	     ((ThisExpression)target).name == null))
	    cd = getThis();
	else
            if (target instanceof ThisExpression) {
		cd = ((ThisExpression)target).getClass(loader);
	    } else {
                Type targettype = target.getType(loader);
                if (targettype == null) throw new Error("No type for target "+targettype);
		if (targettype.isNull()) throw new Error("Target of "+this+" is of type 'null'");
                cd = loader.getClass(loader.getFQCN(targettype).getName());
            }
        if (cd == null) {
            throw new Error("Can't find class for target "+target);
        }
	List argtypes = getArgTypes(loader);
        Type t;
        do {
            t = cd.findMethodType(name.getString(), argtypes, loader);
            if (t != null) {
                return loader.getFQCN(t);
            }
            List supers = loader.getSupers(new Type(cd.getFQCN()));
            if (supers != null && supers.size() > 0) {
                Type supert = (Type)supers.get(0);
                cd = loader.getClass(supert.getName());
            } else break;
        } while (cd != null);
        throw new Error("No type found for "+cd+"."+name+argtypes);
    }

    public boolean isConstructorInvocation() {
	return name.equals("<init>");
    }
    /**
     * Target of the method invocation.
     */
    protected Expression target;
    /**
     * Name of the method being invoked.
     */
    protected Identifier name;
    /**
     * Argument list.
     */
    protected ArgumentList arguments;
    protected String default_string = "";
}
