/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uiuc.Jumbo.Ast;
import uiuc.Jumbo.Util.*;

/**
 * A qualified name that ends up in an expression.
 *
 * @author      Ava Arachne Jarvis
 * @version     %I%, %G%
 * @since       JDK 1.3
 */

public class QualifiedNameExpression extends Expression  {
    /**
     * @param n  the name of the expression
     */
    public QualifiedNameExpression( Identifier n ) { 
        constructor = "QualifiedName";
        setName(n);
    }

    public IR toIR() {
        return IR.make("get" + constructor, new ArrayMonoList()
                        .add(name.toIRString()));
    }

    public boolean replaceUC(Node old_child, Node new_child) {
	if (name == old_child) {
	    setName((Identifier)new_child);
	    return true;
	}
	return false;
    }

    private void setName(Identifier i) {
        name = i;
        setParent(i);
    }

    public void toSource() {
        name.toSource();
    }

    public String toString() {
        return name.toString();
    }

    /**
     * Returns a new type object based on this name.
     */
    public Type toType() {
        return new Type( name );
    }

    /** Get the variable declaration for this QualifiedNameExpression
     * without depending on outside variable lookup.
     * This is *not* the same as getVariableDeclaration.  This will
     * actually look through outer classes as well, and check for fields.
     * @param from The point in the AST to find variables in.  In the
     * recursive calls, this will be the class definition node.
     * @param loader A ClassEnv instance.
     * @return The node that declares the variable in the first part
     * of the QNE, or null if it's not a variable.
     */
    public Variable getVarDecl(Node from, ClassEnv loader) {
        String first = name.getField(0).getString();
        Variable v = from.getVariableDeclaration(first);
        if (v == null) {
            ClassDefiner thisclass = from.getEnclosingClass();
            // This can happen with imports statements.
            if (thisclass == null) return null;
            Type t;
            try {
                t = thisclass.findFieldType(first, loader);
            } catch (Error e) {
                t = null;
            }
            if (t != null) {
                return null;
            }
            ClassDefiner enclosing = ((Node)thisclass).getEnclosingClass();
            if (enclosing == null) {
                return null;
            }
            return getVarDecl((Node)thisclass, loader);
        } 
        return v;
    }

    /** Returns the type of the beginning variable or field.
     */
    public Type getFieldOrVarType(Node from,
                                  Identifier name, ClassEnv loader) {
        ClassDefiner thisclass = from.getEnclosingClass();
        // This can happen with imports statements.
        if (thisclass == null) return null;
        String first = name.getField(0).getString();
        Variable v = from.getVariableDeclaration(first);
        if (v == null) {
            Type t;
            try {
                t = thisclass.findFieldType(first, loader);
            } catch (Error e) {
                t = null;
            }
            if (t != null) return t;
            return getFieldOrVarType((Node)thisclass, name, loader);
        } 
        return v.type;
    }

    public class QNEMarker {
	/** The first part of the QNE which can be a field.
	 * If the QNE doesn't resolve to any usable thing, an
	 * error will be thrown.
	 * This may be = name.size(), if there are no parts of the QNE
	 * left to be fields.
	 */
	int field;
	/** The type of the starting part of the QNE, if
	 * firstfield !+ -1.
	 */
	Type type;
	/** True if the first part of the QNE is a variable.
	 */
	boolean isVar;
    }

    /** Returns the first variable or field with the given name, or null.
     * This function scans through outer classes too.  
     */
    public QNEMarker getFirstField(ClassEnv loader) {
	QNEMarker marker = new QNEMarker();
        int firstfield = -1;
        Type t = getFieldOrVarType(this, name, loader);
        if (t != null) {
	    if (getVarDecl(this, loader) != null) {
		firstfield = 1;
		marker.isVar = true;
	    } else {
		firstfield = 0;
	    }
        } else {
            for (int i = name.size()-1; i >= 0; i--) {
		String classname = name.getField(0).getString();
		for (int j = 1; j <= i; j++) {
		    classname += "."+name.getField(j).getString();
		}
                ClassDefiner cd = loader.getClass(classname);
                if (cd != null) {
                    t = new Type(cd.getFQCN());
                    firstfield = i+1;
                    break;
                }
            }
	    if (t == null) {
		if (getEnclosingClass() == null)
		    System.out.println("Parents: "+getParents());
		throw new Error("Can't resolve QNE "+name+" in "+getEnclosingClass());
	    }
        }
	marker.field = firstfield;
	marker.type = t;
	return marker;
    }

    public Type getType(ClassEnv loader) {
	QNEMarker first = getFirstField(loader);
	Type t = first.type;
        if (first.field < name.size()) {
	    int i = first.field;
	    if (i == 0) // First part is field, type is looked up already
		i++;
            for (; i < name.size(); i++) {
                if (t.isArray() && name.getField(i).getString().equals("length"))
                    return new Type("int");
                ClassDefiner cd = loader.getClass(t.getName());
                String s = name.getField(i).getString();
                try {
                    t = cd.findFieldType(s, loader);
                } catch (Error e) {
                    System.out.println(e);
                }
                if (t == null) {
                    throw new Error("Null type for "+this+" (first field "+first.field+")");
                }
            }
        }
        return t;
    }

    /**
     * The qualified name.
     */
    protected Identifier name;
}
