/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/**
 * Represents a list of statements.
 *
 * @author      Ava Arachne Jarvis
 * @version     %I%, %G%
 * @since       JDK 1.3
 */

package uiuc.Jumbo.Ast;

import java.util.*;
import java.util.Iterator;

import uiuc.Jumbo.Util.*;
import uiuc.Jumbo.Compiler.*;

public class StatementList extends Statement {
    /**
     * Creates an empty StatementList.
     */
    public StatementList() { 
        statements = new LinkedList();
    }

    protected boolean replaceUC(Node old_child, Node new_child) {
        return replaceInList(statements, old_child, new_child);
    }

    /**
     * Creates a StatementList with the given statements
     */
    public StatementList(List l) {
        setStatements(l);
    }

    protected boolean replaceUC(List old_child, List new_child) {
        if (old_child == statements) {
            setStatements(new_child);
            return true;
        }
        return false;
    }

    /** Sets the statements */
    public void setStatements(List l) {
        statements = l;
        setParents(l);
    }

    /**
     * Prepends a Statement to the StatementList.
     */
    public void prependStatement( Statement s ) {
        statements.add(0, s );
        setParent(s);
    }

    /**
     * Appends a Statement to the StatementList.
     */
    public void appendStatement( Statement s ) {
        statements.add( s );
        setParent(s);
    }

    public void appendStatements( StatementList s) {
        for (Iterator i = s.statements.iterator(); i.hasNext(); ) {
            Statement st = (Statement)i.next();
            appendStatement( st );
        }
    }

    /**
     * Adds a statement to the end of the StatementList.
     */
    protected void add( Statement s) {
        appendStatement(s);
    }

    /**
     * Adds an expression to the end of the StatementList.
     */
    protected void add( Expression e) {
        appendStatement(new ExpressionStatement(e));
    }

     public IR toIR() {
        IRCode ir = IR.make("statements", new ArrayMonoList());
        if (use_singles && statements.size() == 1) {
            ir.addArg(((Node)statements.get(0)).toIR());
        } else {
            ir.addArg(toIRList(statements));
        }
        return ir;
    }

    /** Prints out the corresponding source code
    */
    public void toSource() {
        Iterator i = statements.iterator();
        while (i.hasNext()) {
            Statement s = (Statement)i.next();
            s.toSource();
            if (s.needsSemicolon())
                print(";");
            if (!(s instanceof EmptyStatement))
		linebreak();
        }
    }

    public void toCommaSource() {
        Iterator i = statements.iterator();
        while (i.hasNext()) {
            Statement s = (Statement)i.next();
            s.toSource();
            if (i.hasNext())
                print(", ");
        }
    }

    /** Return the imports in effect for this node.
     */
    public List getImports() {
        if (parent == null) {
            List imports = new ArrayList();
            List jl = new ArrayList();
            jl.add("java");
            jl.add("lang");
            imports.add(new ImportDeclaration(new Identifier(jl), true));
            for (Iterator i = statements.iterator(); i.hasNext(); ) {
                Node n = (Node)i.next();
                if (n instanceof ImportDeclaration) {
                    imports.add(n);
                } else if (n instanceof PackageDeclaration) {
                    imports.add(n);
                }
            }
            return imports;
        } else {
            return parent.getImports();
        }
    }

    public boolean needsSemicolon() { return false; }

    public Variable getVariableDeclaration(String name, Node from) {
        Variable v = null;
        for (Iterator i = statements.iterator(); i.hasNext(); ) {
            Statement s = (Statement)i.next();
            if (s == from) {
                break;
            }
            if (v != null) continue; // Keep checking for from
            if ((s instanceof StatementList) &&
                (!(s instanceof ScopedStatementList))){
                v = ((StatementList)s).getContainedVariableDeclaration(name);
            }
            if (s instanceof LocalVariableDeclaration)
                v = ((LocalVariableDeclaration)s).getContainedVariableDeclaration(name);
        }
        if (v != null) return v;
        // from == null indicates that we came from an enclosing StatementList.
        // In that case, we don't want to check outside.
        if (from != null)
            return super.getVariableDeclaration(name);
        else
            return null;
    }
    
    public Variable getContainedVariableDeclaration(String name) {
        Variable v = null;
        for (Iterator i = statements.iterator(); i.hasNext(); ) {
            Statement s = (Statement)i.next();
            if ((s instanceof StatementList) &&
                (!(s instanceof ScopedStatementList))){
                v = ((StatementList)s).getContainedVariableDeclaration(name);
            }
            if (s instanceof LocalVariableDeclaration)
                v = ((LocalVariableDeclaration)s).getContainedVariableDeclaration(name);
            if (v != null) return v;
        }
        return null;
    }

    /**
     * List of statements.
     */
    protected List statements;

    public int size() { return statements.size(); }

    public String toString() {
        String s = "[";
        for (Iterator i = statements.iterator(); i.hasNext(); ) {
            s += i.next();
            if (i.hasNext());
            s += ", ";
        }
        s += "]";
        return s;
    }
}
