/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uiuc.Jumbo.Ast;

import uiuc.Jumbo.Util.*;

/**
 * Represents a trinary conditional expression.
 *
 * @author      Ava Arachne Jarvis
 * @version     %I%, %G%
 * @since       JDK 1.3
 */

public class TrinaryExpression extends Expression  {
    /**
     * @param te         Test expression
     * @param t          "True" branch expression
     * @param f          "False" branch expression
     */
    public TrinaryExpression( Expression te, Expression t, Expression f ) { 
	setTest(te);
	setTrueExpression(t);
	setFalseExpression(f);
    }

    private void setTest(Expression te) {
	test = te;
	setParent(test);
    }

    private void setTrueExpression(Expression t) {
	true_expression = t;
	setParent(true_expression);
    }

    private void setFalseExpression(Expression f) {
	false_expression = f;
	setParent(false_expression);
    }

    protected boolean replaceUC(Node old_child, Node new_child) {
	if (old_child == test) {
	    setTest((Expression)new_child);
	    return true;
	}
	else if (old_child == true_expression) {
	    setTrueExpression((Expression)new_child);
	    return true;
	}
	else if (old_child == false_expression) {
	    setFalseExpression((Expression)new_child);
	    return true;
	}
	else {
	    return false;
	}
    }

    public IR toIR() {
        return IR.make("trinaryConditional", new ArrayMonoList()
                       .add(test.toIR())
                       .add(true_expression.toIR())
                       .add(false_expression.toIR()));
    }

    public void toSource() {
        print("(");
        test.toSource();
        print("?");
        true_expression.toSource();
        print(":");
        false_expression.toSource();
        print(")");
    }

    public Type getType(ClassEnv loader) {
        Type t1 = true_expression.getType(loader);
        Type t2 = false_expression.getType(loader);
        if (t1.isPrimitive() && t2.isPrimitive()) {
            if (t1.equals(t2)) return t1;
            if (t1.getKind() == Type.BOOLEAN ||
                t2.getKind() == Type.BOOLEAN) {
                throw new Error("Can't unify "+t1+" and "+t2+" in trinary");
            }
            /* There's really a special case for downcasting constant ints
             * to byte/short/char, but to hell with that.
             */
            if (t1.getKind() == Type.SHORT && t2.getKind() == Type.BYTE) 
                return t1;
            if (t2.getKind() == Type.SHORT && t1.getKind() == Type.BYTE) 
                return t2;
            if (t1.getKind() == Type.DOUBLE) return t1;
            if (t2.getKind() == Type.DOUBLE) return t2;
            if (t1.getKind() == Type.FLOAT) return t1;
            if (t2.getKind() == Type.FLOAT) return t2;
            if (t1.getKind() == Type.LONG) return t1;
            if (t2.getKind() == Type.LONG) return t2;
            if (t1.getKind() == Type.INT) return t1;
            if (t2.getKind() == Type.INT) return t2;
            throw new Error("Can't unify "+t1+" and "+t2+" in trinary");
        }
        if (t1.isNull() && t2.isReference()) {
            return t2;
        }
        if (t2.isNull() && t1.isReference()) {
            return t1;
        }
        if (t1.isReference() && t2.isReference()) {
            if (t1.conforms(t2, loader)) return t2;
            if (t2.conforms(t1, loader)) return t1;
        }
        throw new Error("Mismatched types in trinary "+t1+" and "+t2);
    }

    public String toString() {
        return "("+test+"?"+true_expression+":"+false_expression+")";
    }

    /**
     * Test expression.
     */
    protected Expression test;
    /**
     * "True" condition.
     */
    protected Expression true_expression;
    /**
     * "False" condition.
     */
    protected Expression false_expression;
}
