/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uiuc.Jumbo.Ast;

/**
 * Represents a type. 
 *
 * @author      Ava Arachne Jarvis
 * @version     %I%, %G%
 * @since       JDK 1.3
 */

import java.util.*;
import java.lang.reflect.*;

public class Type extends Node {
    /** A type that is not determined yet */
    public static final byte UNKNOWN    = -1;
    /** A type for NULL */
    public static final byte NULL    = 0;
    /*public static final byte ARRAY   = 1;*/
    public static final byte OBJECT  = 2;
    public static final byte VOID    = 3;
    public static final byte BOOLEAN = 4;
    public static final byte CHAR    = 5;
    public static final byte FLOAT   = 6;
    public static final byte DOUBLE  = 7;
    public static final byte BYTE    = 8;
    public static final byte SHORT   = 9;
    public static final byte INT     = 10;
    public static final byte LONG    = 11;

    private int kind;
    
    /** The name of this type, if it is an object type */
    private String name;

    /** The dimensions of this type (0 = not an array) */
    private int dimensions;

    /**
     * Creates a completely blank primitive type for 
     * filling in later.
     */
    public Type() {
        this( "", 0 );
    }

    /**
     * Creates a type with no dimensions.
     *
     * @param tname   name of the type
     */
    public Type( Object tname ) { 
        this( tname, 0 );
    }

    public Type( Object tname, int dims ) {
        setName(tname);
        dimensions += dims;
    }

    /** Create a type with specific kind, name and dimensions. 
     * Don't use setName here, as this is only used from copyType() and
     * getElementType() */
    protected Type( int kind, String tname, int dims ) {
        this.kind = kind;
        this.name = tname;
        this.dimensions += dims;
    }

    public Type( int kind, int dims ) {
        this.kind = kind;
        this.dimensions += dims;
        if (!isPrimitive() && !isVoid()) throw new Error("Lacking name for creating object type");
    }

    public Type copyType() {
        Type t = new Type(kind, name, dimensions);
        return t;
    }

    /** Return a type representing the same type as the Class given.
     * This is for Class objects given by reflection.
     */
    public static Type fromClass(Class c) {
	Class base = c;
	int arraydims = 0;
	while (base.isArray()) {
	    base = base.getComponentType();
	    arraydims++;
	}
	if (base.isPrimitive()) {
	    if (base == Boolean.TYPE) return new Type(BOOLEAN, arraydims);
	    if (base == Float.TYPE) return new Type(FLOAT, arraydims);
	    if (base == Double.TYPE) return new Type(DOUBLE, arraydims);
	    if (base == Character.TYPE) return new Type(CHAR, arraydims);
	    if (base == Byte.TYPE) return new Type(BYTE, arraydims);
	    if (base == Short.TYPE) return new Type(SHORT, arraydims);
	    if (base == Integer.TYPE) return new Type(INT, arraydims);
	    if (base == Long.TYPE) return new Type(LONG, arraydims);
	    if (base == Void.TYPE) return new Type(VOID, arraydims);
	    throw new Error("Unknown primitive type "+base);
	} else {
	    return new Type(OBJECT, base.getName(), arraydims);
	}
    }

    /** Return a type representing the given kind
     */
    public static Type fromKind(int k) {
	int arraydims = 0;
	if (k == BOOLEAN) return new Type(BOOLEAN, arraydims);
	if (k == FLOAT) return new Type(FLOAT, arraydims);
	if (k == DOUBLE) return new Type(DOUBLE, arraydims);
	if (k == CHAR) return new Type(CHAR, arraydims);
	if (k == BYTE) return new Type(BYTE, arraydims);
	if (k == SHORT) return new Type(SHORT, arraydims);
	if (k == INT) return new Type(INT, arraydims);
	if (k == LONG) return new Type(LONG, arraydims);
	if (k == VOID) return new Type(VOID, arraydims);
	throw new Error("Unknown Type kind: "+k);
    }

    /**
     * Increments the number of dimension by 1.
     */
    public void incrementDimensions() {
        dimensions += 1;
    }

    /**
     * Sets the number of dimensions.
     *
     * @param dims  Number of dimensions to set this type to
     */
    public void setDimensions( int dims ) {
        dimensions = dims;
    }

    /**
     * Sets the name of the type.
     *
     * @param tname  New name of this type
     */
    public void setName( Object tname ) {
        if (tname instanceof Class) {
            setBytecodeName(((Class)tname).getName());
            return;
        }
        if (tname == null || tname.equals("")) {
            kind = UNKNOWN;
        } else {
            name = null;
            if (tname.equals("void") ||
                tname == Void.TYPE) kind = VOID;
            else if (tname.equals("boolean") ||
                     tname == Boolean.TYPE) kind = BOOLEAN;
            else if (tname.equals("byte") ||
                     tname == Byte.TYPE) kind = BYTE;
            else if (tname.equals("char") ||
                     tname == Character.TYPE) kind = CHAR;
            else if (tname.equals("short") ||
                     tname == Short.TYPE) kind = SHORT;
            else if (tname.equals("int") ||
                     tname == Integer.TYPE) kind = INT;
            else if (tname.equals("long") ||
                     tname == Long.TYPE) kind = LONG;
            else if (tname.equals("float") ||
                     tname == Float.TYPE) kind = FLOAT;
            else if (tname.equals("double") ||
                     tname == Double.TYPE) kind = DOUBLE;
            else {
                kind = OBJECT;
                if (tname instanceof String)                   
                    name = (String)tname;
                else if (tname instanceof Identifier)
                    name = ((Identifier)tname).getString();
                else if (tname instanceof Type)
                    name = ((Type)tname).getName();
                else {
                    System.err.println("Warning: Got a "+tname.getClass()+" in Type.setName()");
                    name = tname.toString();
                }
            }
        }
    }

    public void setBytecodeName(String s) {
        String rest = s;
        while (rest.charAt(0) == '[') {
            incrementDimensions();
            rest = rest.substring(1);
        }
        if (rest.endsWith(";")) {
            kind = OBJECT; name = rest.substring(1, rest.length()-1); 
            return;
        }
        if (rest.length() == 1) {
            switch (rest.charAt(0)) {
            case 'B': kind = BYTE; break;
            case 'C': kind = CHAR; break;
            case 'D': kind = DOUBLE; break;
            case 'F': kind = FLOAT; break;
            case 'I': kind = INT; break;
            case 'J': kind = LONG; break;
            case 'S': kind = SHORT; break;
            case 'V': kind = VOID; break;
            case 'Z': kind = BOOLEAN; break;
            default: throw new Error("Illegal type string "+s);
            }
            return;
        }
        if (s.equals("void")) kind = VOID;
        else if (s.equals("boolean")) kind = BOOLEAN;
        else if (s.equals("byte")) kind = BYTE;
        else if (s.equals("char")) kind = CHAR;
        else if (s.equals("short")) kind = SHORT;
        else if (s.equals("int")) kind = INT;
        else if (s.equals("long")) kind = LONG;
        else if (s.equals("float")) kind = FLOAT;
        else if (s.equals("double")) kind = DOUBLE;
        else {
            kind = OBJECT; 
            name = rest;
        }
        return;
    }

    /**
     * Creates a Jaemus.Type object for this type
     */
    public uiuc.Jumbo.Jaemus.Type toType() {
        uiuc.Jumbo.Jaemus.Type t;
        switch (kind) {
        case NULL: t = uiuc.Jumbo.Jaemus.Type.null_type; break;
        case OBJECT: t = new uiuc.Jumbo.Jaemus.Type(name); break;
        case VOID: t = uiuc.Jumbo.Jaemus.Type.void_type; break;
        case BOOLEAN: t = uiuc.Jumbo.Jaemus.Type.boolean_type; break;
        case CHAR: t = uiuc.Jumbo.Jaemus.Type.char_type; break;
        case FLOAT: t = uiuc.Jumbo.Jaemus.Type.float_type; break;
        case DOUBLE: t = uiuc.Jumbo.Jaemus.Type.double_type; break;
        case BYTE: t = uiuc.Jumbo.Jaemus.Type.byte_type; break;
        case SHORT: t = uiuc.Jumbo.Jaemus.Type.short_type; break;
        case INT: t = uiuc.Jumbo.Jaemus.Type.int_type; break;
        case LONG: t = uiuc.Jumbo.Jaemus.Type.long_type; break;
        default: throw new Error("Unknown kind "+kind);
        }
        for (int i = 0; i < dimensions; i++) {
            t = new uiuc.Jumbo.Jaemus.Type(t);
        }
        return t;
    }

    public IR toIR() {
        return IR.make(toType());
    }

    private String[] typenames = {
        "null",
        "array",
        "object",
        "void",
        "boolean",
        "char",
        "float",
        "double",
        "byte",
        "short",
        "int",
        "long"
    };

    /** 
     * Outputs this type as source
     */
    public void toSource() {
        String result;
        if (isPrimitive() || isVoid() || isNull())
            result = typenames[kind];
        else
            result = name;
        for (int i = 0; i < dimensions; i++) {
            result += "[]";
        }
        print(result);
    }

    public String toString() {
        String result;
        if (isPrimitive() || isVoid() || isNull())
            result = typenames[kind];
        else
            result = "class "+name;
        for (int i = 0; i < dimensions; i++) {
            result += "[]";
        }
        return result;
    }

    /** Return a Class representing this type.
     * This will only work for types representing existing classes
     * and primitive types.  For other classes, it returns the first
     * superclass that is an existing class (i.e. ClassWrapper)
     */
    public Class toClass(ClassEnv loader) {
        try {
            if (isArray()) {
                String dims = "";
                for (int i = 0; i < dimensions; i++) {
                    dims += "[";
                }
                switch (kind) {
                case OBJECT:
                    try {
                        return Class.forName(dims+"L"+name+";");
                    } catch (ClassNotFoundException e) {
                        throw new Error("Null class for type "+this);
                    }
                case VOID: return Class.forName(dims+"V");
                case BOOLEAN: return Class.forName(dims+"Z");
                case CHAR: return Class.forName(dims+"C");
                case FLOAT: return Class.forName(dims+"F");
                case DOUBLE: return Class.forName(dims+"D");
                case BYTE: return Class.forName(dims+"B");
                case SHORT: return Class.forName(dims+"S");
                case INT: return Class.forName(dims+"I");
                case LONG: return Class.forName(dims+"J");
                default: throw new Error("Unclassable array kind for "+this);
                }
            }
        } catch (ClassNotFoundException e) {
            throw new Error("Can't find class for array: "+e);
        }
        switch (kind) {
        case OBJECT: 
	    ClassDefiner cd = loader.getClass(name);
	    while (cd != null &&
		   !(cd instanceof ClassWrapper)) {
		if (cd instanceof InterfaceDeclaration) {
		    // Punt on superinterfaces -- don't know which to take
		    System.out.println("Using Object as superclass of "+cd);
		    cd = loader.getClass("java.lang.Object");
		} else {
		    cd = loader.getClass(cd.getSuperName());
		}
	    }
	    if (cd != null)
		return ((ClassWrapper)cd).thisclass;
            try {
                return Class.forName(name);
            } catch (ClassNotFoundException e) {
                throw new Error("Couldn't forName("+name+")");
            }
        case VOID: return Void.TYPE;
        case BOOLEAN: return Boolean.TYPE;
        case CHAR: return Character.TYPE;
        case FLOAT: return Float.TYPE;
        case DOUBLE: return Double.TYPE;
        case BYTE: return Byte.TYPE;
        case SHORT: return Short.TYPE;
        case INT: return Integer.TYPE;
        case LONG: return Long.TYPE;
        default: throw new Error("Unclassable kind for "+this);
        }
    }

    /**
     * Returns whether this type is an array type or not.
     *
     * @return True if this type is an array type, false otherwise.
     */
    public boolean isArray() {
       return (dimensions > 0);
    }


    /**
     * Returns whether this type is a primitive type or not.
     * This also returns true for an array of primitive types.
     *
     * @return True if this type is a primitive type, false otherwise.
     */
    public boolean isPrimitive() {
        return kind > VOID;
    }

    /**
     * Returns true if this type is a reference type (object, array or null).
     */
    public boolean isReference() {
        return kind <= OBJECT || dimensions > 0;
    }
    
    /**
     * Returns whether this type is the 'void' type.
     *
     * @return True if this type is 'void', false otherwise.
     */
    public boolean isVoid() {
        return kind == VOID;
    }

    /** Returns true if this type is the type of 'null'.
     */
    public boolean isNull() {
        return kind == NULL;
    }

    /**
     * Returns a new type representing the type of the contents of this array.
     */
    public Type getElementType() {
        if (dimensions == 0) throw new Error("Taking element type of non-array type "+this);
        return new Type(kind, name, dimensions-1);
    }

    /** Returns the name in this type, if non-primitive.
     * For an array type, returns the name of the array element.
     */
    public String getName() {
        if (isPrimitive()) throw new Error("Taking class name of primitive type "+this);
	if (isNull()) throw new Error("Taking class name of null type");
        return name;
    }

    /** Returns the dimensions of this type -- 0 for non-arrays.
     */
    public int getDimensions() {
        return dimensions;
    }

    public int getKind() {
        return kind;
    }

    public boolean equals(Object o) {
        if (!(o instanceof Type)) return false;
        Type t = (Type)o;
        if (t.kind != kind) return false;
        if (t.isPrimitive()) return true;
        if (t.dimensions != dimensions) return false;
        return t.name.equals(name);
    }

    /** May this primitive type be widened to the given primitive kind? */
    public boolean mayWidenPrimitive(int kind) {
        if (this.kind == kind) return true;
        switch (this.kind) {
        case BOOLEAN:
            return false; // Only boolean conforms to boolean
        case BYTE:
            return (kind == SHORT ||
                    kind == INT ||
                    kind == LONG ||
                    kind == FLOAT ||
                    kind == DOUBLE);
        case SHORT: 
        case CHAR:
            return (kind == INT ||
                    kind == LONG ||
                    kind == FLOAT ||
                    kind == DOUBLE);
        case INT:
            return (kind == LONG ||
                    kind == FLOAT ||
                    kind == DOUBLE);
        case LONG:
            return (kind == FLOAT ||
                    kind == DOUBLE);
        case FLOAT:
            return (kind == DOUBLE);
        case DOUBLE:
            return false; // Only double conforms to double
        default:
            return false;
        }
    }
    
    /** Return the common ancestor of two types.
     */
    public Type commonAncestor(Type other, TypeEnv typeenv) {
        if (other.conforms(this, typeenv)) return this;
        if (conforms(other, typeenv)) return other;
        List supers = typeenv.getSupers(this);
        List othersupers = typeenv.getSupers(other);
        /* Check if we reached the ceiling */
        if (supers.size() == 0) return this;
        if (othersupers.size() == 0) return other;
        Type supert = (Type)supers.get(0);
        Type othert = (Type)othersupers.get(0);
        Type common = supert.commonAncestor(othert, typeenv);
        return common;
    }

    /** Check that this type conforms (methodwise) to the given type.
     * @param t A Type.
     * @param superfinder An implementation of SuperFinder.
     * @return True if t is the same type as this, or a supertype.
     */
    public boolean conforms(Type t, TypeEnv typeenv) {
        if (isArray()) {
            if (t.dimensions > dimensions) return false;
            if (t.dimensions < dimensions) {
                // t's elements must be Object
                if (t.name.equals("Object") ||
                    t.name.equals("java.lang.Object"))
                    return true;
                return false;
            }
            // Otherwise, we must check the objects anyway
        }
        // If the array dims match, check the elements
        if (isPrimitive()) {
            // Check for casting of primitive types
            return mayWidenPrimitive(t.kind);
        }
        // Since this isn't primitive, it cannot conform to a primitive
        if (t.isPrimitive()) return false;
        // Null conforms to any object type
        if (kind == NULL) return true;
        // Objects must check for subclassing, even inside arrays
        Type othert = typeenv.getFQCN(t);
        Type thist = typeenv.getFQCN(this);
        if (othert == null) {
            System.out.println("Null FQCN for "+othert);
            return false;
        }
        if (thist == null) throw new Error("Null thist for "+this);
        if (thist.getName() == null) throw new Error("Null getName for "+thist);
        if (thist.getName().equals(othert.getName())) return true;
        List supers = typeenv.getSupers(this);
        for (Iterator i = supers.iterator(); i.hasNext(); ) {
            Type supert = (Type)i.next();
            if (supert.conforms(othert, typeenv)) return true;
        }
        return false;
    }

    /** Return a literal of the zero value for this type.
     */
    public Literal zeroLiteral() {
        switch (kind) {
        case BOOLEAN:
            return new BooleanLiteral(false);
        case BYTE:
        case SHORT: 
        case INT:
            return new IntegerLiteral(0);
        case CHAR:
            return new CharacterLiteral('\u0000');
        case LONG:
            return new LongLiteral(0l);
        case FLOAT:
            return new FloatLiteral(0f);
        case DOUBLE:
            return new DoubleLiteral(0f);
        case OBJECT:
        case NULL:
            return new NullLiteral();
        default:
            throw new Error("Uninstantiable type "+this);
        }
    }
}


/** Interface for functions for finding environment information on Types.*/
interface TypeEnv {
    /** Returns the list of superclass types (including interfaces).
     * For an object, the superclass is always first */
    List getSupers(Type t);
    /** Returns the type with any class names in FQCN form
     * Primitive types are returned unchanged.  */
    Type getFQCN(Type t);
}
