/* Copyright (C) 2003, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package uiuc.Jumbo.Ast;
import uiuc.Jumbo.Util.*;

/**
 * A simple variable, in this or an outer method.  This represents
 * the use of a variable, not the declaration.  It is essentially
 * a one-field QNE that is known to be a variable.
 *
 * @author      Lars Clausen
 * @version     %I%, %G%
 * @since       JDK 1.3
 */

public class VarExpression extends QualifiedNameExpression  {
    /**
     * @param n  the name of the expression
     */
    public VarExpression( Identifier n ) { 
	super(n);
    }

    public VarExpression( String s) {
	super(new Identifier(Util.unjoin(s, ".")));
    }

    /**
     * Returns a new type object based on this name.
     */
    public Type toType() {
	throw new Error("Trying to turn a variable into a type.");
    }

    /** Get the variable declaration for this QualifiedNameExpression
     * without depending on outside variable lookup.
     * This is *not* the same as getVariableDeclaration.  This will
     * actually look through outer classes as well, and check for fields.
     * @param from The point in the AST to find variables in.  In the
     * recursive calls, this will be the class definition node.
     * @param loader A ClassEnv instance.
     * @return The node that declares the variable in the first part
     * of the QNE, or null if it's not a variable.
     */
    public Variable getVarDecl(Node from, ClassEnv loader) {
        String first = name.getField(0).getString();
        Variable v = from.getVariableDeclaration(first);
        if (v == null) {
            ClassDefiner thisclass = from.getEnclosingClass();
            ClassDefiner enclosing = ((Node)thisclass).getEnclosingClass();
            if (enclosing == null) {
                return null;
            }
	    return getVarDecl((Node)thisclass, loader);
        }
        return v;
    }

    /** Returns the declaration of a variable.
     */
    public Variable getDecl() {
	return getVarDecl(this, null);
    }

    /** Returns the type of the beginning variable or field.
     */
    public Type getFieldOrVarType(Node from,
                                  Identifier name, ClassEnv loader) {
        ClassDefiner thisclass = from.getEnclosingClass();
        // This can happen with imports statements.
        if (thisclass == null) return null;
        String first = name.getField(0).getString();
        Variable v = from.getVariableDeclaration(first);
        if (v == null) {
            return getFieldOrVarType((Node)thisclass, name, loader);
        } 
        return v.type;
    }

    /** Returns the first variable or field with the given name, or null.
     * This function scans through outer classes too.  
     */
    public QNEMarker getFirstField(ClassEnv loader) {
	QNEMarker marker = new QNEMarker();
	marker.field = 1;
	marker.type = getType(loader);
	marker.isVar = true;
	return marker;
    }

    public Type getType(ClassEnv loader) {
	Variable v = getVarDecl(this, loader);
	if (v == null) {
	    String first = name.getField(0).getString();
	    traceVariableDeclaration(first);
	    throw new Error("No decl found for "+this+" at "+sourceline +" in "+parent);
	}
	return v.type;
    }
}
