/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Copyright (C) 1998-2001  Gerwin Klein <lsf@jflex.de>                    *
 * All rights reserved.                                                    *
 *                                                                         *
 * This program is free software; you can redistribute it and/or modify    *
 * it under the terms of the GNU General Public License. See the file      *
 * COPYRIGHT for more information.                                         *
 *                                                                         *
 * This program is distributed in the hope that it will be useful,         *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc., *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                 *
 *                                                                         *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

package uiuc.Jumbo.Parser;

import java.io.*;
import java.util.*;
import uiuc.Jumbo.Ast.*;
import uiuc.Jumbo.Compiler.*;
import uiuc.Jumbo.Jaemus.ClassInfo;

/**
 * Simple driver for the java parser. 
 *
 * Usage: <code>java Parser.JavaParser [-d &lt;<i>directory</i>&gt;] &lt;<i>file1</i>&gt; &lt;<i>file2</i>&gt; ...</code>
 *
 * NOTES: 
 * <ul>
 *   <li> <i>directory</i> must exist.
 *   <li> Use <code>-</code> for standard input.
 *   <li> Use <code>-c</code> to create constructors directly.
 *   <li> Use <code>-s</code> to output source directly.
 * </ul>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License. See the file
 * COPYING for more information.
 */
public class JavaParser {
    private static final int OUTPUT_SOURCE = 0;
    private static final int CREATE_CONSTRUCTORS = 1;
    private static final int OUTPUT_LIFTED = 2;

    public static Node parse(Reader r) 
        throws java.lang.Exception {
        Scanner s = new Scanner(new UnicodeEscapes(r));
        parser p = new parser(s);
        return (Node)p.parse().value;
    }

    public static void main(String argv[]) {
        boolean errors = false;
        int output = CREATE_CONSTRUCTORS;
        String sourcepath = "";
	int i;
        for (i = 0; i < argv.length; i++) {
            if (argv[i].equals( "-d" )) {
                // set the output directory.
                i++;
                ClassInfo.outputdir = argv[i] + File.separator;
            } else if (argv[i].equals( "-sourcepath" )) {
                // Set input directory
                i++;
                sourcepath = argv[i] + File.separator;
            } else if (argv[i].equals( "-v" )) {
                System.out.println("Jumbo version 0.8.0");
                System.exit(0);
            } else if (argv[i].equals( "-s" )) {
                output = OUTPUT_SOURCE;
            } else if (argv[i].equals( "-l" )) {
                output = OUTPUT_LIFTED;
	    } else if (argv[i].equals("--")) {
		i++;
		break;
            } else if (argv[i].length() > 1 && argv[i].charAt(0) == '-') {
                System.out.println("Usage: jumbo [-sourcepath <dir>|-d <dir>|-c|-l|-v|-h] files");
                System.exit(0);
            } else {
		// Have come to end of options.  Now deal with input.
		break;
	    }
	}
	/** All options parsed, now parse input */
	List inputs = new ArrayList();
	for (; i < argv.length; i++) {
	    Node result = null;
	    try {
		Reader r = null;
		if ( argv[i].equals( "-" ) ) {
		    BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
		    StringBuffer buf = new StringBuffer();
		    String s;
		    do {
			s = br.readLine();
			if (s != null) { buf.append(s); buf.append("\n"); }
		    } while (s != null);
		    r = new StringReader(buf.toString());
		    Node.setOutputToSystemOut();
		} else {
		    r = new FileReader( sourcepath + argv[i] );

		    String filename = argv[i];
		    int index = filename.lastIndexOf(File.separatorChar);
		    if ( index >= 0 ) {
			if (index == filename.length()) {
			    System.err.println( "Will not parse a filename ending in "+File.separator+"; skipping " + argv[i] );
			    continue;
			}
			filename = filename.substring( index+1 );
		    }

		}
		result = parse(r);
	    } catch (Throwable e) {
		System.err.println(argv[i] + ":Parse error:"+e );
		e.printStackTrace(System.out);
		errors = true;
		continue;
	    }
	    inputs.add(result);
	}
	if (output == OUTPUT_SOURCE) {
	    Node.setOutputToSystemOut();
	}
	if (inputs.size() > 0) {
	    Node fullinput;
	    if (inputs.size() == 1) 
		fullinput = (Node)inputs.get(0);
	    else {
		StatementList l = new StatementList();
		for (Iterator ins = inputs.iterator(); ins.hasNext(); ) {
		    Node n = (Node)ins.next();
		    l.appendStatement(new CompilationUnit(n));
		}
		fullinput = l;
	    }
	    switch (output) {
	    case OUTPUT_SOURCE:
		outputSource(fullinput);
		Node.endOutput();
		break;
	    case CREATE_CONSTRUCTORS:
		Code c;
		try {
		    c = (Code)fullinput.toIR().toObject();
		} catch (Throwable e) {
		    e.printStackTrace(System.out);
		    errors = true;
		    break;
		}
		try {
		    c.generate();
		} catch (Throwable e) {
		    e.printStackTrace(System.out);
		    errors = true;
		    break;
		}
		break;
	    case OUTPUT_LIFTED:
		String s;
		try {
		    s = fullinput.toIR().toSource();
		} catch (Throwable e) {
		    e.printStackTrace(System.out);
		    errors = true;
		    break;
		}

		System.out.println(s);
		break;
	    }
	}
        if (errors) System.exit(1);
    }

    public static void outputSource(Object o) {
        if (o instanceof Node)
            ((Node)o).toSource();
        else if (o instanceof LinkedList)
            Node.printSourceListWithUnknowns((LinkedList)o);
        else
            throw new Error("Unknown type returned from parser: "+o.getClass());
    }

}
