/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*
 * ArrayMonoList.java
 *
 *
 */

package uiuc.Jumbo.Util;

import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 * An array-based implementation of the MonoList interface
 * @version      1.0,  8 Nov 2001
 * @author       Lars Clausen
 *
*/

public final class ArrayMonoList implements MonoList {
    protected Object[] elements;
    protected int first_index, last_index;

    public ArrayMonoList() {
        elements = new Object[10];
    }

    public ArrayMonoList(int start_size) {
        elements = new Object[start_size];
    }

    public ArrayMonoList(Object first_elem) {
        elements = new Object[10];
        add(first_elem);
    }

    public ArrayMonoList(Object first_elem, int start_size) {
        elements = new Object[start_size];
        add(first_elem);
    }
    
    public ArrayMonoList(String elements, char delimiter) {
        this(1);
        int lasti = 0;
        for (int i = elements.indexOf(delimiter);
             i != -1; i = elements.indexOf(delimiter, lasti)) {
            add(elements.substring(lasti, i));
            lasti = i+1;
        }
    }
    
    public MonoList add(Object o) {
        if (last_index >= elements.length) {
            ensureCapacity(elements.length+1);
        }
        elements[last_index++] = o;
        return this;
    }
    public Iterator iterator() {
        return new MonoIterator(this);
    }
    final class MonoIterator implements Iterator {
        int iter_index;
        final ArrayMonoList a;

        MonoIterator(ArrayMonoList t) {
            iter_index = t.first_index;
            a = t;
        }
        public boolean hasNext() {
            return iter_index < a.last_index;
        }
        public Object next() {
            if (iter_index == a.last_index) throw new NoSuchElementException();
            return a.elements[iter_index++];
        }
        public void remove() {
	    if(iter_index == a.first_index) throw new NoSuchElementException();
	    for(int i=iter_index-1; i<a.last_index; i++){
		a.elements[i] = a.elements[i+1];
	    }
	    iter_index--;
	    a.last_index--;
        }
    }
    public Iterator reverseIterator() {
        return new ReverseMonoIterator(this);
    }
    final class ReverseMonoIterator implements Iterator {
        int iter_index;
        final ArrayMonoList a;

        ReverseMonoIterator(ArrayMonoList t) {
            iter_index = t.last_index;
            a = t;
        }
        public boolean hasNext() {
            return iter_index > a.first_index;
        }
        public Object next() {
            if (iter_index == a.first_index) throw new NoSuchElementException();
            return a.elements[--iter_index];
        }
        public void remove() {
            throw new UnsupportedOperationException();
        }
    }
    public Object foldr(FoldOp op, Object zero) {
        Object res = zero;
        for (Iterator i = reverseIterator(); i.hasNext(); ) {
            res = op.eval(i.next(), res);
        }
        return res;
    }
    public Object foldl(FoldOp op, Object zero) {
        Object res = zero;
        for (Iterator i = iterator(); i.hasNext(); ) {
            res = op.eval(res, i.next());
        }
        return res;
    }
    public Object get(int i) {
        return elements[i+first_index];
    }
    public MonoList subList(int from, int to) {
        ArrayMonoList a = new ArrayMonoList(to-from);
        a.elements = elements;
        a.first_index = from+first_index;
        a.last_index = to+first_index;
        return a;
    }
    public int size() {
        return last_index - first_index;
    }
    public MonoList addAll(MonoList l) {
        if (l == null) return this;
        ensureCapacity(size()+l.size());
        if (l instanceof ArrayMonoList) {
            ArrayMonoList a = (ArrayMonoList)l;
            System.arraycopy(a.elements, a.first_index, elements, last_index, l.size());
            last_index += a.size();
        } else {
            for (Iterator i = l.iterator(); i.hasNext(); ) {
                add(i.next());
            }
        }
        return this;
    }
    private void ensureCapacity(int capacity) {
        if (capacity >= elements.length+first_index) {
            Object[] newelements = new Object[capacity*2];
            System.arraycopy(elements, first_index,
                             newelements, 0, last_index-first_index);
            elements = newelements;
            last_index = last_index-first_index;
            first_index = 0;
        }
    }

    public String toString() {
        String result = "[";
        for (int i = first_index; i < last_index; i++) {
            result += elements[i];
            if (i < last_index-1) result+=", ";
        }
        return result+"]";
    }

    public Object[] toArray() {
        Object[] array = new Object[last_index-first_index];
        System.arraycopy(elements, first_index, array, 0, array.length);
        return array;
    }

    public MonoList copy() {
        ArrayMonoList newl = new ArrayMonoList(size());
        System.arraycopy(elements, first_index, newl.elements, 0, size());
        newl.first_index = 0;
        newl.last_index = size();
        return newl;
    }
}

/* End of class ArrayMonoList*/
