/* Copyright (C) 2003, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*
 * FixedMonoList.java
 *
 *
 */

package uiuc.Jumbo.Util;

import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 * An array-based implementation of the MonoList interface that
 * assumes the length is commonly known in advance, and so optimizes
 * for that case.
 * @version      1.0,  8 Nov 2001
 * @author       Lars Clausen
 *
*/

public final class FixedMonoList implements MonoList {
    final public Object[] elements;
    final public int first_index;
    public int last_index;

    public FixedMonoList() {
        elements = new Object[10];
        first_index = 0;
        last_index = 0;
    }

    public FixedMonoList(int size) {
        elements = new Object[size];
        first_index = 0;
        last_index = 0;
    }

    public FixedMonoList(Object first_elem) {
        elements = new Object[10];
        first_index = 0;
        last_index = 0;
        add(first_elem);
    }

    public FixedMonoList(Object first_elem, int start_size) {
        elements = new Object[start_size];
        first_index = 0;
        last_index = 0;
        add(first_elem);
    }
    
    public FixedMonoList(String elements, char delimiter) {
        this(10);
        int lasti = 0;
        for (int i = elements.indexOf(delimiter);
             i != -1; i = elements.indexOf(delimiter, lasti)) {
            add(elements.substring(lasti, i));
            lasti = i+1;
        }
    }
    
    /** Construct a sublist */
    protected FixedMonoList(int start, FixedMonoList father) {
        elements = father.elements;
        first_index = start;
        last_index = father.last_index;
    }

    /** Construct a sublist */
    protected FixedMonoList(int start, int end, FixedMonoList father) {
        elements = father.elements;
        first_index = start;
        last_index = end;
    }

    /** Create a new longer list */
    protected FixedMonoList(Object o, FixedMonoList father) {
        elements = new Object[father.elements.length*2];
        first_index = 0;
        last_index = father.last_index-father.first_index;
        System.arraycopy(father.elements, father.first_index,
                         elements, 0, last_index);
        add(o);
    }

    public MonoList add(Object o) {
        if (last_index >= elements.length) {
            return new FixedMonoList(o, this);
        } else {
            elements[last_index] = o;
            last_index++;
            return this;
        }
    }
    public Iterator iterator() {
        return new MonoIterator(this);
    }
    class MonoIterator implements Iterator {
        int iter_index;
        final int last_index;
        final Object[] a;

        MonoIterator(FixedMonoList t) {
            iter_index = t.first_index;
            a = t.elements;
            last_index = t.last_index;
        }
        public boolean hasNext() {
            return iter_index < last_index;
        }
        public Object next() {
            if (iter_index == last_index) throw new NoSuchElementException();
            else return elements[iter_index++];
        }
        public void remove() {
            throw new UnsupportedOperationException();
        }
    }
    public Iterator reverseIterator() {
        return new ReverseMonoIterator(this);
    }
    class ReverseMonoIterator implements Iterator {
        int iter_index;
        final int first_index;
        final Object[] a;

        ReverseMonoIterator(FixedMonoList t) {
            iter_index = t.last_index;
            a = t.elements;
            first_index = t.first_index;
        }
        public boolean hasNext() {
            return iter_index > first_index;
        }
        public Object next() {
            if (iter_index == first_index) throw new NoSuchElementException();
            return elements[--iter_index];
        }
        public void remove() {
            throw new UnsupportedOperationException();
        }
    }
    public Object foldr(FoldOp op, Object zero) {
        Object res = zero;
        for (Iterator i = reverseIterator(); i.hasNext(); ) {
            res = op.eval(i.next(), res);
        }
        return res;
    }
    public Object foldl(FoldOp op, Object zero) {
        Object res = zero;
        for (Iterator i = iterator(); i.hasNext(); ) {
            res = op.eval(res, i.next());
        }
        return res;
    }
    public Object get(int i) {
        return elements[i+first_index];
    }
    public MonoList subList(int from, int to) {
        return new FixedMonoList(first_index+from,
                                 first_index+to,
                                 this);
    }
    public int size() {
        return last_index - first_index;
    }
    public MonoList addAll(MonoList l) {
        if (l == null) return this;
        if (l.size() == 0) return this;
        if (last_index+l.size() > elements.length) {
            FixedMonoList newl = new FixedMonoList(2*(size()+l.size()));
            for (Iterator i = iterator(); i.hasNext(); ) {
                newl.add(i.next());
            }
            for (Iterator i = l.iterator(); i.hasNext(); ) {
                newl.add(i.next());
            }
            return newl;
        } else {
            for (Iterator i = l.iterator(); i.hasNext(); ) {
                add(i.next());
            }
            return this;
        }
    }
    public String toString() {
        String result = "[";
        for (int i = first_index; i < last_index; i++) {
            result += elements[i];
            if (i < last_index-1) result+=", ";
        }
        return result+"]";
    }

    public Object[] toArray() {
        Object[] array = new Object[last_index-first_index];
        System.arraycopy(elements, first_index, array, 0, array.length);
        return array;
    }

    public MonoList copy() {
        FixedMonoList newl = new FixedMonoList(size());
        System.arraycopy(elements, first_index, newl.elements, 0, size());
        return newl;
    }
}

/* End of class FixedMonoList*/
