/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*
 * LinkedMonoList.java
 *
 *
 */

package uiuc.Jumbo.Util;

import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 * An linked-list-based implementation of the MonoList interface.
 * This implementation keeps the list internally in *reverse* order.
 * This allows add() to be constant time, always, but the iterator
 * is funky.  On the other hand, the foldr and foldl operators become
 * linear time.
 * @version      1.0,  8 Nov 2001
 * @author       Lars Clausen
 *
*/

public final class LinkedMonoList implements MonoList {
    final protected Object data;
    final protected LinkedMonoList next;

    public LinkedMonoList(Object elem) {
        data = elem;
        next = null;
    }

    private LinkedMonoList(final Object elem, final LinkedMonoList next) {
        this.data = elem;
        this.next = next;
    }

    public LinkedMonoList(final String elements, final char delimiter) {
        final int i = elements.indexOf(delimiter);
        if (i == -1) {
            data = elements;
            next = null;
        } else {
            data = elements.substring(0, i);
            next = new LinkedMonoList(elements.substring(i+1), delimiter);
        }
    }
    
    public MonoList add(final Object o) {
        if (next != null)
            return new LinkedMonoList(data, (LinkedMonoList)next.add(o));
        else
            return new LinkedMonoList(data, new LinkedMonoList(o));
    }
    public Iterator iterator() {
        return new LinkedMonoIterator(this);
    }

    private LinkedMonoList reverse(LinkedMonoList acc) {
        if (next == null) return new LinkedMonoList(data, acc);
        else return next.reverse(new LinkedMonoList(data, acc));
    }
    /* Not side-effect free */
    final class LinkedMonoIterator implements Iterator {
        LinkedMonoList a;

        LinkedMonoIterator(final LinkedMonoList t) {
            a = t;
        }
        public boolean hasNext() {
            return a != null;
        }
        public Object next() {
            if (a == null) throw new NoSuchElementException();
            final Object d = a.data;
            a = a.next;
            return d;
        }
        public void remove() {
            throw new UnsupportedOperationException();
        }
    }
    public Object foldr(FoldOp op, Object zero) {
        if (next == null)
            return op.eval(data, zero);
        else
            return op.eval(data, next.foldr(op, zero));
    }
    public Object foldl(FoldOp op, Object zero) {
        if (next == null)
            return op.eval(zero, data);
        else
            return next.foldl(op, op.eval(zero, data));
    }
    public Object get(final int i) {
        if (i == 0) return data;
        else if (next == null) throw new IndexOutOfBoundsException(""+i);
        else return next.get(i-1);
    }
    public MonoList subList(final int from) {
        if (from > 0) {
            if (next == null) throw new IndexOutOfBoundsException(""+from);
            else return next.subList(from-1);
        } else {
            return this;
        }
    }
    public MonoList subList(final int from, final int to) {
        if (from > 0) {
            if (next == null) throw new IndexOutOfBoundsException(""+from);
            else return next.subList(from-1, to-1);
        } else {
            if (to == 1)
                return new LinkedMonoList(data);
            else if (next == null && to > 0) throw new IndexOutOfBoundsException(""+to);
            else return new LinkedMonoList(data, (LinkedMonoList)next.subList(0, to-1));
        }
    }
    public int size() {
        if (next == null) return 1;
        else return next.size()+1;
    }
    public boolean isEmpty() {
        return false;
    }
    /* Not side-effect free */
    public MonoList addAll(MonoList l) {
        MonoList l1 = this;
        for (Iterator i = l.iterator(); i.hasNext(); ) {
            l1 = l1.add(i.next());
        }
        return l1;
    }

    private String toStringRest(String s) {
        if (next != null) 
            return next.toStringRest(s+", "+this);
        else
            return s+", "+this+"]";
    }

    public String toString() {
        final String result = "["+data;
        if (next != null) return next.toStringRest(result);
        else return result+"]";
    }

    /* Sideffecting */
    public Object[] toArray() {
        final Object[] array = new Object[size()];
        int index = 0;
        for (Iterator i = iterator(); i.hasNext(); ) {
            array[index++] = i.next();
        }
        return array;
    }

    public MonoList copy() {
        return this;
    }
}

/* End of class LinkedMonoList*/
