/* Copyright (C) 2002, University of Illinois.  All rights reserved. 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*
 * Constructor.java
 *
 * Sam Kamin <kamin@cs.uiuc.edu>, Ava Arachne Jarvis <ajar@uiuc.edu>,
 * Lars Clausen <lrclause@cs.uiuc.edu>
 *
 * The compiler is currently 4-pass:
 * 1) Find all classes and define them, with appropriate package prefixes.
 *    In this pass, also find the possible free variables for inner classes.
 *    We are being slightly conservative, as we can't know the fields.
 * 2) Normalize all super class names.
 * 3) Find all members (fields, methods, inners) and define those,
 *    normalizing class names to contain the full package.
 * 4) Evaluate the code itself.
 *
 * Optimization ideas (for increased compilation speed):
 * - Find all inner clases in the first pass and store them in their
 *   enclosing class (method?).  Then second and third pass doesn't 
 *   need to go through the code.
 * - Store the methods and fields in the first pass, but without normalizing
 *   types.  In the second pass, normalize types and add inner
 *   constructors.
 *   
 */

package uiuc.Jumbo.Compiler;

import uiuc.Jumbo.Jaemus.*;
import java.util.*;
import uiuc.Jumbo.Util.*;
import java.io.*;

public final class Constructor {
    /** For inner class qualified name lookup, to tell whether to
     * restrict to none, variables, or fields.
     */
    private static final int ALLOW_BOTH = 0;
    private static final int ALLOW_VARIABLES = 1;
    private static final int ALLOW_FIELDS = 2;
    private static final boolean use_fold = true;

    protected static Error error(String message, CodeEnvironment env) {
        if (env == null) return new Error(message);
        else return env.error(message);
    }

    protected static boolean flagsLegal(int legalflags, int actualflags) {
        return (actualflags & (~legalflags)) == 0;
    }

    protected static boolean flagsCompatible(int flagsset, int flagsforbidden,
                                           int actualflags) {
        return ((flagsset & actualflags) == 0 ||
                (flagsforbidden & actualflags) == 0);
    }

    /** Converts a list of objects into a 'join'-separated string */
    protected static String toJoinedString(final Object[] l, final String join) {
        String res = "";
        for (int i = 0; i < l.length; i++) {
            res += l[i];
            if (i < l.length-1)
                res += join;
        }
        return res;
    }

    /** Converts a list of objects into a 'join'-separated string */
    protected static String toJoinedString(final MonoList l, final String join) {
        String res = "";
        for (Iterator i = l.iterator(); i.hasNext(); ) {
            res += i.next();
            if (i.hasNext())
                res += join;
        }
        return res;
    }

    /** Inverts the test in the boolean sense (true => false, false => true)
     *  The tests are arranged so that flipping the least bit gives inverse! 
     */
    protected static final int invertCmp(int cmp) {
        return cmp ^ 1;
    };

    /** Reverses the order of a comparision.
     *  EQ, NE stays the same, GT <=> LE, LT <=> GE
     */
    protected static final int reverseCmp(int cmp) {
        if ((cmp >> 1) == (Instr.CMP_EQ >> 1)) return cmp;
        return cmp ^ 1;
    };

    public static Code integerConstant(final int constval) {
        return new Code() {
                public String getName() { return "integerConstant"; }
                public ClosedCode eval(Environment env) {
                    return new ClosedCode(Instr.genConst(constval),
                                          env, Type.int_type,
                                          ConstVal.make(constval));
                }
                public String toString() { return Integer.toString(constval);}
            };
    }
    public static Code constant(final int constval) {
        return integerConstant(constval);
    }

    public static Code booleanConstant(final boolean constval) {
        return new Code() {
                public String getName() { return "booleanConstant"; }
                public ClosedCode eval(Environment env) {
                    return new ClosedCode(Instr.genConst(constval?1:0),
                                          env, Type.boolean_type,
                                          ConstVal.make(constval));
                }
                public ClosedCode evalWithLabels(Environment env,
                                                 Instr truelabel,
                                                 Instr falselabel,
                                                 Instr nextlabel) {
                    Instr code;
                    if ((constval && truelabel == nextlabel) ||
                        (!constval && falselabel == nextlabel))
                        code = Instr.genNil("Goto next");
                    else
                        code = Instr.genGoto(constval?truelabel:falselabel);
                    return new ClosedCode(code,
                                          env, Type.void_type,
                                          ConstVal.make(constval));
                }
                public String toString() { return (constval?"true":"false"); }
            };
    }
    public static Code constant(final boolean constval) {
        return booleanConstant(constval);
    }

    public static Code floatConstant(final float constval) {
        return new Code() {
                public String getName() { return "floatConstant"; }
                public ClosedCode eval(Environment env) {
                    return new ClosedCode(Instr.genConst(constval),
                                          env, Type.float_type,
                                          ConstVal.make(constval));
                }
                public String toString() {return Float.toString(constval)+"f";}
            };
    }
    public static Code constant(final float constval) {
        return floatConstant(constval);
    }

    public static Code longConstant(final long constval) {
        return new Code() {
                public String getName() { return "longConstant"; }
                public ClosedCode eval(Environment env) {
                    return new ClosedCode(Instr.genConst(constval),
                                          env, Type.long_type,
                                          ConstVal.make(constval));
                }
                public String toString() { return Long.toString(constval)+"l";}
            };
    }
    public static Code constant(final long constval) {
        return longConstant(constval);
    }

    public static Code doubleConstant(final double constval) {
        return new Code() {
                public String getName() { return "doubleConstant"; }
                public ClosedCode eval(Environment env) {
                    return new ClosedCode(Instr.genConst(constval),
                                          env, Type.double_type,
                                          ConstVal.make(constval));
                }
                public String toString() { return Double.toString(constval);}
            };
    }
    public static Code constant(final double constval) {
        return doubleConstant(constval);
    }

    public static Code characterConstant(final char constval) {
        return new Code() {
                public String getName() { return "characterConstant"; }
                public ClosedCode eval(Environment env) {
                    return new ClosedCode(Instr.genConst(constval),
                                          env, Type.char_type,
                                          ConstVal.make(constval));
                }
                public String toString() {
                    if (Character.isISOControl(constval)) {
                        if (constval == '\n') return "'\\n'";
                        if (constval == '\r') return "'\\r'";
                        if (constval == '\t') return "'\\t'";
                        if (constval == '\b') return "'\\b'";
                        return "'\\"+Integer.toHexString((int)constval);
                    }
                    return "'"+constval+"'"; }
            };
    }
    public static Code constant(final char constval) {
        return characterConstant(constval);
    }

    public static Code stringConstant(final String constval) {
        return new Code() {
                public String getName() { return "stringConstant"; }
                public ClosedCode eval(Environment env) {
                    return new ClosedCode(Instr.genConst(constval),
                                          env, Type.string_type,
                                          ConstVal.make(constval));
                }
                public String toString() { return "\""+constval+"\"";}
            };
    }
    public static Code constant(final String constval) {
        return stringConstant(constval);
    }

    public static Code nullConstant() {
        return new Code() {
                public String getName() { return "nullConstant"; }
                public ClosedCode eval(Environment env) {
                    return new ClosedCode(Instr.genNull(),
                                          env, Type.null_type,
                                          ConstVal.makeNull());
                }
                public String toString() { return null;}
            };
    }

    public static Code antiquoted(Code c) { return c; }
    public static Code antiquoted(String c) { return constant(c); }
    public static Code antiquoted(boolean c) { return constant(c); }
    public static Code antiquoted(byte c) { return constant(c); }
    public static Code antiquoted(char c) { return constant(c); }
    public static Code antiquoted(short c) { return constant(c); }
    public static Code antiquoted(int c) { return constant(c); }
    public static Code antiquoted(long c) { return constant(c); }
    public static Code antiquoted(float c) { return constant(c); }
    public static Code antiquoted(double c) { return constant(c); }
    
    public static Code noop() {
        return new Code() {
                public String getName() { return "noop"; }
                public ClosedCode eval(Environment env) {
                    return new ClosedCode(Instr.genNil("Noop"),
                                          env, Type.void_type);
                }
                public ClosedCode evalWithLabels(Environment env,
                                                 Instr truelabel,
                                                 Instr falselabel,
                                                 Instr nextlabel) {
                    if (truelabel == nextlabel)
                        return new ClosedCode(Instr.genNil("Noop"),
                                              env, Type.void_type);
                    else
                        return new ClosedCode(Instr.genGoto(truelabel),
                                              env, Type.void_type);
                }
                public String toString() 
                {
                    return "";
                }
            };
    }

    /** Bitwise complment.
     * @param c Code to bitwise complement
     */
    public static Code bitwiseComplement(final Code c) {
        return new Code() {
                public String getName() { return "bitwiseComplement"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return c.defineClasses(e);
                }
                public Environment defineSupers(Environment e) {
                    return c.defineSupers(e);
                }
                public Environment defineMembers(Environment e) {
                    return c.defineMembers(e);
                }
                public ClosedCode eval(Environment env) {
                    final ClosedCode c1 = c.eval(env);
                    if (!c1.type.isDecimal())
                        throw error("Doing bitwise complement on non-decimal type "+c1.type, env);
                    if (c1.isConstant()) {
                        final ConstVal cc = c1.constantValue().bitwiseNegOp();
                        return new ClosedCode(cc.genLoadCode(),
                                              env, c1.type, cc)
                            /*.addInfo(c1)*/;
                    }
                    else {
                        return new ClosedCode(c1.bytecode.
                                              addArith(Instr.NEG, c1.type).
                                              addConst(1, c1.type).
                                              addMath(Instr.SUB, c1.type),
                                              env, c1.type)/*.addInfo(c1)*/;
                    }
                }
                public String toString() { return "~"+c; }
            };
    }

    /** Mathematical negation (-x)
     * @param c Code to negate
     */
    public static Code negate(final Code c) {
        return new Code() {
                public String getName() { return "negate"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return c.defineClasses(e);
                }
                public Environment defineSupers(Environment e) {
                    return c.defineSupers(e);
                }
                public Environment defineMembers(Environment e) {
                    return c.defineMembers(e);
                }
                public ClosedCode eval(Environment env) {
                    final ClosedCode c1 = c.eval(env);
                    if (!c1.type.isNumeric())
                        throw error("Negating non-numeric type "+c1.type, env);
                    if (c1.isConstant()) {
                        final ConstVal cc = c1.constantValue().numericalNegOp();
                        return new ClosedCode(cc.genLoadCode(),
                                              env, c1.type, cc)
                            /*.addInfo(c1)*/;
                    }
                    else
                        return new ClosedCode(c1.bytecode.
                                              addMath(Instr.NEG, c1.type),
                                              env, c1.type)/*.addInfo(c1)*/;
                }
                public String toString() { return "-"+c; }
            };
    }

    /** Do-while loop with a specified endlabel.
     * @param label label name to jump to on break/termination of loop
     * @param test  test expression of the loop
     * @param body  body statement of the loop
     */
    public static Code doWhileLabel(final String label, 
                                    final Code test, 
                                    final Code body) {
        return new Code() {
                public String getName() { return "doWhileLabel"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return body.defineClasses(test.defineClasses(e));
                }
                public Environment defineSupers(Environment e) {
                    return body.defineSupers(test.defineSupers(e));
                }
                public Environment defineMembers(Environment e) {
                    return body.defineMembers(test.defineMembers(e));
                }
                public ClosedCode eval(Environment env) {
                    final Instr startlabel = Instr.genNil("Start of do-while");
                    final Instr testlabel = Instr.genNil("Test of do-while");
                    final Instr endlabel = Instr.genNil("End of do-while");
                    final ClosedCode test1 = test.evalWithLabels(env, startlabel, endlabel, endlabel);
                    final Environment bodyenv = env.defineBreakLabel(label, endlabel);
                    final Environment bodyenv1 = bodyenv.defineContinueLabel(label, testlabel);
                    final ClosedCode body1 = body.evalNoUse(bodyenv1);
                    return new ClosedCode(startlabel.
                                          append(body1.bytecode).
                                          append(testlabel).
                                          append(test1.bytecode).
                                          append(endlabel),
                                          env, Type.void_type)
                        /*.addInfo(test1).addInfo(body1)*/;
                }
                public String toString() { 
                    return (label.length()>0?label+":":"")+"do {\n"+
                        body+"\n} while ("+body+");"; }
            };
    }

    /** Do-while loop.
     * @param test test expression of the loop
     * @param body body statement of the loop
     */
    public static Code doWhile(final Code test, 
                               final Code body) {
        return doWhileLabel("", test, body);
    }

    /*
      (** Constructor: Closes another constructor wrt the environment 
      {@type Code -> env -> Code
      *)
      fun close e env =
      fn env' => e env handle e => raise e;
    */

    protected static String normalizeClassNames(String classname, Environment e) {
        try {
            return e.get(classname).getName();
        } catch (ClassNotFoundException err) {
            throw error(err.toString(), e);
        }
    }

    protected static Type normalizeClassNames(Type t, Environment e) {
        if (t.hasClassName()) {
            final String n = t.getClassName();
            final String n1 = normalizeClassNames(n, e);
            if (!n1.equals(n))
                return t.replaceClassName(n1);
            else
                return t;
        } else return t;
    }
    
    protected static MethodType normalizeClassNames(MethodType mt, Environment e) {
        boolean changed = false;
        final Type ret = normalizeClassNames(mt.returntype, e);
        if (ret != mt.returntype) changed = true;
        Type[] newargs = new Type[mt.arguments.length];
        for (int i = 0; i < newargs.length; i++) {
            newargs[i] = normalizeClassNames(mt.arguments[i], e);
            if (newargs[i] != mt.arguments[i])
                changed = true;
        }
        if (changed)
            return new MethodType(ret, newargs);
        return mt;
    }

    public static Code defineVariable(final int flags,
                                      final Type type,
                                      final String name,
                                      final Code init) {
        return new Code() {
                public String getName() { return "defineVariable"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    e = e.defineVar(name, type, flags);
                    if (init != null) return init.defineClasses(e);
                    else return e;
                }
                public Environment defineSupers(Environment e) {
                    if (init != null) return init.defineSupers(e);
                    else return e;
                }
                public Environment defineMembers(Environment e) {
                    if (init != null) return init.defineMembers(e);
                    else return e;
                }
                public ClosedCode eval(Environment env) {
                    if (!flagsLegal(Flags.FINAL, flags))
                        throw error("Illegal flags for variable "+name+": "+flags, env);
                    if (env.varDefined(name)) {
                        throw error("Variable "+name+" already defined", env);
                    }
                    final Type t = normalizeClassNames(type, env);
                    final Environment initenv = env.defineVar(name, t,
                                                              flags, null);
                    if (init != null) {
                        Var v = initenv.getNewestVar();
                        final ClosedCode initval = init.eval(initenv);
                        final ClosedCode initval1 = valueConvert(initval, t);
                        if (initval1.isConstant()) {
                            if ((flags & Flags.FINAL) != 0) {
                                final Environment env1 = 
                                    env.defineVar(name, t, flags, 
                                                  initval1.constantValue());
                                return new ClosedCode(Instr.genNil("Constant final init"),
                                                      env1, Type.void_type);
                            } else {
                                return new ClosedCode(initval1.constantValue().genLoadCode().
                                                      addStore(v.index, t),
                                                      initenv, Type.void_type);
                            }
                        } else {
                            return new ClosedCode(initval1.bytecode.
                                                  addStore(v.index, t),
                                                  initenv, Type.void_type);
                        }
                    } else {
                        return new ClosedCode(Instr.genNil("Vardef w/o init"),
                                              initenv, Type.void_type);
                    }
                }
                public String toString() { 
                    return Flags.toString(flags)+" "+
                        type+" "+name+(init!=null?" = "+init.toString():""); }
            };
    }

    public static Code defineVariable(final int flags,
                                      final Type type,
                                      final Name name,
                                      final Code init) {
        return new Code() {
                public String getName() { return "defineVariable"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    e = name.define(type, flags, e);
                    if (init != null) return init.defineClasses(e);
                    else return e;
                }
                public Environment defineSupers(Environment e) {
                    if (init != null) return init.defineSupers(e);
                    else return e;
                }
                public Environment defineMembers(Environment e) {
                    if (init != null) return init.defineMembers(e);
                    else return e;
                }
                public ClosedCode eval(Environment env) {
                    if (!flagsLegal(Flags.FINAL, flags))
                        throw error("Illegal flags for variable "+name+": "+flags, env);
                    
                    final Type t = normalizeClassNames(type, env);
                    final Environment initenv = env.defineVar(name.getName(), t, flags, null);
                    if (init != null) {
                        Var v = initenv.getNewestVar();
                        final ClosedCode initval = init.eval(initenv);
                        final ClosedCode initval1 = valueConvert(initval, t);
                        if (initval1.isConstant()) {
                            if ((flags & Flags.FINAL) != 0) {
                                final Environment env1 = 
                                    env.defineVar(name.getName(), t, flags, initval1.constantValue());
                                return new ClosedCode(Instr.genNil("Constant final init"),
                                                      env1, Type.void_type);
                            } else {
                                return new ClosedCode(initval1.constantValue().genLoadCode().
                                                      addStore(v.index, t),
                                                      initenv, Type.void_type);
                            }
                        } else {
                            return new ClosedCode(initval1.bytecode.
                                                  addStore(v.index, t),
                                                  initenv, Type.void_type);
                        }
                    } else {
                        return new ClosedCode(Instr.genNil("Vardef w/o init"),
                                              initenv, Type.void_type);
                    }
                }
                public String toString() { 
                    return Flags.toString(flags)+" "+
                        type+" "+name+(init!=null?" = "+init.toString():""); }
            };
    }

    /** Define a new final variable in the current scope.
     * @param name The name of the new variable.
     * @param type The type of the new variable.
     * @param initval Code for the value that this variable should be 
     * initialized to.
     */       
    public static Code defineFinalVariable(final Type type,
                                           final String name, 
                                           final Code initval) {
        return defineVariable(Flags.FINAL, type, name, initval);
    }

    /** Define a new variable in the current scope.
     */
    public static Code defineVariable(final int flags, 
                                      final Type type,
                                      final String name) {
        return defineVariable(flags, type, name, null);
    }

    /** Define a new variable in the current scope.
     */
    public static Code defineVariable(final String name, 
                                      final Type type) {
        return defineVariable(0, type, name, null);
    }

    /*private*/
    public static ClosedCode makeStringValue(ClosedCode c) {
        //        if (c.type.isObjectType("java/lang/String")) return c;
        Type t = c.type;
        if (t == Type.byte_type || t == Type.short_type)
            t = Type.int_type;
        if (t.isReference() && (!t.isObjectType("java/lang/String"))) 
            t = Type.object_type;
        return new ClosedCode(c.bytecode.
                              addInvokevirtual("java/lang/StringBuffer",
                                               "append",
                                               new MethodType(new Type("java/lang/StringBuffer"),
                                                              t)),
                              c.env, Type.string_type);
    }

    public static final int CMP_EQ = 0;
    public static final int CMP_NE = 1;
    public static final int CMP_LT = 2;
    public static final int CMP_GE = 3;
    public static final int CMP_GT = 4;
    public static final int CMP_LE = 5;
    public static final int ADD = 6;
    public static final int SUB = 7;
    public static final int MUL = 8;
    public static final int DIV = 9;
    public static final int REM = 10;
    public static final int NEG = 11;
    public static final int SHL = 12;
    public static final int SHR = 13;
    public static final int USHR = 14;
    public static final int AND = 15;
    public static final int OR = 16;
    public static final int XOR = 17;

    /** Any binary operation 
     * `op' is one of the Constructor.OP constants.
     */
    public static Code binOp(final int op, final Code c1, final Code c2) {
        switch (op) {
        case CMP_EQ: return binCompare(Instr.CMP_EQ, c1, c2);
        case CMP_NE: return binCompare(Instr.CMP_NE, c1, c2);
        case CMP_LT: return binCompare(Instr.CMP_LT, c1, c2);
        case CMP_GE: return binCompare(Instr.CMP_GE, c1, c2);
        case CMP_GT: return binCompare(Instr.CMP_GT, c1, c2);
        case CMP_LE: return binCompare(Instr.CMP_LE, c1, c2);
        case ADD: return mathOp(Instr.ADD, c1, c2);
        case SUB: return mathOp(Instr.SUB, c1, c2);
        case MUL: return mathOp(Instr.MUL, c1, c2);
        case DIV: return mathOp(Instr.DIV, c1, c2);
        case REM: return mathOp(Instr.REM, c1, c2);
        case NEG: return mathOp(Instr.NEG, c1, c2);
        case SHL: return bitwiseOp(Instr.SHL, c1, c2);
        case SHR: return bitwiseOp(Instr.SHR, c1, c2);
        case USHR: return bitwiseOp(Instr.USHR, c1, c2);
        case AND: return bitwiseOp(Instr.AND, c1, c2);
        case OR: return bitwiseOp(Instr.OR, c1, c2);
        case XOR: return bitwiseOp(Instr.XOR, c1, c2);
        default:
            throw error("Illegal binop "+op, null);
        }
    }

    protected static final String[] mathopnames = {
        "==",
        "!=",
        "<",
        ">",
        ">=",
        "<=",
        "+",
        "-",
        "*",
        "/",
        "%",
        "-",
        "<<",
        ">>",
        ">>>",
        "&",
        "|",
        "^",
    };
    
    /** A math (addition, substraction, division, multiplication, etc)
     *  binary operation.
     * @param op  operation to perform (see class Instr)
     * @param c1  first operand
     * @param c2  second operand
     */
    public static Code mathOp(final int op, final Code c1, final Code c2) {
        return new Code() {
                public String getName() { return "mathOp"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return c2.defineClasses(c1.defineClasses(e));
                }
                public Environment defineSupers(Environment e) {
                    return c2.defineSupers(c1.defineSupers(e));
                }
                public Environment defineMembers(Environment e) {
                    return c2.defineMembers(c1.defineMembers(e));
                }
                public ClosedCode evalInAdd(Environment env) {
                    final ClosedCode cp1;
                    if (op == Instr.ADD) {
                        cp1 = c1.evalInAdd(env);
                    } else {
                        cp1 = c1.eval(env);
                    }
                    final ClosedCode cp2 = c2.eval(env);
                    if (cp1.isConstant() && cp2.isConstant()) {
                        final ConstVal c = cp1.constantValue().mathBinOp(op, cp2.constantValue());
                        return new ClosedCode(c.genLoadCode(),
                                              env, c.getType(), c)
                            /*.addInfo(cp1).addInfo(cp2)*/;
                    } else if (op == Instr.ADD) {
                        if (cp1.type.isObjectType("java/lang/StringBuffer")) {
                            final ClosedCode str2 = makeStringValue(cp2);
                            return new ClosedCode(cp1.bytecode.
                                                  append(str2.bytecode),
                                                  env, cp1.type);
                        } else if (cp1.type.isObjectType("java/lang/String") ||
                                   cp2.type.isObjectType("java/lang/String")) {
                            final ClosedCode str1 = makeStringValue(cp1);
                            final ClosedCode str2 = makeStringValue(cp2);
                            return new ClosedCode(Instr.genNew("java/lang/StringBuffer").
                                                  addDup(1,0).
                                                  addInvokenonvirtual("java/lang/StringBuffer", "<init>", new MethodType(Type.void_type)).
                                                  append(str1.bytecode).
                                                  append(str2.bytecode),
                                                  env, Type.stringbuffer_type);
                        } 
                    } 
                    final Type t = cp1.type.lub(cp2.type);
                    
                    return new ClosedCode(cp1.bytecode.
                                          addConvert(cp1.type, t).
                                          append(cp2.bytecode).
                                          addConvert(cp2.type, t).
                                          addMath(op, t),
                                          env, t);
                }
                public ClosedCode eval(Environment env) {
                    final ClosedCode cp = evalInAdd(env);
                    if (op == Instr.ADD &&
                        cp.type.isObjectType("java/lang/StringBuffer")) {
                        return new ClosedCode(cp.bytecode.
                                              addInvokevirtual("java/lang/StringBuffer", "toString", new MethodType(Type.string_type)),
                                              env, Type.string_type);
                    }
                    else return cp;
                }
                public String toString() {
                    return "("+c1.toString()+mathopnames[op]+c2.toString()+")";
                }
            };
    }

    /** A math (negation) unary operation.
     * @param op  operation to perform (see class Instr)
     * @param c   operand
     */
    public static Code mathOp(final int op, final Code c) {
        return new Code() {
                public String getName() { return "mathOp"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return c.defineClasses(e);
                }
                public Environment defineSupers(Environment e) {
                    return c.defineSupers(e);
                }
                public Environment defineMembers(Environment e) {
                    return c.defineMembers(e);
                }
                public ClosedCode eval(Environment env) {
                    final ClosedCode cp = c.eval(env);
                    if (cp.isConstant()) {
                        final ConstVal v = cp.constantValue().bitwiseNegOp();
                        if (v.isInt())
                            return new ClosedCode(v.genLoadCode(),
                                                  env, Type.int_type, v)
                                /*.addInfo(cp)*/;
                        else
                            return new ClosedCode(v.genLoadCode(),
                                                  env, Type.long_type, v)
                                /*.addInfo(cp)*/;
                    }
                    else
                        return new ClosedCode(cp.bytecode.
                                              addMath(op, cp.type),
                                              env, cp.type)
                            /*.addInfo(cp)*/;
                }
            };
    }

    protected static final String[] bitwiseopnames = {
        "<<",
        ">>",
        ">>>",
        "&",
        "|",
        "^",
    };

    /** A bitwise (and, or, xor, shl, shr, shrr) binary operation.
     * @param op  operation to perform (see class Instr)
     * @param c1  first operand
     * @param c2  second operand
     */
    public static Code bitwiseOp(final int op, final Code c1, final Code c2) {
        return new Code() {
                public String getName() { return "bitwiseOp"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return c2.defineClasses(c1.defineClasses(e));
                }
                public Environment defineSupers(Environment e) {
                    return c2.defineSupers(c1.defineSupers(e));
                }
                public Environment defineMembers(Environment e) {
                    return c2.defineMembers(c1.defineMembers(e));
                }
                public ClosedCode eval(Environment env) {
                    ClosedCode cp1 = c1.eval(env);
                    ClosedCode cp2 = c2.eval(env);

                    ConstVal cv = null;
                    final Type t = cp1.type.lub(cp2.type);
                    Instr bytecode;
                    if (cp1.isConstant() && cp2.isConstant()) {
                        cv = cp1.constantValue().bitwiseBinOp(op, cp2.constantValue());
                        return new ClosedCode(cv.genLoadCode(), env, t, cv);
                    } else if (op <= Instr.USHR) {
                        bytecode = cp1.bytecode.
                            addConvert(cp1.type, cp1.type).
                            append(cp2.bytecode).
                            addConvert(cp2.type, Type.int_type).
                            addBitwise(op, t);
                    } else {                    
                        bytecode = cp1.bytecode.
                            addConvert(cp1.type, t).
                            append(cp2.bytecode).
                            addConvert(cp2.type, t).
                            addBitwise(op, t);
                    }
                    return new ClosedCode(bytecode, env, t, cv)
                        /*.addInfo(cp1).addInfo(cp2)*/;
                }
                public String toString() {
                    return "("+c1.toString()+bitwiseopnames[op-Instr.SHL]+c2.toString()+")";
                }
            };
    }

    /** A bitwise (negation) unary operation.
     * @param op  operation to perform (see class Instr)
     * @param c   operand
     */
    public static Code bitwiseOp(final int op, final Code c) {
        return new Code() {
                public String getName() { return "bitwiseOp"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return c.defineClasses(e);
                }
                public Environment defineSupers(Environment e) {
                    return c.defineSupers(e);
                }
                public Environment defineMembers(Environment e) {
                    return c.defineMembers(e);
                }
                public ClosedCode eval(Environment env) {
                    final ClosedCode cp = c.eval(env);

                    return new ClosedCode(cp.bytecode.
                                          addBitwise(op, cp.type),
                                          env, cp.type)/*.addInfo(cp)*/;
                }
                public String toString() { return "~"+c; }
            };
    }

    /** Boolean (logical) not.
     * @param c Code to boolean not.
     */
    public static Code booleanNot(final Code c) {
        return new Code() {
                public String getName() { return "booleanNot"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return c.defineClasses(e);
                }
                public Environment defineSupers(Environment e) {
                    return c.defineSupers(e);
                }
                public Environment defineMembers(Environment e) {
                    return c.defineMembers(e);
                }
                public ClosedCode eval(Environment env) {
                    final Instr falselabel = Instr.genNil("False of booleanNot");
                    final Instr endlabel = Instr.genNil("End of booleanNot");
                    final ClosedCode c1 = c.eval(env);
                    if (c1.isConstant()) {
                        final ConstVal v = c1.constantValue().logicalNegOp();
                        return new ClosedCode(v.genLoadCode(),
                                              env, Type.boolean_type, v)
                            /*.addInfo(c1)*/;
                    }

                    return new ClosedCode(c1.bytecode.
                                          addIf(Instr.CMP_EQ, falselabel).
                                          addConst(0).
                                          addGoto(endlabel).
                                          append(falselabel).
                                          addConst(1).
                                          append(endlabel),
                                          env, Type.boolean_type)
                        /*.addInfo(c1)*/;
                }
                public ClosedCode evalWithLabels(Environment env,
                                                 Instr truelabel,
                                                 Instr falselabel,
                                                 Instr nextlabel) {
                    final ClosedCode cc =
                        c.evalWithLabels(env, falselabel, truelabel,
                                         nextlabel);
                    return new ClosedCode(cc.bytecode, env, 
                                          Type.boolean_type,
                                          (cc.isConstant()?
                                           cc.constantValue().logicalNegOp():
                                           null))/*.addInfo(cc)*/;
                }
                public String toString() { return "!"+c; }
            };
    }

    protected static ClosedCode intcomp(final int compare_kind,
                                      final ClosedCode c1,
                                      final ClosedCode c2,
                                      Environment env) {
        final Instr falselabel = Instr.genNil("Int compare false branch"),
            endlabel = Instr.genNil("End of int compare");
        if ((c1.isConstant() && c1.constantValue().intVal() == 0) ||
            (c2.isConstant() && c2.constantValue().intVal() == 0)) {
            final ClosedCode c = (c1.isConstant()?c2:c1);
            final int cmp = (c1.isConstant()?reverseCmp(compare_kind):compare_kind);
            return new ClosedCode(
                                  c.bytecode.
                                  addIf(cmp, falselabel).
                                  addIconst(0).
                                  addGoto(endlabel).
                                  append(falselabel).
                                  addIconst(1).
                                  append(endlabel),
                                  Type.boolean_type,
                                  env)/*.addInfo(c)*/;

        }
        return new ClosedCode(
                              c1.bytecode.
                              append(c2.bytecode).
                              addIfIcomp(compare_kind, falselabel).
                              addIconst(0).
                              addGoto(endlabel).
                              append(falselabel).
                              addIconst(1).
                              append(endlabel),
                              Type.boolean_type,
                              env)/*.addInfo(c1).addInfo(c2)*/;
    }

    protected static ClosedCode objcomp(final int compare_kind,
                                      final ClosedCode c1,
                                      final ClosedCode c2,
                                      Environment env) {
        final Instr falselabel = Instr.genNil("Int compare false branch"),
            endlabel = Instr.genNil("End of int compare");
        if ((c1.isConstant() && c1.constantValue().isNull()) ||
            (c2.isConstant() && c2.constantValue().isNull())) {
            // We know they're not both constant, or we wouldn't be here
            final ClosedCode c = (c1.isConstant()?c2:c1);
            return new ClosedCode(
                                  c.bytecode.
                                  addIfNull(compare_kind, falselabel).
                                  addIconst(0).
                                  addGoto(endlabel).
                                  append(falselabel).
                                  addIconst(1).
                                  append(endlabel),
                                  Type.boolean_type,
                                  env)/*.addInfo(c)*/;
        }
        return new ClosedCode(
                              c1.bytecode.
                              append(c2.bytecode).
                              addIfAcomp(compare_kind, falselabel).
                              addIconst(0).
                              addGoto(endlabel).
                              append(falselabel).
                              addIconst(1).
                              append(endlabel),
                              Type.boolean_type,
                              env)/*.addInfo(c1).addInfo(c2)*/;
    }

    public static Instr genNormalizeBool (int compare_kind) {
        // Normalize to (0,non-0) after a three-valued comparison
        // Input is -1, 0, or 1
        //  If compare_kind is NE, then nothing
        //  If compare_kind is EQ, then logically negate
        //  If compare_kind is GE, {0,+} -> 1, {-} -> 0
        //  If compare_kind is GT, {+} -> 1, {0,-} -> 0
        //  If compare_kind is LE, {0,-} -> 1, {+} -> 0
        //  If compare_kind is LT, {-} -> 1, {0,+} -> 0

        if (compare_kind == Instr.CMP_NE) 
            return Instr.genNil("NE compare happens directly");
        if (compare_kind == Instr.CMP_LT)
            return Instr.genConst(2).addBitwise(Instr.AND, Type.int_type);
        if (compare_kind == Instr.CMP_LE)
            return Instr.genConst(1).addBitwise(Instr.XOR, Type.int_type);
        if (compare_kind == Instr.CMP_GE)
            return Instr.genConst(-1).addBitwise(Instr.XOR, Type.int_type);
        final Instr falselabel = Instr.genNil("Int compare false branch"),
            endlabel = Instr.genNil("End of int compare");
        return Instr.genIf(compare_kind, falselabel).
            addIconst(0).
            addGoto(endlabel).
            append(falselabel).
            addIconst(1).
            append(endlabel);
    }

    protected static final String[] compareopnames = {
        "==",
        "!=",
        "<",
        ">=",
        ">",
        "<="
    };

    public static Code binCompare(final int compare_kind,
                                  final Code c1,
                                  final Code c2) {
        return new Code() {
                public String getName() { return "binCompare"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return c2.defineClasses(c1.defineClasses(e));
                }
                public Environment defineSupers(Environment e) {
                    return c2.defineSupers(c1.defineSupers(e));
                }
                public Environment defineMembers(Environment e) {
                    return c2.defineMembers(c1.defineMembers(e));
                }
                public ClosedCode eval(Environment env) {
                    final ClosedCode c1c = c1.eval(env);
                    final ClosedCode c2c = c2.eval(env);
                    if (c1c.isConstant() && c2c.isConstant()) {
                        final ConstVal c = c1c.constantValue().compareBinOp(compare_kind, c2c.constantValue());
                        return new ClosedCode(c.genLoadCode(),
                                              env, c.getType(), c)
                            /*.addInfo(c1c).addInfo(c2c)*/;

                    }
                    final Type t = c1c.type.lub(c2c.type);
                    if (t.bytecodeKind() == Type.INT)
                        return intcomp(compare_kind, c1c, c2c, env);
                    if (t.bytecodeKind() == Type.OBJECT)
                        return objcomp(compare_kind, c1c, c2c, env);
                    else return
                             new ClosedCode(
                                            c1c.bytecode.
                                            addConvert(c1c.type, t).
                                            append(c2c.bytecode).
                                            addConvert(c2c.type, t).
                                            addCmp(t, Instr.L).
                                            append(genNormalizeBool(compare_kind)),
                                            env,
                                            Type.boolean_type)
                             /*.addInfo(c1c).addInfo(c2c)*/;
                }
                public ClosedCode evalWithLabels(Environment env,
                                                 Instr truelabel,
                                                 Instr falselabel,
                                                 Instr nextlabel) {
                    final ClosedCode c1c = c1.eval(env);
                    final ClosedCode c2c = c2.eval(env);
                    if (c1c.isConstant() && c2c.isConstant()) {
                        final ConstVal c = c1c.constantValue().compareBinOp(compare_kind, c2c.constantValue());
                        return new ClosedCode(c.genLoadCode(),
                                              env, c.getType(), c)
                            /*.addInfo(c1c).addInfo(c2c)*/;

                    }
                    final Type t = c1c.type.lub(c2c.type);
                    Instr retcode = c1c.bytecode;
                    Instr jumplabel = truelabel;
                    int cmp = compare_kind;
                    if (truelabel == nextlabel) {
                        jumplabel = falselabel;
                        cmp = invertCmp(cmp);
                    }
                    if (t.bytecodeKind() == Type.INT) {
                        if (c1c.isConstant() &&
                            c1c.constantValue().bytecodeIntVal() == 0) {
                            retcode = c2c.bytecode;
                            retcode.addIf(reverseCmp(cmp), jumplabel);
                        } else if (c2c.isConstant() &&
                                   c2c.constantValue().bytecodeIntVal() == 0) {
                            retcode.addIf(cmp, jumplabel);
                        } else {
                            retcode.
                                append(c2c.bytecode).
                                addIfIcomp(cmp, jumplabel);
                        }
                    }
                    else if (t.bytecodeKind() == Type.OBJECT) {
                        if (c1c.isConstant() &&
                            c1c.constantValue().isNull()) {
                            retcode = c2c.bytecode;
                            retcode.addIfNull(cmp, jumplabel);
                        } else if (c2c.isConstant() &&
                                   c2c.constantValue().isNull()) {
                            retcode.addIfNull(cmp, jumplabel);
                        } else {
                            retcode.
                                append(c2c.bytecode).
                                addIfAcomp(cmp, jumplabel);
                        }
                    }
                    else
                        retcode.
                            addConvert(c1c.type, t).
                            append(c2c.bytecode).
                            addConvert(c2c.type, t).
                            addCmp(t, Instr.L).
                            addIf(cmp, jumplabel);
                    if (truelabel != nextlabel && falselabel != nextlabel) {
                        // A double jump.  If jumped to true.
                        retcode.addGoto(falselabel);
                    }
                    final ClosedCode retval = new ClosedCode(retcode, env, Type.void_type);
                    return retval/*.addInfo(c1c).addInfo(c2c)*/;
                }
                public String toString() {
                    return "("+c1.toString()+compareopnames[compare_kind]+c2.toString()+")";
                }
            };
    }

    public static Code instanceOf(final Code value, final Type type) {
        return new Code() {
                public String getName() { return "instanceOf"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return value.defineClasses(e);
                }
                public Environment defineSupers(Environment e) {
                    return value.defineSupers(e);
                }
                public Environment defineMembers(Environment e) {
                    return value.defineMembers(e);
                }
                public ClosedCode eval(Environment env) {
                    final ClosedCode value1 = value.eval(env);
                    if (!value1.type.isReference()) 
                        throw error("Doing instanceof on non-reference value "+value1.type, env);
                    final Type type1 = normalizeClassNames(type, env);
                    return new ClosedCode(value1.bytecode
                                          .addInstanceof(type1),
                                          env, Type.boolean_type)
                        /*.addInfo(value1)*/;
                }
                public String toString() {
                    return "("+value+" instanceof "+type+")";
                }
            };
    }

    public static Code cast(final Code value, final Type type) {
        return new Code() {
                public String getName() { return "cast"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return value.defineClasses(e);
                }
                public Environment defineSupers(Environment e) {
                    return value.defineSupers(e);
                }
                public Environment defineMembers(Environment e) {
                    return value.defineMembers(e);
                }
                public ClosedCode eval(Environment env) {
                    final ClosedCode value1 = value.eval(env);
                    // XXX Check here that cast is legal
                    final Type type1 = normalizeClassNames(type, env);
                    if (value1.isConstant() &&
                        (type1.isPrimitive() || 
                         type1.isObjectType("java/lang/String"))) {
                        final ConstVal cv = ConstVal.make(value1.constantValue(), type1);
                        return new ClosedCode(cv.genLoadCode(),
                                              env, type1, cv)
                            /*.addInfo(value1)*/;
                    } else {
                        if (type1.isReference()) {
                            return new ClosedCode(value1.bytecode.
                                                  addCheckcast(type1.castName()),
                                                  env, type1)
                                /*.addInfo(value1)*/;
                        } else {
                            return new ClosedCode(value1.bytecode.
                                                  addConvert(value1.type, type1), 
                                                  env, type1)
                                /*.addInfo(value1)*/;
                        }
                    }
                }
                public String toString() {
                    return "(("+type+")"+value+")";
                }
            };
    }

    public static Code ifThenElse(final Code c1,
                                  final Code c2,
                                  final Code c3) {
        return new Code() {
                public String getName() { return "ifThenElse"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return c3.defineClasses
                        (c2.defineClasses
                         (c1.defineClasses(e)));
                }
                public Environment defineSupers(Environment e) {
                    return c3.defineSupers
                        (c2.defineSupers
                         (c1.defineSupers(e)));
                }
                public Environment defineMembers(Environment e) {
                    return c3.defineMembers
                        (c2.defineMembers
                         (c1.defineMembers(e)));
                }
                public ClosedCode eval(Environment env) {
                    final Instr truelabel = Instr.genNil("True label");
                    final Instr falselabel = Instr.genNil("False label");
                    final ClosedCode c1c = c1.evalWithLabels(env, truelabel,
                                                             falselabel,
                                                             truelabel);
                    if (c1c.isConstant()) {
                        if (c1c.constantValue().booleanVal()) {
                            return c2.evalNoUse(env);
                        } else {
                            return c3.evalNoUse(env);
                        }
                    } else {
                        final ClosedCode c2c = c2.evalNoUse(env);
                        final ClosedCode c3c = c3.evalNoUse(env);
                        final Instr endlabel = Instr.genNil("End of if");
                        Type finaltype = c2c.type.bottomLub(c3c.type);
                        if (finaltype != Type.bottom)
                            finaltype = Type.void_type;
                        final Instr retcode = c1c.bytecode.
                            append(truelabel).
                            append(c2c.bytecode).
                            append(c2c.type != Type.bottom?Instr.genGoto(endlabel):
                                   Instr.genNil("Unreachable end of true branch")).
                            append(falselabel).
                            append(c3c.bytecode).
                            append(endlabel);
                        return new ClosedCode(retcode, env, finaltype)
                            /*.addInfo(c1c).addInfo(c2c).addInfo(c3c)*/;
                    }
                }
                public String toString() {
                    return "if ("+c1+") {\n"+c2+";\n} else {\n"+c3+";\n}";
                }
            };
    }


    public static Code ifThen(final Code c1,
                              final Code c2) {
        return new Code() {
                public String getName() { return "ifThen"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return c2.defineClasses(c1.defineClasses(e));
                }
                public Environment defineSupers(Environment e) {
                    return c2.defineSupers(c1.defineSupers(e));
                }
                public Environment defineMembers(Environment e) {
                    return c2.defineMembers(c1.defineMembers(e));
                }
                public ClosedCode eval(Environment env) {
                    final Instr truelabel = Instr.genNil("True label");
                    final Instr endlabel = Instr.genNil("End of if");
                    final ClosedCode c1c = c1.evalWithLabels(env,
                                                             truelabel,
                                                             endlabel,
                                                             truelabel);
                    if (c1c.isConstant()) {
                        if (c1c.constantValue().booleanVal()) {
                            return c2.evalNoUse(env);
                        } else {
                            return new ClosedCode(Instr.genNil("Empty constant if"),
                                                  env, Type.void_type)
                                /*.addInfo(c2c)*/;
                        }
                    }
                    final ClosedCode c2c = c2.evalNoUse(env);
                    return new ClosedCode(c1c.bytecode.
                                          append(truelabel).
                                          append(c2c.bytecode).
                                          append(endlabel),
                                          env,
                                          Type.void_type)
                        /*.addInfo(c1c).addInfo(c2c)*/;
                }
                public String toString() {
                    return "if ("+c1+") {\n"+c2+";\n}";
                }
            };
    }

    public static Code trinaryConditional(final Code c1,
                                          final Code c2,
                                          final Code c3) {
        return new Code() {
                public String getName() { return "trinaryConditional"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return c3.defineClasses(c2.defineClasses(c1.defineClasses(e)));
                }
                public Environment defineSupers(Environment e) {
                    return c3.defineSupers(c2.defineSupers(c1.defineSupers(e)));
                }
                public Environment defineMembers(Environment e) {
                    return c3.defineMembers(c2.defineMembers(c1.defineMembers(e)));
                }
                public ClosedCode eval(Environment env) {
                    final ClosedCode c2c = c2.eval(env),
                        c3c = c3.eval(env);
                    Type t = c2c.type.commonAncestor(c3c.type);
                    ClosedCode c2c1 = valueConvert(c2c, t);
                    ClosedCode c3c1 = valueConvert(c3c, t);
                    final ClosedCode c1c = c1.evalWithLabels(env, c2c1.bytecode,
                                                             c3c1.bytecode,
                                                             c2c1.bytecode);
                    // XXX Check here if the result values are compatible
                    if (c1c.isConstant()) {
                        if (c1c.constantValue().booleanVal()) {
                            return c2c;
                        } else {
                            return c3c;
                        }
                    }
                    final Instr endlabel = Instr.genNil("End of ternary");
                    return new ClosedCode (c1c.bytecode.
                                           append(c2c1.bytecode).
                                           addGoto(endlabel).
                                           append(c3c1.bytecode).
                                           append(endlabel),
                                           env,
                                           t)
                        /*.addInfo(c1c).addInfo(c2c).addInfo(c3c)*/;
                }
                public String toString() {
                    return "("+c1+"?"+c2+":"+c3+")";
                }
            };
    }

    public static Code shortcutOr(final Code c1,
                                  final Code c2) {
        return new Code() {
                public String getName() { return "shortcutOr"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return c2.defineClasses(c1.defineClasses(e));
                }
                public Environment defineSupers(Environment e) {
                    return c2.defineSupers(c1.defineSupers(e));
                }
                public Environment defineMembers(Environment e) {
                    return c2.defineMembers(c1.defineMembers(e));
                }
                public ClosedCode eval(Environment env) {
                    final ClosedCode c1c = c1.eval(env),
                        c2c = c2.eval(env);
                    if (c1c.type.kind != Type.BOOLEAN) 
                        throw error("Or operands must be of boolean type, not "+c1c.type, env);
                    if (c2c.type.kind != Type.BOOLEAN) 
                        throw error("Or operands must be of boolean type, not "+c2c.type, env);
                    if (c1c.isConstant()) {
                        if (c1c.constantValue().booleanVal()) {
                            return new ClosedCode(Instr.genConst(1),
                                                  env, Type.boolean_type,
                                                  ConstVal.make(true))
                                /*.addInfo(c1c).addInfo(c2c)*/;
                        } else {
                            return c2c;
                        }
                    }
                    final Instr endlabel = Instr.genNil("End of shortcut Or");
                    return new ClosedCode (c1c.bytecode.
                                           addDup(1,0).
                                           addIf(Instr.CMP_NE, endlabel).
                                           addPop(c1c.type).
                                           append(c2c.bytecode).
                                           append(endlabel),
                                           env,
                                           Type.boolean_type)
                        /*.addInfo(c1c).addInfo(c2c)*/;
                }
                public ClosedCode evalWithLabels(Environment env,
                                                 Instr truelabel,
                                                 Instr falselabel,
                                                 Instr nextlabel) {
                    final Instr label2 = Instr.genNil("Shortcut Or middle");
                    final ClosedCode c1c = c1.evalWithLabels(env,
                                                             truelabel,
                                                             label2,
                                                             label2);
                    final ClosedCode c2c = c2.evalWithLabels(env,
                                                             truelabel,
                                                             falselabel,
                                                             nextlabel);
                    if (c1c.isConstant()) {
                        if (c1c.constantValue().booleanVal()) {
                            // This introduces a spurios goto.  Too bad.
                            return c1c;
                        } else {
                            return c2c;
                        }
                    }
                    // c2c being constant doesn't matter -- it will contain
                    // the appropriate jump anyway, and
                    // we must evaluate c1c.
                    return new ClosedCode(c1c.bytecode.
                                          append(label2).
                                          append(c2c.bytecode),
                                          env, Type.void_type);

                }
                public String toString() {
                    return "("+c1+"||\n"+c2+")";
                }
            };
    }

    public static Code shortcutAnd(final Code c1,
                                   final Code c2) {
        return new Code() {
                public String getName() { return "shortcutAnd"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return c2.defineClasses(c1.defineClasses(e));
                }
                public Environment defineSupers(Environment e) {
                    return c2.defineSupers(c1.defineSupers(e));
                }
                public Environment defineMembers(Environment e) {
                    return c2.defineMembers(c1.defineMembers(e));
                }
                public ClosedCode eval(Environment env) {
                    final ClosedCode c1c = c1.eval(env),
                        c2c = c2.eval(env);
                    if (c1c.type.kind != Type.BOOLEAN) 
                        throw error("And operands must be of boolean type, not "+c1c.type, env);
                    if (c2c.type.kind != Type.BOOLEAN) 
                        throw error("And operands must be of boolean type, not "+c2c.type, env);
                    if (c1c.isConstant()) {
                        if (c1c.constantValue().booleanVal()) {
                            return c2c;
                        } else {
                            return new ClosedCode(Instr.genConst(0),
                                                  env, Type.boolean_type,
                                                  ConstVal.make(false))
                                /*.addInfo(c1c).addInfo(c2c)*/;
                        }
                    }
                    final Instr endlabel = Instr.genNil("End of shortcut And");
                    return new ClosedCode (
                                           c1c.bytecode.
                                           addDup(1,0).
                                           addIf(Instr.CMP_EQ, endlabel).
                                           addPop(c1c.type).
                                           append(c2c.bytecode).
                                           append(endlabel),
                                           env,
                                           Type.boolean_type)
                        /*.addInfo(c1c).addInfo(c2c)*/;
                }
                public ClosedCode evalWithLabels(Environment env,
                                                 Instr truelabel,
                                                 Instr falselabel,
                                                 Instr nextlabel) {
                    final Instr label2 = Instr.genNil("Shortcut And middle");
                    final ClosedCode c1c = c1.evalWithLabels(env,
                                                             label2,
                                                             falselabel,
                                                             label2);
                    final ClosedCode c2c = c2.evalWithLabels(env,
                                                             truelabel,
                                                             falselabel,
                                                             nextlabel);
                    if (c1c.isConstant()) {
                        if (!c1c.constantValue().booleanVal()) {
                            // This introduces a spurios goto.  Too bad.
                            return c1c;
                        } else {
                            return c2c;
                        }
                    }
                    // c2c being constant doesn't matter -- it will contain
                    // the appropriate jump anyway, and we must evaluate c1c.
                    return new ClosedCode(c1c.bytecode.
                                          append(label2).
                                          append(c2c.bytecode),
                                          env, Type.void_type);

                }
                public String toString() {
                    return "("+c1+"&&\n"+c2+")";
                }
            };
    }

    public static Code whileLabel(final String label,
                                  final Code test,
                                  final Code body) {
        return new Code() {
                public String getName() { return "whileLabel"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return body.defineClasses(test.defineClasses(e));
                }
                public Environment defineSupers(Environment e) {
                    return body.defineSupers(test.defineSupers(e));
                }
                public Environment defineMembers(Environment e) {
                    return body.defineMembers(test.defineMembers(e));
                }
                public ClosedCode eval(Environment env) {
                    final Instr startlabel = Instr.genNil("Start of while");
                    final Instr testlabel = Instr.genNil("Test of while");
                    final Instr endlabel = Instr.genNil("End of while");
                    final ClosedCode test1 = test.evalWithLabels(env,
                                                                 startlabel,
                                                                 endlabel,
                                                                 endlabel);
                    final Environment bodyenv = env.defineBreakLabel(label, endlabel);
                    final Environment bodyenv1 = bodyenv.defineContinueLabel(label, testlabel);
                    final ClosedCode body1 = body.evalNoUse(bodyenv1);
                    return new ClosedCode(
                                          Instr.genGoto(testlabel).
                                          append(startlabel).
                                          append(body1.bytecode).
                                          append(testlabel).
                                          append(test1.bytecode).
                                          append(endlabel),
                                          env, Type.void_type)
                        /*.addInfo(test1).addInfo(body1)*/;
                }
                public String toString() {
                    return (label.length()>0?label+": ":"")+
                        "while ("+test+") {\n"+body+"}";
                }
            };
    }

    public static Code whileDo (final Code test, 
                                final Code body) {
        return whileLabel("", test, body);
    }

    public static Code forLabel(final String label,
                                final Code h1,
                                final Code h2,
                                final Code h3,
                                final Code c) {
        return new Code() {
                public String getName() { return "forLabel"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return c.defineClasses(h3.defineClasses(h2.defineClasses(h1.defineClasses(e))));
                }
                public Environment defineSupers(Environment e) {
                    return c.defineSupers(h3.defineSupers(h2.defineSupers(h1.defineSupers(e))));
                }
                public Environment defineMembers(Environment e) {
                    return c.defineMembers(h3.defineMembers(h2.defineMembers(h1.defineMembers(e))));
                }
                public ClosedCode eval(Environment env) {
                    final Instr startlabel = Instr.genNil("Start of for");
                    final Instr iterlabel = Instr.genNil("Iter of for");
                    final Instr testlabel = Instr.genNil("Test of for");
                    final Instr endlabel = Instr.genNil("End of for");
                    final ClosedCode h1c = h1.evalNoUse(env);
                    final Environment bodyenv = h1c.env.defineBreakLabel(label, endlabel);
                    final Environment bodyenv1 = bodyenv.defineContinueLabel(label, iterlabel);

                    final ClosedCode c1 = c.evalNoUse(bodyenv1);
                    final ClosedCode h2c = h2.evalWithLabels(h1c.env,
                                                             c1.bytecode,
                                                             endlabel,
                                                             endlabel);
                    final ClosedCode h3c = h3.evalNoUse(h1c.env);
                    return new ClosedCode(
                                          h1c.bytecode.
                                          addGoto(testlabel).
                                          append(startlabel).
                                          append(c1.bytecode).
                                          append(iterlabel).
                                          append(h3c.bytecode).
                                          append(testlabel).
                                          append(h2c.bytecode).
                                          append(endlabel),
                                          env,
                                          Type.void_type)
                        /*.addInfo(h1c).addInfo(h2c).addInfo(h3c).addInfo(c1)*/;
                }
                public String toString() {
                    return (label.length()>0?label+": ":"")+
                        "for ("+h1+"; "+h2+"; "+h3+") {\n"+c+"}";
                }
            };
    }

    public static Code forLoop(final Code h1,
                               final Code h2,
                               final Code h3,
                               final Code c) {
        return forLabel("", h1, h2, h3, c);
    }


    protected static Instr makeFinallyJumps(MonoList handlers) {
        final Instr jumps = Instr.genNil("Start of finally handler jumps");
        for (Iterator i = handlers.iterator(); i.hasNext(); ) {
            final Instr ins = (Instr)i.next();
            jumps.addJsr(ins);
        }
        return jumps;
    }

    protected static ClosedCode makeLoopControl(MonoList labels, Environment env) {
        if (labels.size() > 1) {
            final Instr jumps = makeFinallyJumps(labels.subList(0, labels.size()-1));
            return new ClosedCode(jumps.addGoto((Instr)labels.get(labels.size()-1)),
                                  env, Type.bottom);
        } else {
            return new ClosedCode(Instr.genGoto((Instr)labels.get(0)),
                                  env, Type.bottom);
        }
    }

    public static Code breakStatementWithLabel(final String label) {
        return new Code() {
                public String getName() { return "breakStatementWithLabel"; }
                public ClosedCode eval(Environment env) {
                    try {
                        final MonoList labels = env.getBreakLabel(label);
                        return makeLoopControl(labels, env);
                    } catch (LabelNotFoundException e) {
                        throw error(e.toString(), env);
                    }
                }
                public String toString() {
                    return "break"+(label.length()>0?" "+label:"");
                }
            };
    }


    public static Code breakStatement() {
        return breakStatementWithLabel("");
    }

    public static Code continueStatementWithLabel(final String label) {
        return new Code() {
                public String getName() { return "continueStatementWithLabel"; }
                public ClosedCode eval(Environment env) {
                    try {
                        final MonoList labels = env.getContinueLabel(label);
                        return makeLoopControl(labels, env);
                    } catch (LabelNotFoundException e) {
                        throw error(e.toString(), env);
                    }
                }
                public String toString() {
                    return "continue"+(label.length()>0?" "+label:"");
                }
            };
    }

    public static Code continueStatement() {
        return continueStatementWithLabel("");
    }

    public static Code blockLabel(final String label,
                                  final Code c) {
        return new Code() {
                public String getName() { return "blockLabel"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return c.defineClasses(e);
                }
                public Environment defineSupers(Environment e) {
                    return c.defineSupers(e);
                }
                public Environment defineMembers(Environment e) {
                    return c.defineMembers(e);
                }
                public ClosedCode eval(Environment env) {
                    final Instr endlabel = Instr.genNil("End of block "+label);
                    final Environment bodyenv = env.defineBreakLabel(label, endlabel);
                    final ClosedCode cc = c.evalNoUse(bodyenv);
                    return new ClosedCode(cc.bytecode.
                                          append(endlabel),
                                          env,
                                          cc.type)/*.addInfo(cc)*/;
                }
                public String toString () {
                    return label+": {\n"+c+"\n}";
                }
            };
    }

    public static Code synchronizedBlock(final Code monitor,
                                         final Code body) {
        return new Code() {
                public String getName() { return "synchronizedBlock"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return body.defineClasses(monitor.defineClasses(e));
                }
                public Environment defineSupers(Environment e) {
                    return body.defineSupers(monitor.defineSupers(e));
                }
                public Environment defineMembers(Environment e) {
                    return body.defineMembers(monitor.defineMembers(e));
                }
                public ClosedCode eval(Environment env) {
                    final ClosedCode mon = monitor.eval(env);
                    if (!mon.type.isReference()) 
                        throw error("Monitor type must be reference, not "+mon.type, env);
                    final Environment bodyenv = env.generateVar("monitor", new Type("java/lang/Object"));
                    final Var synchvar = bodyenv.getNewestVar();
                    final ClosedCode bod = body.evalNoUse(bodyenv);
                    return new ClosedCode(
                                          mon.bytecode.
                                          addStore(synchvar.index, synchvar.type).
                                          addLoad(synchvar.index, synchvar.type).
                                          addMonitorenter().
                                          append(bod.bytecode).
                                          addLoad(synchvar.index, synchvar.type).
                                          addMonitorexit(),
                                          env,
                                          Type.void_type)/*.addInfo(mon).addInfo(bod)*/;
                }
                public String toString () {
                    return "synchronized ("+monitor+") {\n"+body+"\n}";
                }
            };
    }

    protected static ClosedCode addCases(SwitchCase cases, SortedMap casemap,
                                       Environment env) {
        final ClosedCode code = cases.code.evalNoUse(env);
        env = code.env;
        for (Iterator i = cases.caselist.iterator(); i.hasNext(); ) {
            final Code c = (Code)i.next();
            final ClosedCode c1 = c.eval(env);
            if (!c1.type.isInteger())
                throw error("Switch case must be integral, not "+c1.type, env);
            env = c1.env;
            if (c1.isConstant() && c1.constantValue().isInt()) {
                casemap.put(new Integer(c1.constantValue().intVal()), code.bytecode);
                //              code/*.addInfo(c1)*/;
            } else {
                throw error("Non-constant or non-int switch case: "+c, env);
            }
        }
        return code;
    }

    /** Generate a switch statement
     * @param label The label of the whole statement
     * @param switchee The expression being switched on
     * @param cases The cases, a list of SwitchCase objects
     * @param default Code for the default case
     *
     * Known bugs:  The default statement is always placed at the end.
     */
    public static Code switchLabel(final String label, final Code switchee,
                                   final MonoList cases, final Code defaultcode) {
        return new Code() {
                public String getName() { return "switchLabel"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    ClassEnvironment e1 = switchee.defineClasses(e);
                    e1 = e1.openScope();
                    for (Iterator i = cases.iterator(); i.hasNext(); ) {
                        SwitchCase sc = (SwitchCase)i.next();
                        e1 = sc.code.defineClasses(e1);
                    }
                    return e.closeScope(defaultcode.defineClasses(e1));
                }
                public Environment defineSupers(Environment e) {
                    Environment e1 = switchee.defineSupers(e);
                    for (Iterator i = cases.iterator(); i.hasNext(); ) {
                        SwitchCase sc = (SwitchCase)i.next();
                        e1 = sc.code.defineSupers(e1);
                    }
                    return defaultcode.defineSupers(e1);
                }
                public Environment defineMembers(Environment e) {
                    Environment e1 = switchee.defineMembers(e);
                    for (Iterator i = cases.iterator(); i.hasNext(); ) {
                        SwitchCase sc = (SwitchCase)i.next();
                        e1 = sc.code.defineMembers(e1);
                    }
                    return defaultcode.defineMembers(e1);
                }
                public ClosedCode eval(Environment env) {
                    final ClosedCode s = switchee.eval(env);
                    if (!s.type.isInteger())
                        throw error("Switch argument must be integral, not "+s.type, env);
                    final Instr endlabel = Instr.genNil("End of switch");
                    Environment env1 = env.defineBreakLabel(label, endlabel);
                    if (s.isConstant()) {
                        // XXX Fixme: Find the right switch case.
                    }
                    final SortedMap casesmap = new TreeMap();
                    final ClosedCode d = defaultcode.evalNoUse(env1);
                    final Instr casescode = d.bytecode;
                    if (d.type != Type.bottom)
                        casescode.addGoto(endlabel);
                    final MonoList casesclosed = new ArrayMonoList();
                    final ClosedCode result =
                        new ClosedCode(s.bytecode,
                                       env, Type.void_type)/*.addInfo(s).addInfo(d)*/;
                    for (Iterator i = cases.iterator(); i.hasNext(); ) {
                        final ClosedCode cc = addCases((SwitchCase)i.next(), casesmap, env1);
                        env1 = cc.env;
                        casescode.append(cc.bytecode);
                        //result/*.addInfo(cc)*/;
                    }
                    result.bytecode.addSwitch(casesmap, d.bytecode).
                        append(casescode).
                        append(endlabel);
                    return result;
                }
                public String toString() {
                    String s = "";
                    if (label != null && label.length() > 0)
                        s = label+": ";
                    s += "switch ("+switchee+") {\n";
                    for (Iterator i = cases.iterator(); i.hasNext(); ) {
                        final SwitchCase cs = (SwitchCase)i.next();
                        for (Iterator j = cs.caselist.iterator(); j.hasNext(); ) {
                            s += "case "+((Code)j.next())+":\n";
                        }
                        s += cs.code+"\n";
                    }
                    if (defaultcode != null)
                        s += "default: "+defaultcode;
                    s += "}";
                    return s;
                }
            };
    }

    public static Code switchNoLabel(Code switches, MonoList cases, Code defaultcode) {
        return switchLabel("", switches, cases, defaultcode);
    }

    public static final boolean autoboxing = false;

    protected static ClosedCode box(ClosedCode c, Type t) {
        String typeClass;
        switch (c.type.kind) {
        case Type.BYTE: typeClass =  "java/lang/Byte"; break;
        case Type.CHAR: typeClass =  "java/lang/Char"; break;
        case Type.SHORT: typeClass =  "java/lang/Short"; break;
        case Type.INT : typeClass =  "java/lang/Integer"; break;
        case Type.LONG: typeClass =  "java/lang/Long"; break;
        case Type.FLOAT: typeClass =  "java/lang/Float"; break;  
        case Type.DOUBLE: typeClass =  "java/lang/Double"; break; 
        case Type.BOOLEAN: typeClass =  "java/lang/Boolean"; break;
        default: throw error("Internal: Not primitive after all: "+c.type, null);
        }

        final MethodType inittype = new MethodType(Type.void_type, c.type);
        MethodInfo initializer;
        try {
            initializer = c.env.findMethod(typeClass, "<init>", inittype);
        } catch (MethodNotFoundException e) {
            throw error("ARGH! Can't find "+typeClass+".<init>("+c.type+": "+e, null);
        }

        return new ClosedCode(Instr.genNew(typeClass).
                              addDup(1, 0).
                              append(c.bytecode).
                              addInvokenonvirtual(initializer),
                              c.env, new Type(typeClass))/*.addInfo(c)*/;
    }

    protected static ClosedCode unbox(ClosedCode c, Type t) {
        final String typeClass;
        switch (t.kind) {
        case Type.BYTE: typeClass =  "java/lang/Byte"; break;
        case Type.CHAR: typeClass =  "java/lang/Char"; break;
        case Type.SHORT: typeClass =  "java/lang/Short"; break;
        case Type.INT : typeClass =  "java/lang/Integer"; break;
        case Type.LONG: typeClass =  "java/lang/Long"; break;
        case Type.FLOAT: typeClass =  "java/lang/Float"; break;  
        case Type.DOUBLE: typeClass =  "java/lang/Double"; break; 
        case Type.BOOLEAN: typeClass =  "java/lang/Boolean"; break;
        default: throw error("Internal: Not primitive unbox after all: "+t, null);
        }

        final MethodType valuetype = new MethodType(c.type);
        try {
            final MethodInfo valuemethod = c.env.findMethod(typeClass, t+"Value", valuetype);
            return new ClosedCode(c.bytecode.addInvokevirtual(valuemethod),
                                  c.env, new Type(typeClass))/*.addInfo(c)*/;
        } catch (MethodNotFoundException e) {
            throw error("ARGH! Can't find "+typeClass+"."+t+"Value(): "+e, null);
        }
    }

    protected static ClosedCode valueConvert(ClosedCode c, Type t) {
        if (c.type.isPrimitive() && !(c.type.equals(t))) {
            if (c.isConstant()) {
                final ConstVal converted = ConstVal.make(c.constantValue(), t);
                return new ClosedCode(converted.genLoadCode(),
                                      c.env, t, converted);
            } else {
                if (autoboxing) {
                    if (!t.isPrimitive())
                        return box(c, t);
                }
                return new ClosedCode(c.bytecode.addConvert(c.type, t),
                                      c.env, t)/*.addInfo(c)*/;
            }
        } else {
            if (autoboxing) {
                if (t.isPrimitive() && !c.type.isPrimitive()) {
                    return unbox(c, t);
                }
            }
            return c;
        }
    }

    /* *********** Field, variable & array access. *********** */

    /* Allows assignment to any variable without knowing its kind.
     * @param lhs A getQualifiedName(), getSuperField() etc Code value.
     * @param value The expression assigned to the LHS
     */
    public static Code putValue(final Code lhs, final Code value) {
        return new Code() {
                public String getName() { return "putValue"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return value.defineClasses(lhs.defineClasses(e));
                }
                public Environment defineSupers(Environment e) {
                    return value.defineSupers(lhs.defineSupers(e));
                }
                public Environment defineMembers(Environment e) {
                    return value.defineMembers(lhs.defineMembers(e));
                }
                public ClosedCode eval(Environment env) {
                    return lhs.evalLHS(value, -1, env);
                }
                public ClosedCode evalNoUse(Environment env) {
                    return lhs.evalLHSNoUse(value, -1, env);
                }
                public String toString() {
                    return lhs+" = "+value;
                }
            };
    }

    /* Allows assignment to any variable without knowing its kind.
     * @param lhs A getQualifiedName(), getSuperField() etc Code value.
     * @param operator The operator to the left of the equals sign.
     * @param value The expression assigned to the LHS
     */
    public static Code putValue(final Code lhs, final int op,
                                final Code value) {
        return new Code() {
                public String getName() { return "putValue"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return value.defineClasses(lhs.defineClasses(e));
                }
                public Environment defineSupers(Environment e) {
                    return value.defineSupers(lhs.defineSupers(e));
                }
                public Environment defineMembers(Environment e) {
                    return value.defineMembers(lhs.defineMembers(e));
                }
                public ClosedCode eval(Environment env) {
                    return lhs.evalLHS(value, op, env);
                }
                public ClosedCode evalNoUse(Environment env) {
                    return lhs.evalLHSNoUse(value, op, env);
                }
                public String toString() {
                    return lhs+" "+mathopnames[op]+"= "+value;
                }
            };
    }

    /** Returns true is the string is the name of an existing class */
    public static boolean isClassName(String name, Environment env) {
        return env.isClass(name);
    }

    public static Code dotClass(final String name) {
        return new Code() {
                public String getName() { return "dotClass"; }
                public ClosedCode eval(Environment env) {
                    if (Type.typeclasses.containsKey(name)) {
                        String classname = 
                            (String)Type.typeclasses.get(name);
                        return new ClosedCode
                            (Instr.genGetstatic("java/lang/"+classname,
                                                "TYPE",
                                                Type.class_type),
                             env, Type.class_type);
                    } else {
                        final String s = normalizeClassNames(name, env).
                            replace('/', '.');
                        return new ClosedCode
                            (Instr.genConst(s).
                             addInvokestatic("java/lang/Class",
                                             "forName",
                                             new MethodType(Type.class_type,
                                                            Type.string_type)),
                             env, Type.class_type);
                    }
                }
                public String toString() {
                    return name+".class";
                }
            };
    }

    /** Returns code to load the 'this' variable */
    public static Code getThis() {
        return new Code() {
                public String getName() { return "getThis"; }
                public ClosedCode eval(Environment env) {
                    return new ClosedCode(Instr.genLoad(0, env.getSelfType()),
                                          env, env.getSelfType());
                }
                public String toString() {
                    return "this";
                }
            };
    }

    /** Returns code to load a qualified 'this' variable */
    public static Code getThis(final String name) {
        return new Code() {
                public String getName() { return "getThis"; }
                public ClosedCode eval(Environment env) {
		    Environment e = env;
		    Type t = new Type(name);
		    while (e != null && !e.getSelfType().equals(t)) {
			e = (Environment)e.getOuter();
		    }
		    if (e == null) {
			throw error("No outer class "+name+" found", env);
		    }
		    
		    if (e == env)
			return new ClosedCode(Instr.genLoad(0, env.getSelfType()),
					      env, env.getSelfType());
		    else {
			Environment outer = env;
			Instr loadcode = Instr.genLoad(0, env.getSelfType());
			while (outer != e) {
			    if (outer.getOuter() == null) {
				throw error("Internal: Outer env not found?", env);
			    }
			    loadcode.addGetfield(outer.getCurrentClass().getName(),
						 "this$"+(outer.numEnclosingClasses()-1),
						 outer.getOuter().getSelfType());
			    outer = (Environment)outer.getOuter();
			}
			return new ClosedCode(loadcode, env,
					      outer.getSelfType());
		    }
                }
                public String toString() {
                    return name+".this";
                }
            };
    }

    protected static ClosedCode putVariable(String name,
                                            final int op,
                                            Code value,
                                            Environment env,
                                            boolean isUsed) {
        final ClosedCode value1 = value.eval(env);
        final Var v = env.lookupVar(name);
        if (v == null) {
            throw error("Can't find variable "+name, env);
        } else {
            if (v.isFinal() && op != -1) {
                throw error("Trying to op= final variable "+v, env);
            }
            final ClosedCode value2 = valueConvert(value1, v.type);
            ConstVal cv = null;
            Instr bytecode = value2.bytecode;
            if (value2.isConstant()) {
                bytecode = value2.constantValue().genLoadCode();
            }
            return new ClosedCode((op == -1?Instr.genNil("No arith"):
                                   Instr.genLoad(v.index, v.type)).
                                  append(bytecode).
                                  append(op == -1?Instr.genNil("No arith"):
                                         Instr.genArith(op, v.type)).
                                  append(isUsed?Instr.genDup(value2.type):
                                         Instr.genNil("Not used")).
                                  addStore(v.index, v.type),
                                  env, isUsed?v.type:Type.void_type)
                /*.addInfo(value1)*/;
        }
    }

    

    protected static ClosedCode getLocalVariable(String name, Environment env) {
        final Var v = env.lookupVar(name);
        if (v != null) {
            if (v.isConstant()) {
                return new ClosedCode(v.constantValue().genLoadCode(),
                                      env, v.type, v.constantValue());
            } else {
                return new ClosedCode(Instr.genLoad(v.index, v.type),
                                      env, v.type);
            }
        }
        return null;
    }
    
    protected static ClosedCode getVariable(String name, Environment env) {
        ClosedCode cc = getLocalVariable(name, env);
        if (cc != null) return cc;
        /* We must obey the shadowing rules here. */
        return getOuterFieldOrVariable(name, env, ALLOW_VARIABLES,
                                       Instr.genNil("outer"));
    }

    public static Code getVariable(final String name) {
        return new Code() {
                public String getName() { return "getVariable"; }
                public ClosedCode evalLHSNoUse(final Code value, final int op,
                                                  Environment env) {
                    return putVariable(name, op, value, env, false);
                }
                public ClosedCode evalLHS(final Code value, final int op,
                                             Environment env) {
                    return putVariable(name, op, value, env, true);
                }
                public ClosedCode eval(Environment env) {
                    return getVariable(name, env);
                }
                public String toString() {
                    return name;
                }
            };
    }

    /** Look for a field or variable in an outer class, mangling names
     * appropriately.
     * @param name Name of the field
     * @param env Current environment
     * @param forv Whether to ALLOW_FIELDS, ALLOW_VARIABLES, or ALLOW_BOTH
     * @param loadcode Code to load any inner objects onto the stack.
     * @return Closed code to load an outer field or variable value, or
     * null if none is found.  An error is thrown if a field is found where a 
     * variable is expected or vice-versa.
     */
    protected static ClosedCode getOuterFieldOrVariable(String name,
                                                        Environment env,
                                                        int forv,
                                                        Instr loadcode) {
        Environment outer = (Environment)env.getOuter();
        if (outer == null) return null;
        final Var v = outer.lookupVar(name);
        int enclosing = outer.numEnclosingClasses();
        if (v != null) {
            if (forv == ALLOW_FIELDS) 
                throw error("Accessing shadowed field "+name, env);
            if (v.isConstant()) {
                return new ClosedCode(v.constantValue().genLoadCode(),
                                      env, v.type, v.constantValue());
            } else {
                return new ClosedCode(loadcode.
                                      addGetfield(env.getCurrentClass().getName(),
                                                  "val$"+name,
                                                  v.type),
                                      env, v.type);
            }
        } else {
	    final FieldInfo f = outer.findField(outer.getCurrentClass(), name);
	    if (f != null) {
		if (forv == ALLOW_VARIABLES) 
		    throw error("Accessing shadowed variable "+name, env);
		/* Get field from enclosing instance */
		if (f.isConstant() && ConstVal.legalConstantType(f.getType())) {
		    final ConstVal cv = ConstVal.make(f.constantValueObject(), f.getType());
		    return new ClosedCode(cv.genLoadCode(), env, f.getType(), cv);
		} else if (f.isStatic()) {
		    return new ClosedCode(Instr.genGetstatic(f), env, f.getType());
		} else {
		    return new ClosedCode(loadcode.
					  addGetfield(env.getCurrentClass().getName(),
						      "this$"+enclosing,
						      outer.getSelfType()).
					  addGetfield(outer.getCurrentClass().getName(),
						      name,
						      f.getType()),
					  env, f.getType());
		}
	    } else {
		// Recurse, loading the appropriate enclosing instance.
		ClosedCode cc = getOuterFieldOrVariable(name, outer, forv,
							Instr.genGetfield(env.getCurrentClass().getName(),
									  "this$"+enclosing,
									  outer.getSelfType()));
		return cc;
	    }
	}
    }
    /*protected*/
    public static ClosedCode getFieldOrVariable(String name, Environment env) {
        final ClosedCode cc = getLocalVariable(name, env);
        if (cc != null) return cc;
        else {
            final FieldInfo f = env.findField(env.getCurrentClass(), name);
            if (f == null) {
                ClosedCode outerget = getOuterFieldOrVariable(name, env,
                                                              ALLOW_BOTH,
                                                              Instr.genLoad(0, env.getSelfType()));
                return outerget;
            } else {
		if (f.isConstant() && ConstVal.legalConstantType(f.getType())) {
		    final ConstVal cv = ConstVal.make(f.constantValueObject(), f.getType());
		    return new ClosedCode(cv.genLoadCode(), env, f.getType(), cv);
                                      
		} else if (f.isStatic()) {
		    return new ClosedCode(Instr.genGetstatic(f),
					  env, f.getType());
		} else {
		    return new ClosedCode(Instr.genLoad(0, env.getSelfType()).
					  addGetfield(f),
					  env, f.getType());
		}
	    }
	}
    }

    /** Create the code to actually get the field, including popping
     *  any leftover value from static field lookup targets.
     */
    protected static Instr doGetField(FieldInfo f, Type t) {
        if (f.isStatic()) {
            return Instr.genPop(t).addGetstatic(f);
        } else {
            return Instr.genGetfield(f);
        }
    }

    /** Create code to lookup a number of fields from a value */
    protected static ClosedCode getFields(ClosedCode c, MonoList names) {
        Instr code = c.bytecode;
        Type classtype = c.type;
        ConstVal exprconstval = null;
        for (Iterator i = names.iterator(); i.hasNext(); ) {
            final String name = (String)i.next();
            final FieldInfo f = c.env.findField(classtype.className(), name);
            if (f == null)
                throw error("Can't find field "+classtype+"."+name, null);
            classtype = f.getType();
            if (f.isConstant() && ConstVal.legalConstantType(f.getType())) {
                ConstVal cv = ConstVal.make(f.constantValueObject(), f.getType());
                code = c.bytecode;
                code = code.addPop(c.type).append(cv.genLoadCode());
                exprconstval = cv;
            } else {
                code = code.append(doGetField(f, c.type));
                exprconstval = null;
            }
        }
        return new ClosedCode(code, c.env, classtype, exprconstval)/*.addInfo(c)*/;
    }

    protected static ClosedCode getField(ClosedCode c, String name) {
        final Type classtype = c.type;
        final FieldInfo f = c.env.findField(classtype.className(), name);
        if (f == null) {
            ClosedCode cc = getOuterFieldOrVariable(name, c.env, ALLOW_FIELDS,
                                                    Instr.genNil("Buggy?"));
            if (cc == null)
                throw error("Can't find field "+classtype+"."+name, null);
            System.out.println("Not fully implemented.");
            return cc;
        } else {
	    if (f.isConstant() && ConstVal.legalConstantType(f.getType())) {
		final ConstVal cv = ConstVal.make(f.constantValueObject(), f.getType());
		return new ClosedCode(c.bytecode.
				      addPop(c.type).
				      append(cv.genLoadCode()),
				      c.env, f.getType(), cv)/*.addInfo(c)*/;
	    } else {
		return new ClosedCode(c.bytecode.
				      append(doGetField(f, c.type)),
				      c.env, f.getType())/*.addInfo(c)*/;
	    }
	}
    }

    protected static ClosedCode qualifiedNameLookup1(MonoList names, 
                                                     String prefix,
                                                     Environment env) {
        if (names.size() == 0) return null;
        try {
            final ClassInfo c = env.get(prefix);
            final ClosedCode cc = 
                new ClosedCode(Instr.genNil("Static lookup"),
                               env, new Type(prefix, true));
            return getFields(cc, names);
        } catch (ClassNotFoundException e) {
            System.out.println("No class "+prefix+": "+e);
        }
        final String name = (String)names.get(0);
        final MonoList rest = names.subList(1, names.size());
        return qualifiedNameLookup1(rest, prefix+"/"+name, env);
    }

    protected static ClosedCode qualifiedNameLookup(MonoList names,
                                                    Environment env) {
        final String name = (String)names.get(0);
        final ClosedCode cc = getFieldOrVariable(name, env);
        if (cc != null) {
            if (names.size() > 1) {
                final MonoList rest = names.subList(1, names.size());
                return getFields(cc, rest);
            } else {
                return cc;
            }
        } else {
            final MonoList rest = names.subList(1, names.size());
            return qualifiedNameLookup1(rest, name, env);
        }
    }
    /** Create code to read a field from something known to be an object.
     *  An expression like 'a.b.c' can't use this, but 'a.foo().c' can.
     */
    public static Code getField(final Code target,
                                final String name) {
        return new Code() {
                public String getName() { return "getField"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return target.defineClasses(e);
                }
                public Environment defineSupers(Environment e) {
                    return target.defineSupers(e);
                }
                public Environment defineMembers(Environment e) {
                    return target.defineMembers(e);
                }
                public ClosedCode evalLHSNoUse(final Code value, final int op,
                                                  Environment env) {
                    return putFieldClosed(target.eval(env), name, op, 
                                          value.eval(env), false);
                }
                public ClosedCode evalLHS(final Code value, final int op,
                                             Environment env) {
                    return putFieldClosed(target.eval(env), name, op,
                                          value.eval(env), true);
                }
                public ClosedCode eval(Environment env) {
                    final ClosedCode target1 = target.eval(env);
                    if (name.equals("length") &&
                        target1.type.isArray()) {
                        return new ClosedCode(target1.bytecode.
                                              addArrayLength(),
                                              env, Type.int_type);
                    }
                    return getField(target1, name);
                }
                public String toString() {
                    return target+"."+name;
                }
            };
    }

    public static Code getQualifiedNameOrClass(final MonoList names) {
        return new Code() {
                public String getName() { return "getQualifiedNameOrClass"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) { 
                    Local v = e.findFreeVar((String)names.get(0));
                    if (v != null && (v.flags & Flags.FINAL) != 0 ) {
                        e = e.addFree(v);
                    }
                    return e; 
                }
                public Environment defineSupers(Environment e) { 
                    return e; 
                }
                public Environment defineMembers(Environment e) { 
                    return e; 
                }
                public ClosedCode eval(Environment env) {
                    ClosedCode cc = null;
                    try {
                        cc = getQualifiedName(names).eval(env);
                    } catch (Error e) {
                        // This is an ugly hack, we should really restructure
                        // the stuff around here.  QualifiedNameLookup should
                        // return null, then errors come outside it.
                    }
                    if (cc == null) {
                        final String classname = toJoinedString(names, "/");
                        if (isClassName(classname, env)) {
                            cc = new ClosedCode(Instr.genNil("Class name lookup"),
                                                env, new Type(classname, true));
                        } else {
                            throw error("Can't find class "+classname, env);
                        }
                    }
                    return cc;
                }
                public String toString() {
                    return toJoinedString(names, ".");
                }
            };
    }

    public static Code getQualifiedNameOrClass(final String name) {
        return new Code() {
                public String getName() { return "getQualifiedNameOrClass"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    Local v = e.findFreeVar(name);
                    if (v != null && (v.flags & Flags.FINAL) != 0 ) {
                        e = e.addFree(v);
                    }
                    return e;
                }
                public Environment defineSupers(Environment e) {
                    return e; 
                }
                public Environment defineMembers(Environment e) {
                    return e; 
                }
                public ClosedCode eval(Environment env) {
                    final ClosedCode cc = getFieldOrVariable(name, env);
                    if (cc == null) {
                        if (isClassName(name, env)) {
                            return new ClosedCode(Instr.genNil("Class name lookup"),
                                                  env, new Type(name, true));
                        }
                    }
                    return cc;
                }
                public String toString() {
                    return name;
                }
            };
    }

    /** Create code to get a value from a dotted list.
     *  This does not allow the result to just be a class type.
     */
    public static Code getQualifiedName(final MonoList names) {
        return new Code() {
                public String getName() { return "getQualifiedName"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) { 
                    Local v = e.findFreeVar((String)names.get(0));
                    if (v != null && (v.flags & Flags.FINAL) != 0 ) {
                        e = e.addFree(v);
                    }
                    return e; 
                }
                public Environment defineSupers(Environment e) { 
                    return e; 
                }
                public Environment defineMembers(Environment e) { 
                    return e; 
                }
                public ClosedCode evalLHSNoUse(final Code value, final int op,
                                                  Environment env) {
                    return putQualifiedNameClosed(names, op, value, env, false);
                }
                public ClosedCode evalLHS(final Code value, final int op,
                                             Environment env) {
                    return putQualifiedNameClosed(names, op, value, env, true);
                }
                public ClosedCode eval(Environment env) {
                    /* The `length' operator is special */
                    if (names.get(names.size()-1).equals("length")) {
                        final MonoList fieldnames = names.subList(0, names.size()-1);
                        final ClosedCode q = getQualifiedNameOrClass(fieldnames).eval(env);
                        if (q.type.isArray()) {
                            return new ClosedCode(q.bytecode.
                                                  addArrayLength(),
                                                  env, Type.int_type);
                        } else {
                            return getField(q, "length");
                        }
                    } else {
                        ClosedCode cc = qualifiedNameLookup(names, env);
                        if (cc == null)
                            throw error("Can't find qualified name "+
                                        toJoinedString(names, "."), env);
                        return cc;
                    }
                }
                public String toString() {
                    return toJoinedString(names, ".");
                }
            };
    }

    public static Code getQualifiedName(final Name name) {
        return new Code() {
                public String getName() { return "getQualifiedName"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    Local v = e.findFreeVar(name.getName());
                    if (v != null && (v.flags & Flags.FINAL) != 0 ) {
                        e = e.addFree(v);
                    }
                    return e; 
                }
                public Environment defineSupers(Environment e) {
                    return e; 
                }
                public Environment defineMembers(Environment e) {
                    return e; 
                }
                public ClosedCode evalLHSNoUse(final Code value, final int op,
                                               Environment env) {
                    return putVariable(name.getName(), op, value, env, false);
                }
                public ClosedCode evalLHS(final Code value, final int op,
                                          Environment env) {
                    return putVariable(name.getName(), op, value, env, true);
                }
                public ClosedCode eval(Environment env) {
                    final ClosedCode cc = getVariable(name.getName(), env);
                    if (cc == null) 
                        throw error("Not such name: "+name, env);
                    return cc;
                }
                public String toString() {
                    return name.toString();
                }
            };
    }

    /** Create code to get a value from a single-element dotted list.
     *  This does not allow the result to just be a class type.
     */
    public static Code getQualifiedName(final String name) {
        return new Code() {
                public String getName() { return "getQualifiedName"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    Local v = e.findFreeVar(name);
                    if (v != null && (v.flags & Flags.FINAL) != 0 ) {
                        e = e.addFree(v);
                    }
                    return e; 
                }
                public Environment defineSupers(Environment e) {
                    return e; 
                }
                public Environment defineMembers(Environment e) {
                    return e; 
                }
                public ClosedCode evalLHSNoUse(final Code value, final int op,
                                                  Environment env) {
                    return putFieldOrVariable(name, op, value, env, false);
                }
                public ClosedCode evalLHS(final Code value, final int op,
                                             Environment env) {
                    return putFieldOrVariable(name, op, value, env, true);
                }
                public ClosedCode eval(Environment env) {
                    final ClosedCode cc = getFieldOrVariable(name, env);
                    if (cc == null) 
                        throw error("Not such name: "+name, env);
                    return cc;
                }
                public String toString() {
                    return name;
                }
            };
    }

    protected static ClosedCode putFieldClosed(ClosedCode get, String name,
                                               final int op, ClosedCode value, 
                                               boolean isUsed) {
        final FieldInfo f = get.env.findField(get.type.className(), name);
        if (f == null)
            throw error("Can't find field "+get.type+"."+name, null);
        value = valueConvert(value, f.getType());
        Instr bytecode = value.bytecode;
        ConstVal cv = null;
        if (f.isFinal() && op != -1) {
            throw error("Trying to op= final field "+get.type+"."+name, null);
        }
        if (value.isConstant()) {
            cv = value.constantValue();
            if (f.isStaticFinal())
                f.makeConstant(cv.constantObject());
            bytecode = cv.genLoadCode();
        }
        if (f.isStatic()) {
            return new ClosedCode(get.bytecode.
                                  addPop(get.type).
                                  append(bytecode).
                                  append(op == -1?Instr.genNil("No arith"):
                                         Instr.genGetstatic(f).
                                         addArith(op, f.getType())).
                                  append(isUsed?Instr.genDup(value.type):
                                         Instr.genNil("Not used")).
                                  addPutstatic(f),
                                  value.env, isUsed?value.type:Type.void_type)
                /*.addInfo(get).addInfo(value)*/;
        } else if (op == Instr.ADD &&
                   f.getType().isObjectType("java/lang/String")) {
            final ClosedCode str2 = makeStringValue(value);
            return new ClosedCode(get.bytecode
                                  .addDup(get.type)
                                  .addGetfield(f)
                                  .append(str2.bytecode)
                                  .addInvokevirtual("java/lang/String",
                                                    "concat",
                                                    new MethodType(Type.string_type,
                                                                   Type.string_type))
                                  .append(isUsed?Instr.genDup(Type.string_type, get.type):
                                          Instr.genNil("Not used"))
                                  .addPutfield(f),
                                  value.env, 
                                  isUsed?Type.string_type:Type.void_type);
        }
        return new ClosedCode(get.bytecode.
                              append(op == -1?Instr.genNil("No arith"):
                                     Instr.genDup(get.type).
                                     addGetfield(f)).
                              append(bytecode).
                              append(op == -1?Instr.genNil("No arith"):
                                     Instr.genArith(op, f.getType())).
                              append(isUsed?Instr.genDup(value.type, get.type):
                                     Instr.genNil("Not used")).
                              addPutfield(f),
                              value.env, isUsed?value.type:Type.void_type)
            /*.addInfo(get).addInfo(value)*/;
    }

    public static Code putField(final String name, final Code value) {
        return new Code() {
                public String getName() { return "putField"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return value.defineClasses(e);
                }
                public Environment defineSupers(Environment e) {
                    return value.defineSupers(e);
                }
                public Environment defineMembers(Environment e) {
                    return value.defineMembers(e);
                }
                public ClosedCode eval(Environment e) {
                    return putSelfField(name, -1, value.eval(e), false);
                }
                public String toString() {
                    return name+" = "+value;
                }
            };
    }

    protected static ClosedCode putSelfField(String name, final int op,
                                             ClosedCode value,
                                             boolean isUsed) {
        final FieldInfo f = value.env.findField(name);
        if (f == null)
            throw error("Can't find field "+value.env.getCurrentClass()+
                  "."+name, null);
        if (f.isFinal() && op != -1) {
            throw error("Trying to op= final field "+value.env.getCurrentClass()
                  +"."+name, null);
        }
        value = valueConvert(value, f.getType());
        ConstVal cv = null;
        Instr bytecode = value.bytecode;
        if (value.isConstant()) {
            cv = value.constantValue();
            if (f.isStaticFinal()) {
                f.makeConstant(cv.constantObject());
            }
            bytecode = cv.genLoadCode();
        }
        if (f.isStatic()) {
            return new ClosedCode(bytecode.
                                  append(op == -1?Instr.genNil("No arith"):
                                         Instr.genGetstatic(f).
                                         addArith(op, f.getType())).
                                  append(isUsed?Instr.genDup(value.type):
                                         Instr.genNil("Not used")).
                                  addPutstatic(f), 
                                  value.env, isUsed?f.getType():Type.void_type)
                /*.addInfo(value)*/;
        } else {
            return new ClosedCode(Instr.genLoad(0, value.env.getSelfType()).
                                  append(op == -1?Instr.genNil("No arith"):
                                         Instr.genDup(value.env.getSelfType()).
                                         addGetfield(f)).
                                  append(bytecode).
                                  append(op == -1?Instr.genNil("No arith"):
                                         Instr.genArith(op, f.getType())).
                                  append(isUsed?Instr.genDup(value.type, value.env.getSelfType()):
                                         Instr.genNil("Not used")).
                                  addPutfield(f),
                                  value.env, isUsed?f.getType():Type.void_type)
                /*.addInfo(value)*/;
        }
    }

    protected static ClosedCode putFieldOrVariable(String name,
                                                 final int op,
                                                 Code value,
                                                 Environment env,
                                                 boolean isUsed) {
        final Var v = env.lookupVar(name);
        if (v == null) {
            final ClosedCode value1 = value.eval(env);
            return putSelfField(name, op, value1, isUsed);
        } else {
            ClosedCode value1;
            if (op != -1) {
                if (v.isFinal()) {
                    throw error("Trying to op= final variable "+v, env);
                }
                value1 = binOp(op, getVariable(name), value).eval(env);
            } else {
                value1 = value.eval(env);
            }
            final ClosedCode value2 = valueConvert(value1, v.type);
            ConstVal cv = null;
            Instr bytecode = value2.bytecode;
            if (value2.isConstant()) {
                bytecode = value2.constantValue().genLoadCode();
            }
            return new ClosedCode(bytecode.
                                  append(isUsed?Instr.genDup(value2.type):
                                         Instr.genNil("Not used")).
                                  addStore(v.index, v.type),
                                  env, isUsed?v.type:Type.void_type)
                /*.addInfo(value1)*/;
        }
    }

    protected static ClosedCode putQualifiedNameClosed(final MonoList names,
                                                     final int op, 
                                                     final Code value,
                                                     Environment env,
                                                     boolean isUsed) {
        if (names.size() > 1) {
            final ClosedCode value1 = value.eval(env);
            final MonoList target = names.subList(0, names.size()-1);
            ClosedCode c1 = qualifiedNameLookup(target, env);
            if (c1 == null) {
                try {
                    ClassInfo c = env.get(toJoinedString(target, "/"));
                    c1 = new ClosedCode(Instr.genNil("Static lookup"),
                                        env, new Type(c.getName(), true));
                } catch (ClassNotFoundException ce) {
                    throw error("Can't find qualified name "+target, env);
                }
            }
            return putFieldClosed(c1,
                                  (String)names.get(names.size()-1),
                                  op, value1, isUsed);
        } else {
            return putFieldOrVariable((String)names.get(0),
                                      op, value, env, isUsed);
        }
    }

    public static Code getSuperField(final String name) {
        return new Code() {
                public String getName() { return "getSuperField"; }
                public ClosedCode evalLHS(final Code value, final int op, 
                                             Environment env) {
                    return putSuperFieldClosed(name, op, value, env, true);
                }
                public ClosedCode evalLHSNoUse(final Code value, final int op,
                                                  Environment env) {
                    return putSuperFieldClosed(name, op, value, env, false);
                }
                public ClosedCode eval(Environment env) {
                    final Type self = env.getSelfType();
                    final ClassInfo selfclass = env.getClass(self.className());
                    final ClassInfo superclass = env.getClass(selfclass.supername);
                    if (superclass == null) {
                        throw error(selfclass+" has no superclass", env);
                    }
                    final FieldInfo f = env.findField(superclass.getName(), name);
                    if (f == null)
                        throw error("Can't find super field "+superclass.getName()+"."+name, env);
                    if (f.isConstant() && ConstVal.legalConstantType(f.getType())) {
                        final ConstVal cv = ConstVal.make(f.constantValueObject(), f.getType());
                        return new ClosedCode(cv.genLoadCode(),
                                              env, f.getType(), cv);
                    } else if (f.isStatic()) {
                        return new ClosedCode(Instr.genGetstatic(f),
                                              env, f.getType());
                    } else {
                        return new ClosedCode(Instr.genLoad(0, new Type(superclass.getName())).
                                              addGetfield(f),
                                              env, f.getType());
                    }
                }
                public String toString() {
                    return "super."+name;
                }
            };
    }

    protected static ClosedCode putSuperFieldClosed(final String name,
                                                    final int op,
                                                    final Code value,
                                                    Environment env,
                                                    boolean isUsed) {
        final ClosedCode value1 = value.eval(env);
        final Type self = env.getSelfType();
        final ClassInfo selfclass = env.getClass(self.className());
        final ClassInfo superclass = env.getClass(selfclass.supername);
        if (superclass == null) {
            throw error(selfclass+" has no superclass", env);
        }
        final FieldInfo f = env.findField(superclass.getName(), name);
        if (f == null)
            throw error("Can't find super field "+superclass.getName()+"."+name, env);
        if (op != -1 && f.isFinal()) {
            throw error("Trying to op= final field "+superclass.getName()+"."+name, env);
        }
        final ClosedCode value2 = valueConvert(value1, f.getType());
        // Check for constants
        Instr bytecode = value2.bytecode;
        ConstVal cv = null;
        if (value2.isConstant()) {
            cv = value2.constantValue();
            if (f.isFinal()) {
                f.makeConstant(cv.constantObject());
            }
            bytecode = cv.genLoadCode();
        }
        if (f.isStatic()) {
            return new ClosedCode(bytecode.
                                  append(op == -1?Instr.genNil("No arith"):
                                         Instr.genGetstatic(f).
                                         addArith(op, f.getType())).
                                  append(isUsed?Instr.genDup(f.getType()):
                                         Instr.genNil("Not used")).
                                  addPutstatic(f),
                                  env, isUsed?f.getType():Type.void_type)
                /*.addInfo(value2)*/;
        } else {
            final Type supertype = new Type(superclass.getName());
            return new ClosedCode(Instr.genLoad(0, supertype).
                                  append(op == -1?Instr.genNil("No arith"):
                                         Instr.genDup(supertype).
                                         addGetfield(f)).
                                  append(bytecode).
                                  append(op == -1?Instr.genNil("No arith"):
                                         Instr.genArith(op, f.getType())).
                                  append(isUsed?Instr.genDup(f.getType(), supertype):
                                         Instr.genNil("Not used")).
                                  addPutfield(f),
                                  env, isUsed?f.getType():Type.void_type)
                /*.addInfo(value2)*/;
        }
    }
    
    protected static ClosedCode arrayIncrementClosed(final Code arrayexpr,
                                                   final Code indexexpr,
                                                   final int value,
                                                   final boolean ispre,
                                                   Environment env,
                                                   boolean isUsed) {
        final ClosedCode arrayexpr1 = arrayexpr.eval(env);
        final ClosedCode indexexpr1 = indexexpr.eval(env);
        if (!arrayexpr1.type.isArray())
            throw error("Trying to index non-array type "+arrayexpr1.type, env);
        final Type elemtype = arrayexpr1.type.array_contents;
        final Instr pre1 = (!ispre && isUsed?Instr.genDup(elemtype, 2)
                            :Instr.genNil("Not pre or not used"));
        final Instr post1 = (ispre && isUsed?Instr.genDup(elemtype, 2)
                             :Instr.genNil("Not post or not used"));
        if (!indexexpr1.type.isInteger())
            throw error("Non-integer array index "+indexexpr1.type, env);
        if (!elemtype.isNumeric())
            throw error("Incrementing non-numeric array type "+
                  arrayexpr1.type.array_contents, env);
        return new ClosedCode(arrayexpr1.bytecode.
                              append(indexexpr1.bytecode).
                              addConvert(indexexpr1.type, Type.int_type).
                              addDup(2, 0).
                              addALoad(elemtype).
                              append(pre1).
                              addConst(value).
                              addMath(Instr.ADD, elemtype).
                              append(post1).
                              addAStore(elemtype),
                              env, isUsed?elemtype:Type.void_type)
            /*.addInfo(arrayexpr1).addInfo(indexexpr1)*/;
    }
  
    protected static Code arrayIncrement(final Code arrayexpr, final Code indexexpr,
                                       final int value, final boolean ispre) {
        return new Code() {
                public String getName() { return "arrayIncrement"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return indexexpr.defineClasses(arrayexpr.defineClasses(e));
                }
                public Environment defineSupers(Environment e) {
                    return indexexpr.defineSupers(arrayexpr.defineSupers(e));
                }
                public Environment defineMembers(Environment e) {
                    return indexexpr.defineMembers(arrayexpr.defineMembers(e));
                }
                public ClosedCode evalNoUse(Environment env) {
                    return arrayIncrementClosed(arrayexpr, indexexpr, value,
                                                ispre, env, false);
                }
                public ClosedCode eval(Environment env) {
                    return arrayIncrementClosed(arrayexpr, indexexpr, value,
                                                ispre, env, true);
                }
                public String toString() {
                    return (ispre?(value>0?"++":"--"):"")+
                        arrayexpr+"["+indexexpr+"]"+
                        (!ispre?(value==1?"++":(value==-1?"--":"+= "+value)):"");
                }
            };
    }

    public static Code arrayPreincrement(final Code arrayexpr, Code indexexpr,
                                         final int value) {
        return arrayIncrement(arrayexpr, indexexpr, value, true);
    }

    public static Code arrayPostincrement(final Code arrayexpr, Code indexexpr,
                                          final int value) {
        return arrayIncrement(arrayexpr, indexexpr, value, false);
    }

    protected static ClosedCode incrementClosed(final Code getexpr, 
                                              final String fieldname,
                                              final int value,
                                              final boolean ispre,
                                              final Environment env,
                                              final boolean isUsed) {
        if (getexpr == null) {
            if (env.varDefined(fieldname)) {
                final Var v = env.lookupVar(fieldname);
                if (!v.type.isNumeric())
                    throw error("Attempting to increment `"+v+
                          "', which is of type "+v.type, env);
                if (v.type.isInteger()) {
                    if (ispre)
                        return new ClosedCode(Instr.genIinc(v.index, value).
                                              append(isUsed?
                                                     Instr.genLoad(v.index, v.type):
                                                     Instr.genNil("Not used")),
                                              env, v.type);
                    else
                        return new ClosedCode((isUsed?
                                               Instr.genLoad(v.index, v.type):
                                               Instr.genNil("Not used")).
                                              addIinc(v.index, value),
                                              env, v.type);
                } else {
                    return putVariable(fieldname, Instr.ADD,
                                       constant(value), env, isUsed);
                }
            }
            // Not a variable
            final FieldInfo f = env.findField(env.getSelfType().className(), fieldname);
            if (f == null)
                throw error("Can't find field "+env.getSelfType().className()+"."+fieldname, env);
            final Type ft = f.getType();
            if (!ft.isNumeric())
                throw error("Attempting to increment `"+f+"', which is of type "+ft, env);
            ClosedCode constcode = constant(value).eval(env);
            Instr constval = constcode.constantValue().genLoadCode();
            if (f.isStatic()) {
                // Static field
                final Instr pre0 = (!ispre && isUsed?Instr.genDup(ft):
                                    Instr.genNil("Not pre or not used"));
                final Instr post0 = (ispre && isUsed?Instr.genDup(ft):
                                     Instr.genNil("Not post or not used"));
                return new ClosedCode(Instr.genGetstatic(f).
                                      append(pre0).
                                      append(constval).
                                      addMath(Instr.ADD, ft).
                                      append(post0).
                                      addPutstatic(f),
                                      env, ft);
            } else {
                // Instance field
                final Instr pre1 = (!ispre && isUsed?Instr.genDup(ft,1):
                                    Instr.genNil("Not pre or not used"));
                final Instr post1 = (ispre && isUsed?Instr.genDup(ft,1):
                                     Instr.genNil("Not post or not used"));
                return new ClosedCode(Instr.genLoad(0, env.getSelfType()).
                                      addDup(1,0).
                                      addGetfield(f).
                                      append(pre1).
                                      append(constval).
                                      addMath(Instr.ADD, ft).
                                      append(post1).
                                      addPutfield(f),
                                      env, ft);
            }
        } else {  // Non-null getexpr
            final ClosedCode getexpr1 = getexpr.eval(env);
            final FieldInfo f = env.findField(getexpr1.type.className(), fieldname);
            if (f == null)
                throw error("Can't find field "+getexpr1.type.className()+"."+fieldname, env);
            final Type ft = f.getType();
            if (!ft.isNumeric())
                throw error("Attempting to increment `"+f+"', which is of type "+ft, env);
            ClosedCode constcode = constant(value).eval(env);
            Instr constval = constcode.constantValue().genLoadCode();
            if (f.isStatic()) {
                // Static field
                final Instr pre0 = (!ispre && isUsed?Instr.genDup(ft):
                                    Instr.genNil("Not pre or not used"));
                final Instr post0 = (ispre && isUsed?Instr.genDup(ft):
                                     Instr.genNil("Not post or not used"));
                return new ClosedCode(getexpr1.bytecode.
                                      addPop(getexpr1.type).
                                      addGetstatic(f).
                                      append(pre0).
                                      append(constval).
                                      addMath(Instr.ADD, ft).
                                      append(post0).
                                      addPutstatic(f),
                                      env, ft)
                    /*.addInfo(getexpr1)*/;
            } else {
                // Instance field
                final Instr pre1 = (!ispre && isUsed?Instr.genDup(ft,1):
                                    Instr.genNil("Not pre or not used"));
                final Instr post1 = (ispre && isUsed?Instr.genDup(ft,1):
                                     Instr.genNil("Not post or not used"));
                return new ClosedCode(getexpr1.bytecode.
                                      addDup(1,0).
                                      addGetfield(f).
                                      append(pre1).
                                      append(constval).
                                      addMath(Instr.ADD, ft).
                                      append(post1).
                                      addPutfield(f),
                                      env, ft)
                    /*.addInfo(getexpr1)*/;
            }
        }
    }

    protected static Code increment(final Code getexpr, final String fieldname,
                                    final int value, final boolean ispre) {
        return new Code() {
                public String getName() { return "increment"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    if (getexpr != null)
                        return getexpr.defineClasses(e);
                    else return e;
                }
                public Environment defineSupers(Environment e) {
                    if (getexpr != null)
                        return getexpr.defineSupers(e);
                    else return e;
                }
                public Environment defineMembers(Environment e) {
                    if (getexpr != null)
                        return getexpr.defineMembers(e);
                    else return e;
                }
                public ClosedCode eval(Environment env) {
                    return incrementClosed(getexpr, fieldname, value, ispre, env, true);
                }
                public ClosedCode evalNoUse(Environment env) {
                    return incrementClosed(getexpr, fieldname, value, ispre, env, false);
                }
                public String toString() {
                    return (ispre?(value>0?"++":"--"):"")+
                        (getexpr!=null?getexpr+".":"")+fieldname+
                        (!ispre?(value==1?"++":(value==-1?"--":"+= "+value)):"");
                }

            };
    }

    protected static Code increment(final Code getexpr, final Name name,
                                    final int value, final boolean ispre) {
        return new Code() {
                public String getName() { return "increment"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    if (getexpr != null)
                        return getexpr.defineClasses(e);
                    else return e;
                }
                public Environment defineSupers(Environment e) {
                    if (getexpr != null)
                        return getexpr.defineSupers(e);
                    else return e;
                }
                public Environment defineMembers(Environment e) {
                    if (getexpr != null)
                        return getexpr.defineMembers(e);
                    else return e;
                }
                public ClosedCode eval(Environment env) {
                    return incrementClosed(getexpr, name.getName(), value, ispre, env, true);
                }
                public ClosedCode evalNoUse(Environment env) {
                    return incrementClosed(getexpr, name.getName(), value, ispre, env, false);
                }
                public String toString() {
                    return (ispre?(value>0?"++":"--"):"")+
                        (getexpr!=null?getexpr+".":"")+name+
                        (!ispre?(value==1?"++":(value==-1?"--":"+= "+value)):"");
                }

            };
    }

    public static Code preincrement(final Code getexpr, final String fieldname,
                                    final int value) {
        return increment(getexpr, fieldname, value, true);
    }

    public static Code postincrement(final Code getexpr,
                                     final String fieldname,
                                     final int value) {
        return increment(getexpr, fieldname, value, false);
    }

    public static Code preincrement(final Code getexpr, final Name fieldname,
                                    final int value) {
        return increment(getexpr, fieldname, value, true);
    }

    public static Code postincrement(final Code getexpr,
                                     final Name fieldname,
                                     final int value) {
        return increment(getexpr, fieldname, value, false);
    }

    public static Code preincrement(final MonoList names, int value) {
	String lastfield = (String)names.get(names.size()-1);
	MonoList head = names.subList(0, names.size()-1);
	if (head.size() == 1) 
	    return increment(getQualifiedNameOrClass((String)head.get(0)), lastfield, value, true);
	else
	return increment(getQualifiedNameOrClass(head), lastfield, value, true);
    }

    public static Code postincrement(final MonoList names, int value) {
	String lastfield = (String)names.get(names.size()-1);
	MonoList head = names.subList(0, names.size()-1);
	if (head.size() == 1) 
	    return increment(getQualifiedNameOrClass((String)head.get(0)), lastfield, value, false);
	else
	    return increment(getQualifiedNameOrClass(head), lastfield, value, false);
    }

    /** Generate code to load a null object.
     */
    public static Code nullObject() {
        return new Code() {
                public String getName() { return "nullObject"; }
                public ClosedCode eval(Environment env) {
                    return new ClosedCode(Instr.genNull(),
                                          env, Type.null_type);
                }
            };
    }

    public static Code newObject(final String classname,
                                 final MonoList initvalues) {
        return new Code() {
                public String getName() { return "newObject"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    ClassEnvironment e1 = e;
                    for (Iterator i = initvalues.iterator(); i.hasNext(); ) {
                        Code c = (Code)i.next();
                        e1 = c.defineClasses(e1);
                    }
                    return e1;
                }
                public Environment defineSupers(Environment e) {
                    for (Iterator i = initvalues.iterator(); i.hasNext(); ) {
                        Code c = (Code)i.next();
                        e = c.defineSupers(e);
                    }
                    return e;
                }
                public Environment defineMembers(Environment e) {
                    for (Iterator i = initvalues.iterator(); i.hasNext(); ) {
                        Code c = (Code)i.next();
                        e = c.defineMembers(e);
                    }
                    return e;
                }
                public ClosedCode eval(Environment env) {

                    try {
                        final ClassInfo c = env.get(classname);
                        if (c.flagSet(Flags.ABSTRACT))
                            throw error("Cannot initialize abstract class "+classname, env);
                        final Type t = new Type(c.getName());
                        MonoList inits;
                        if (c.isInnerClass()) {
                            inits = new ArrayMonoList().addAll(initvalues);
                            if (!env.getCurrentMethod().isStatic() &&
                                !c.isStatic())
                                inits = inits.add(getThis());
                            for (Local free = env.getFree(); free != null;
                                 free = free.next) {
                                inits = inits.add(getVariable(free.name));
                            }
                        } else {
                            inits = initvalues;
                        }
                        final ClosedCode create = new ClosedCode(Instr.genNew(t.className()).
                                                                 addDup(1, 0),
                                                                 env, t);
                        MonoList argcodes = evalCodeList(inits, env);
                        final MethodInfo m = 
                            getInvokeTarget(env, t, "<init>", argcodes);
                        final ClosedCode invoke =
                            invokeMethodArgsEvaled(create, m,
                                                   argcodes, true);
                        return new ClosedCode(invoke.bytecode, env, t)
                            /*.addInfo(invoke)*/;
                    } catch (ClassNotFoundException e) {
                        throw error("Can't create instance: "+e, env);
                    }
                }
                public String toString() {
                    return "new "+classname+
                        "("+toJoinedString(initvalues, ", ")+")";
                }
            };
    }

    public static Code anonymousClass(final String classname,
                                      final MonoList initvalues,
                                      final MonoList members) {
        return new Code() {
                public String getName() { return "anonymousClass"; }
                ClassInfo c;
                Local free;
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    final ClassInfo outerc = e.getCurrentClass();
                    final String anonname = e.genSym(classname);
                    e = e.updateGenSym();
                    c = new ClassInfo(Flags.PRIVATE,
                                      anonname,
                                      null);
                    e.addClass(c);
                    ClassEnvironment e1 = new ClassEnvironment(e, c);
                    for (Iterator i = members.iterator(); i.hasNext(); ) {
                        Code d = (Code)i.next();
                        e1 = d.defineClasses(e1);
                    }
                    free = null;
                    for (Local v = e1.free; v != null; v = v.next) {
                        if (e.getVar(v.name) != null) {
                            free = new Var(v.name, 0, v.type, v.flags, free);
                        } else {
                            /* Pass as free to outer outer class */
                            e = e.addFree(v);
                        }
                    }
                    this.free = free;
                    e1 = e1.resetFree();
                    /* Add a constructor for these fields */
                    e.closeScope(e1);
                    return e;
                }
                public Environment defineSupers(Environment e) {
                    final String norm_super = normalizeClassNames(classname, e);
                    Environment env = e.startClass(c, free);
                    try {
                        final ClassInfo superc = env.get(norm_super);
                        if (superc.isInterface()) {
                            c.setSuperClass("java/lang/Object");
                            c.implementsInterface(superc.getName());
                        }
                        else
                            c.setSuperClass(norm_super);
                    } catch (ClassNotFoundException err) {
                            throw error("Super class not found: "+err, e);
                    }
                    for (final Iterator i = members.iterator(); i.hasNext(); ) {
                        Code d = (Code)i.next();
                        env = d.defineSupers(env);
                    }
                    return e;
                }
                public Environment defineMembers(Environment e) {
                    if (c == null)
                        throw error("Internal: Couldn't refind class "+
                              classname, e);
                    Environment env = e.startClass(c, free);
                    for (Iterator i = members.iterator(); i.hasNext(); ) {
                        final Code d = (Code)i.next();
                        env = d.defineMembers(env);
                    }
                    return e;
                }
                public ClosedCode eval(Environment e) {
                    final int numargs = initvalues.size();
                    int numconstrargs = numargs;
                    final String supername = normalizeClassNames(classname, e);

                    /* Find the number of actual arguments to constructor */
                    MethodInfo outerm = e.getCurrentMethod();
                    if (outerm != null && !outerm.isStatic())
                        numconstrargs++;
                    if (free != null)
                        numconstrargs += free.size();

                    MonoList argcodes = new ArrayMonoList(numconstrargs);
                    MonoList argdefs = new ArrayMonoList(numconstrargs);
                    Type[] argtypes = new Type[numargs];
                    int index = 0;

                    /* Evaluate given arguments */
                    for (Iterator i = initvalues.iterator(); i.hasNext(); ) {
                        Code initc = (Code)i.next();
                        ClosedCode c1 = initc.eval(e);
                        argcodes = argcodes.add(c1);
                        argdefs = argdefs.add(new ArgDef(Flags.FINAL, c1.type,
                                                         "arg$"+index));
                        argtypes[index++] = c1.type;
                    }

                    /* Evaluate constructed arguments:
                     * Immediately enclosing instance
                     * Free variables
                     */
                    if ((outerm != null && !outerm.isStatic()) ||
                        (outerm == null)){
                        argcodes = argcodes.add(getThis().eval(e));
                    }
                    for (Local free = this.free; free != null; free = free.next) {
                        argcodes = argcodes.add(getVariable(free.name).eval(e));
                    }

                    /* Evaluate class */
                    /*
                    if (e.getCurrentMethod() != null)
                        e = e.defineLocal(classname, new Type(c.getName()), 0);
                    */

                    Environment env = e.startClass(c, free);

                    /* These evals just create methods */
                    evalCodeList(members, env);

                    /* Make code to call the super constructor */
                    MonoList superargs = new ArrayMonoList();
                    for (index = 0; index < numargs; index++) {
                        superargs = superargs.add(getVariable("arg$"+index));
                    }
                    final Code supercall = invokeSuper("", "<init>", 
                                                       superargs);

                    /* Add appropriate constructor.
                     * This adds enclosing instance and free variable
                     * arguments, and stores those appropriately.
                     */
                    final Code d = defineConstructor(0,
                                                     argdefs,
                                                     new ArrayMonoList(),
                                                     supercall,
                                                     noop());
                    // members is not accessed after this, so why add?
                    //  members.add(d);
                    /* XXX Can we really do this in here? */
                    /* No.  It breaks if we have nested anonymous classes */
                    /* This eval call adds a method */
                    env = d.defineMembers(env);
                    d.eval(env);

                    try {
                        c.writeClassFile();
                    } catch (IOException ex) {
                    }

                    /* Create actual constructor call.
                     * This is done outside the class definition. */
                    final Type t = new Type(c.getName());
                    /** Note that this is evaled outside the class env */
                    final ClosedCode create = new ClosedCode(Instr.genNew(c).
                                                             addDup(1, 0),
                                                             e, t);
               
                    final MethodInfo m = getInvokeTarget(e, t, "<init>", argcodes);
                    final ClosedCode invoke =
                        invokeMethodArgsEvaled(create, m,
                                               argcodes, true);

                    return new ClosedCode(invoke.bytecode, e, t);
                    /*                        .addWritable(c);*/
                        /*.addInfo(invoke)*/
                }
                public String toString() {
                    return "new "+classname+
                        "("+toJoinedString(initvalues, ", ")+
                        ") {"+members+"}";
                }
            };
    }

    /** Generate code to create a new array.
     * @param type The type of the elements of the array.
     * @param sizes List of code giving the size in each dimension.
     */
    public static Code newArray(final Type type, final MonoList sizes) {
        return new Code() {
                public String getName() { return "newArray"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    ClassEnvironment e1 = e;
                    for (Iterator i = sizes.iterator(); i.hasNext(); ) {
                        Code c = (Code)i.next();
                        e1 = c.defineClasses(e1);
                    }
                    return e1;
                }
                public Environment defineSupers(Environment e) {
                    for (Iterator i = sizes.iterator(); i.hasNext(); ) {
                        Code c = (Code)i.next();
                        e = c.defineSupers(e);
                    }
                    return e;
                }
                public Environment defineMembers(Environment e) {
                    for (Iterator i = sizes.iterator(); i.hasNext(); ) {
                        Code c = (Code)i.next();
                        e = c.defineMembers(e);
                    }
                    return e;
                }
                public ClosedCode eval(Environment env) {
                    final Instr dimscode = Instr.genNil("Start of array creation");
                    final Type normtype = normalizeClassNames(type, env);
                    Type arraytype = normtype;
                    for (Iterator i = sizes.iterator(); i.hasNext(); ) {
                        final Code c = (Code)i.next();
                        final ClosedCode cc = c.eval(env);
                        if (!cc.type.isInteger())
                            throw error("Array dimension must be integral, not "+cc.type, env);
                        dimscode.append(cc.bytecode);
                        arraytype = new Type(arraytype);
                    }
                    return new ClosedCode(dimscode.
                                          addNewarray(normtype, sizes.size()),
                                          env, arraytype);
                }
                public String toString() {
                    return "new "+type+"["+toJoinedString(sizes, ", ")+"]";
                }
            };
    }
    /** Generate code to create a new array.
     * @param type The type of the elements of the array.
     * @param size Code giving the size of the array.
     */
    public static Code newArray(final Type type, final Code size) {
        return new Code() {
                public String getName() { return "newArray"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return size.defineClasses(e);
                }
                public Environment defineSupers(Environment e) {
                    return size.defineSupers(e);
                }
                public Environment defineMembers(Environment e) {
                    return size.defineMembers(e);
                }
                public ClosedCode eval(Environment env) {
                    final Type normtype = normalizeClassNames(type, env);
                    final ClosedCode cc = size.eval(env);
                    if (!cc.type.isInteger())
                        throw error("Array dimension must be integral, not "+cc.type, env);
                    return new ClosedCode(cc.bytecode.
                                          addNewarray(normtype, 1),
                                          env, new Type(normtype));
                }
                public String toString() {
                    return "new "+type+"["+size+"]";
                }
            };
    }

    /** Generate code to load a value from an array.
     * @param array Code for getting an array.
     * @param index Code for getting the index into the array.
     */
    public static Code getArray(final Code array, final Code index) {
        return new Code() {
                public String getName() { return "getArray"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return index.defineClasses(array.defineClasses(e));
                }
                public Environment defineSupers(Environment e) {
                    return index.defineSupers(array.defineSupers(e));
                }
                public Environment defineMembers(Environment e) {
                    return index.defineMembers(array.defineMembers(e));
                }
                public ClosedCode evalLHSNoUse(final Code rhs, final int op,
                                                  Environment env) {
                    return putArray(array, index, op, rhs, false).eval(env);
                }
                public ClosedCode evalLHS(final Code rhs, final int op,
                                             Environment env) {
                    return putArray(array, index, op, rhs, true).eval(env);
                }
                public ClosedCode eval(Environment env) {
                    final ClosedCode array1 = array.eval(env);
                    if (!array1.type.isArray())
                        throw error("Attempting to index non-array "+array1.type, env);
                    final ClosedCode index1 = index.eval(env);
                    if (!index1.type.isInteger())
                        throw error("Array index must be integer, not "+index1.type, env);
                    final Type elemtype = array1.type.array_contents;
                    return new ClosedCode(array1.bytecode.
                                          append(index1.bytecode).
                                          addALoad(elemtype),
                                          env, elemtype)
                        /*.addInfo(array1).addInfo(index1)*/;
                }
                public String toString() {
                    return array+"["+index+"]";
                }
            };
    }

    /** Generate code to store a value in an array.
     * @param array Code for getting an array.
     * @param index Code for getting the index into the array.
     * @param elem Code for getting the element to be put into the array.
     */
    /*private*/
    public static Code putArray(final Code array, final Code index, 
                                final int op, final Code elem, 
                                final boolean is_used) {
        return new Code() {
                public String getName() { return "putArray"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return elem.defineClasses(index.defineClasses(array.defineClasses(e)));
                }
                public Environment defineSupers(Environment e) {
                    return elem.defineSupers(index.defineSupers(array.defineSupers(e)));
                }
                public Environment defineMembers(Environment e) {
                    return elem.defineMembers(index.defineMembers(array.defineMembers(e)));
                }
                public ClosedCode eval(Environment env) {
                    final ClosedCode array1 = array.eval(env);
                    if (!array1.type.isArray())
                        throw error("Attempting to index non-array "+array1.type, env);
                    final ClosedCode index1 = index.eval(env);
                    if (!index1.type.isInteger())
                        throw error("Array index must be integer, not "+index1.type, env);
                    final ClosedCode elem1 = elem.eval(env);
                    final Type elemtype = array1.type.array_contents;
                    if (!elem1.type.conformsArray(elemtype))
                        throw error("Can't store "+elem1.type+" in array of "+array1.type, env);
                    final ClosedCode elem2 = valueConvert(elem1, elemtype);
                    return new ClosedCode(array1.bytecode.
                                          append(index1.bytecode).
                                          append(op == -1?Instr.genNil("No arith"):
                                                 Instr.genDup(2, 0).
                                                 addALoad(elemtype)).
                                          append(elem2.bytecode).
                                          append(op == -1?Instr.genNil("No arith"):
                                                 Instr.genArith(op, elemtype)).
                                          append(is_used?Instr.genDup(elemtype, 2):
                                                 Instr.genNil("Not used")).
                                          addAStore(elemtype),
                                          env, is_used?elemtype:Type.void_type)
                        /*.addInfo(array1).addInfo(index1).addInfo(elem1)*/;
                }
                public String toString() {
                    return array+"["+index+"] = "+elem;
                }
            };
    }

    /** Constructor that allocates and initializes a one-dimensional array.
     * @param type The type of the array.
     * @param inits List of Code generating the initializing values.
     */
    public static Code initializeArray(final Type type, final MonoList inits) {
        return new Code() {
                public String getName() { return "initializeArray"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    ClassEnvironment e1 = e;
                    for (Iterator i = inits.iterator(); i.hasNext(); ) {
                        Code c = (Code)i.next();
                        e1 = c.defineClasses(e1);
                    }
                    return e1;
                }
                public Environment defineSupers(Environment e) {
                    for (Iterator i = inits.iterator(); i.hasNext(); ) {
                        Code c = (Code)i.next();
                        e = c.defineSupers(e);
                    }
                    return e;
                }
                public Environment defineMembers(Environment e) {
                    for (Iterator i = inits.iterator(); i.hasNext(); ) {
                        Code c = (Code)i.next();
                        e = c.defineMembers(e);
                    }
                    return e;
                }
                public ClosedCode eval (Environment env) {
                    final Instr init_instr = Instr.genNil("Start of array init");
                    int index = 0;
                    final Type normtype = normalizeClassNames(type, env);
                    final ClosedCode endcode = new ClosedCode(Instr.genConst(inits.size()).
                                                              addNewarray(normtype.array_contents, 1),
                                                              env, normtype);
                    for (Iterator i = inits.iterator(); i.hasNext(); ) {
                        final Code c = (Code)i.next();
                        final ClosedCode c1 = c.eval(env);
                        final Type elemtype = normtype.array_contents;
                        if (!c1.type.conformsArray(elemtype))
                            throw error("Can't initialize array of "+elemtype+" with "+c1.type, env);
                        final ClosedCode c2 = valueConvert(c1, elemtype);
                        init_instr.append(Instr.genDup(1,0).
                                          addConst(index).
                                          append(c2.bytecode).
                                          addAStore(elemtype));
                        index++;
                        //endcode/*.addInfo(c1)*/;
                    }
                    endcode.bytecode.append(init_instr);
                    return endcode;
                }
                public String toString() {
                    return "{"+toJoinedString(inits, ", ")+"}";
                }
            };
    }

    /** Constructor that allocates and initializes a one-dimensional array.
     * @param type The type of the array.
     * @param init Code generating the initializing value.
     */
    public static Code initializeArray(final Type type, final Code init) {
        return new Code() {
                public String getName() { return "initializeArray"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return init.defineClasses(e);
                }
                public Environment defineSupers(Environment e) {
                    return init.defineSupers(e);
                }
                public Environment defineMembers(Environment e) {
                    return init.defineMembers(e);
                }
                public ClosedCode eval (Environment env) {
                    final Type normtype = normalizeClassNames(type, env);
                    final ClosedCode endcode = new ClosedCode(Instr.genConst(1).
                                                              addNewarray(normtype.array_contents, 1),
                                                              env, normtype);
                    final ClosedCode c1 = init.eval(env);
                    final Type elemtype = normtype.array_contents;
                    if (!c1.type.conformsArray(elemtype))
                        throw error("Can't initialize array of "+elemtype+" with "+c1.type, env);
                    final ClosedCode c2 = valueConvert(c1, elemtype);
                    endcode.bytecode.append(Instr.genDup(1,0).
                                            addConst(0).
                                            append(c2.bytecode).
                                            addAStore(elemtype));
                    return endcode/*.addInfo(c1)*/;
                }
                public String toString() {
                    return "{"+init+"}";
                }
            };
    }

    /** Arraylength constructor.
     * @param ac Code whose result is an array.q
     */
    public static Code arrayLength(final Code ac) {
        return new Code() { 
                public String getName() { return "arrayLength"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return ac.defineClasses(e);
                }
                public Environment defineSupers(Environment e) {
                    return ac.defineSupers(e);
                }
                public Environment defineMembers(Environment e) {
                    return ac.defineMembers(e);
                }
                public ClosedCode eval (Environment env) {
                    final ClosedCode ac1 = ac.eval(env);
                    if (!ac1.type.isArray())
                        throw error("Can't take array lengh of a "+ac1.type, env);
                    return new ClosedCode ( ac1.bytecode.addArrayLength(), env,
                                            Type.int_type)
                        /*.addInfo(ac1)*/;
                }
                public String toString() {
                    return ac+".length";
                }
            };
    }

    /* Invocation constructors */
    protected static MonoList evalCodeList(MonoList codes, final Environment env) {
        MonoList ccodes = new ArrayMonoList();

        if (use_fold) {
            return (MonoList)
                codes.foldl(new FoldOp() {
                        public Object eval(Object o1, Object o2) {
                            return ((MonoList)o1).add(((Code)o2).eval(env));
                        }
                    }, ccodes);
        } else {
            for (Iterator i = codes.iterator(); i.hasNext(); ) {
                Code c = (Code)i.next();
                ClosedCode c1 = c.eval(env);
                ccodes = ccodes.add(c1);
            }
            return ccodes;
        }
    }

    protected static ClosedCode makeInvokeCode(MethodInfo m,
                                               ClosedCode result,
                                               Instr argscode,
                                               boolean isInvokeSpecial) {
        result.bytecode.append(argscode);
        if (m.isStatic()) {
            result.bytecode.addInvokestatic(m);
        } else {
            /* Can't easily tell if this target is self 
               if (target.env.getCurrentMethod().isStatic()) {
               throw error("Can't call non-static method "+m+" from static method "+target.env.getCurrentMethod());
               }
            */
            if (isInvokeSpecial) {
                result.bytecode.addInvokenonvirtual(m);
            } else if (m.getClassInfo().isInterface()) {
                result.bytecode.addInvokeinterface(m);
            } else {
                result.bytecode.addInvokevirtual(m);
            }
        }
        return result;
    }

    protected static MethodInfo getInvokeTarget(Environment env,
                                                Type target_type,
                                                String name,
                                                MonoList argcodes) {
        MethodInfo m;
        if (!target_type.isReference())
            throw error("Can't invoke method on primitive type "+target_type, env);
        if (target_type.kind == Type.ARRAY)
            target_type = Type.object_type;
        try {
            Type[] argtypes = new Type[argcodes.size()];
            int index = 0;
            for (Iterator i = argcodes.iterator();
                 i.hasNext(); index++ ) {
                argtypes[index] = ((ClosedCode)i.next()).type;
            }
            MethodType mt = new MethodType(Type.void_type, argtypes);
            m = env.findMethod(target_type.className(), name, mt);
        } catch (MethodNotFoundException e) {
            throw error("Invoke error: "+e, env);
        }
        return m;
    }

    protected static ClosedCode invokeMethodArgsEvaled(ClosedCode target,
                                                       MethodInfo m,
                                                       MonoList argcodes,
                                                       boolean isInvokeSpecial) {
        final Instr argbytecodes = Instr.genNil("Start of method args");
        final ClosedCode result = new ClosedCode((target.bytecode != null?
                                                  target.bytecode:
                                                  Instr.genNil("Null invoke target")),
                                                 target.env,
                                                 m.getReturnType());
        if (m.isStatic()) 
            result.bytecode.addPop(target.type);
        //result/*.addInfo(target)*/;
        int index = 0;
        for (Iterator i = argcodes.iterator(); i.hasNext(); ) {
            final ClosedCode cc = (ClosedCode)i.next();
            argbytecodes.append(valueConvert(cc, m.getType().arguments[index++]).bytecode);
            //result/*.addInfo(cc)*/;
        }
        return makeInvokeCode(m, result, argbytecodes, isInvokeSpecial);
    }
    

    /* Invocation constructors */
    /* These constructors will occasionally (for static methods) add
     * a superfluous Aload and Pop.  This could be optimized away fairly
     * easily, together with superfluous fields loads before static loads.
     */

    public static Code invoke(final Code target, final String name,
                              final MonoList arguments) {
        return new Code() {
                public String getName() { return "invoke"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    ClassEnvironment e1 = target.defineClasses(e);
                    for (Iterator i = arguments.iterator(); i.hasNext(); ) {
                        Code c = (Code)i.next();
                        e1 = c.defineClasses(e1);
                    }
                    return e1;
                }
                public Environment defineSupers(Environment e) {
                    e = target.defineSupers(e);
                    for (Iterator i = arguments.iterator(); i.hasNext(); ) {
                        Code c = (Code)i.next();
                        e = c.defineSupers(e);
                    }
                    return e;
                }
                public Environment defineMembers(Environment e) {
                    e = target.defineMembers(e);
                    for (Iterator i = arguments.iterator(); i.hasNext(); ) {
                        Code c = (Code)i.next();
                        e = c.defineMembers(e);
                    }
                    return e;
                }
                public ClosedCode eval (Environment env) {
                    final ClosedCode target1 = target.eval(env);
                    if (target1 == null) {
                        throw error("Can't find method invocation target `"+target+"'", env);
                    }
                    final MonoList argcodes = evalCodeList(arguments, env);
                    final MethodInfo m = getInvokeTarget(env, target1.type, name, argcodes);
                    return invokeMethodArgsEvaled(target1, m, argcodes, false);
                }
                public String toString() {
                    return (target!=null?target+".":"")+name+
                        "("+toJoinedString(arguments, ", ")+")";
                }
            };
    }

    static class SelfMethodInfo {
        MethodInfo method; Environment env; 
        public SelfMethodInfo(MethodInfo m, Environment env) {
            this.method = m;
            this.env = env;
        }
    }

    protected static SelfMethodInfo getSelfMethodCore(String name,
                                                      MethodType type,
                                                      Environment env) {
        try {
            return new SelfMethodInfo(env.findMethod(env.getCurrentClass(),
                                                     name, type),
                                      env);
        } catch (MethodNotFoundException e) {
            if (env.selfIsInner())
                return getSelfMethodCore(name, type,
                                         (Environment)env.getOuter());
            return null;
        }
    }

    protected static SelfMethodInfo getSelfMethod(String name, 
                                                  MonoList argcodes,
                                                  Environment env) {
        final Type[] argtypes = new Type[argcodes.size()];
        int index = 0;
        for (Iterator i = argcodes.iterator();
             i.hasNext(); index++ ) {
            argtypes[index] = ((ClosedCode)i.next()).type;
        }
        final SelfMethodInfo m = getSelfMethodCore(name,
                                                   new MethodType(Type.void_type, argtypes),
                                                   env);
        if (m == null)
            throw error("Can't find self method "+name+
                        "("+toJoinedString(argtypes, ", ")+")", env);
        return m;
    }
    
    protected static ClosedCode makeSelfTarget(SelfMethodInfo m,
                                               Environment env) {
        if (m.method.isStatic()) 
            return new ClosedCode(Instr.genNil("Static self call"),
                                  env, Type.void_type);
        else {
            if (m.env == env)
                return new ClosedCode(Instr.genLoad(0, env.getSelfType()),
                                      env, env.getSelfType());
            else {
                Environment outer = env;
                Instr loadcode = Instr.genLoad(0, env.getSelfType());
                while (outer != m.env) {
                    if (outer.getOuter() == null) {
                        throw error("Internal: Outer env not found?", env);
                    }
                    loadcode.addGetfield(outer.getCurrentClass().getName(),
                                         "this$"+(outer.numEnclosingClasses()-1),
                                         outer.getOuter().getSelfType());
                    outer = (Environment)outer.getOuter();
                }
                return new ClosedCode(loadcode, env,
                                      outer.getSelfType());
            }
        }
    }

    public static Code invokeSelf(final String superclasses,
                                  final String name,
                                  final MonoList arguments) {
        return new Code() {
                public String getName() { return "invokeSelf"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    ClassEnvironment e1 = e;
                    for (Iterator i = arguments.iterator(); i.hasNext(); ) {
			Object nextItem = i.next();
			if(nextItem instanceof Code){
			    Code c = (Code)nextItem;
			    e1 = c.defineClasses(e1);
			}else{
			    i.remove();
			    arguments.addAll((MonoList)nextItem);
			}
                    }
                    return e1;
                }
                public Environment defineSupers(Environment e) {
                    for (Iterator i = arguments.iterator(); i.hasNext(); ) {
                        Code c = (Code)i.next();
                        e = c.defineSupers(e);
                    }
                    return e;
                }
                public Environment defineMembers(Environment e) {
                    for (Iterator i = arguments.iterator(); i.hasNext(); ) {
                        Code c = (Code)i.next();
                        e = c.defineMembers(e);
                    }
                    return e;
                }
                public ClosedCode eval (Environment env) {
                    // XXX Fixme:  Should deal with the superclasses.
                    MonoList argcodes = evalCodeList(arguments, env);
                    final SelfMethodInfo m = getSelfMethod(name, argcodes, env);
                    final Type selftype = m.env.getSelfType();
                    final ClosedCode selfcode = makeSelfTarget(m, env);
                    return invokeMethodArgsEvaled(selfcode, m.method,
                                                  argcodes,false);
                }
                public String toString() {
                    if (superclasses != null && superclasses.length() > 0)
                        return "SUPERSELF!";
                    return name+"("+toJoinedString(arguments, ", ")+")";
                }
            };
    }

    protected static Type getSuperType(Environment env) {
        final Type self = env.getSelfType();
        // XXX Fixme: Should deal with the superclasses.
        final ClassInfo selfclass = env.getClass(self.className());
        final ClassInfo superclass = env.getClass(selfclass.supername);
        if (superclass == null) 
            throw error("No super class for "+selfclass, env);
        return new Type(superclass.getName());
    }

    protected static ClosedCode makeSuperTarget(Type supertype,
                                                Environment env) {
        final MethodInfo m = env.getCurrentMethod();
        if (m.isStatic()) 
            return new ClosedCode(Instr.genNil("Static super call"),
                                  env, Type.void_type);
        else                        
            return new ClosedCode(Instr.genLoad(0, supertype),
                                  env, supertype);
    }

    public static Code invokeSuper(final String superclasses,
                                   final String name,
                                   final MonoList arguments) {
        return new Code() {
                public String getName() { return "invokeSuper"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    ClassEnvironment e1 = e;
                    for (Iterator i = arguments.iterator(); i.hasNext(); ) {
                        Code c = (Code)i.next();
                        e1 = c.defineClasses(e1);
                    }
                    return e1;
                }
                public Environment defineSupers(Environment e) {
                    for (Iterator i = arguments.iterator(); i.hasNext(); ) {
                        Code c = (Code)i.next();
                        e = c.defineSupers(e);
                    }
                    return e;
                }
                public Environment defineMembers(Environment e) {
                    for (Iterator i = arguments.iterator(); i.hasNext(); ) {
                        Code c = (Code)i.next();
                        e = c.defineMembers(e);
                    }
                    return e;
                }
                public ClosedCode eval (Environment env) {
                    final Type supertype;
                    supertype = getSuperType(env);
                    ClosedCode supercode = makeSuperTarget(supertype, env);
                    MonoList argcodes = evalCodeList(arguments, env);
                    final MethodInfo m = getInvokeTarget(env, supertype, name, argcodes);
                    return invokeMethodArgsEvaled(supercode, m, 
                                                  argcodes, true);
                }
                public String toString() {
                    if (superclasses != null && !superclasses.equals(""))
                        return superclasses+"."+"super."+name+"("+toJoinedString(arguments, ", ")+")";
                    return "super."+name+"("+toJoinedString(arguments, ", ")+")";
                }
            };
    }

    /** Init invoke on this */
    public static Code invokeSpecial(final String superclasses,
                                     final String name,
                                     final MonoList arguments) {
        return new Code() {
                public String getName() { return "invokeSpecial"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    ClassEnvironment e1 = e;
                    for (Iterator i = arguments.iterator(); i.hasNext(); ) {
                        Code c = (Code)i.next();
                        e1 = c.defineClasses(e1);
                    }
                    return e1;
                }
                public Environment defineSupers(Environment e) {
                    for (Iterator i = arguments.iterator(); i.hasNext(); ) {
                        Code c = (Code)i.next();
                        e = c.defineSupers(e);
                    }
                    return e;
                }
                public Environment defineMembers(Environment e) {
                    for (Iterator i = arguments.iterator(); i.hasNext(); ) {
                        Code c = (Code)i.next();
                        e = c.defineMembers(e);
                    }
                    return e;
                }
                public boolean isThisConstructorCall() { return true; }
                public ClosedCode eval (Environment env) {
                    final Type self = env.getSelfType();
                    // XXX Fixme: Should deal with the superclasses.
                    final ClassInfo selfclass = env.getClass(self.className());
                    if (name.equals("<init>")) {
                        MonoList argcodes = evalCodeList(arguments, env);
                        final ClosedCode selfcode = new ClosedCode(Instr.genLoad(0, self), env, Type.void_type);
                        final MethodInfo m = getInvokeTarget(env, self, name, argcodes);
                        final ClosedCode invokecode = 
                            invokeMethodArgsEvaled(selfcode, m, argcodes, true);
                        return invokecode;
                    } else {
                        throw error("Can't do invokeSpecial for non-init methods", env);
                    }

                }
                public String toString() {
                    if (superclasses != null) return "SUPERSPECIAL!";
                    return "this("+toJoinedString(arguments, ", ")+")";
                }
            };
    }


    public static Code doThrow(final Code c) {
        return new Code() {
                public String getName() { return "doThrow"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return c.defineClasses(e);
                }
                public Environment defineSupers(Environment e) {
                    return c.defineSupers(e);
                }
                public Environment defineMembers(Environment e) {
                    return c.defineMembers(e);
                }
                public ClosedCode eval (Environment env) {
                    final ClosedCode c1 = c.eval(env);
                    if (c1.type.kind != Type.OBJECT ||
                        !c1.type.subtypeOf(Type.throwable_type))
                        throw error("Thrown object not Throwable: "+c1.type, env);
                    return new ClosedCode(c1.bytecode.addThrow(),
                                          env, Type.bottom)/*.addInfo(c1)*/;
                }
                public String toString() {
                    return "throw "+c;
                }
            };
    }

    public static Code handlers(final Code c, final MonoList handlers) {
        return new Code() {
                public String getName() { return "handlers"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    ClassEnvironment e1 = c.defineClasses(e.openScope());
                    e = e.closeScope(e1);
                    for (Iterator i = handlers.iterator(); i.hasNext(); ) {
                        Handler h = (Handler)i.next();
                        e1 = e.openScope();
                        e1 = e1.defineVar(h.varname, h.type, h.flags);
                        e1 = h.code.defineClasses(e1);
                        e = e.closeScope(e1);
                    }
                    return e;
                }
                public Environment defineSupers(Environment e) {
                    e = c.defineSupers(e);
                    for (Iterator i = handlers.iterator(); i.hasNext(); ) {
                        Handler h = (Handler)i.next();
                        e = h.code.defineSupers(e);
                    }
                    return e;
                }
                public Environment defineMembers(Environment e) {
                    e = c.defineMembers(e);
                    for (Iterator i = handlers.iterator(); i.hasNext(); ) {
                        Handler h = (Handler)i.next();
                        e = h.code.defineMembers(e);
                    }
                    return e;
                }
                public ClosedCode eval (Environment env) {
                    final ClosedCode c1 = c.evalNoUse(env);
                    final Instr endlabel = Instr.genNil("End of try statement");
                    final Instr handlercodes = Instr.genNil("Try statement handlers");
                    final ClosedCode result = new ClosedCode(Instr.genNil("Try"),
                                                             env, Type.void_type);
                    //result/*.addInfo(c1)*/;
                    for (Iterator i = handlers.iterator(); i.hasNext(); ) {
                        final Handler h = (Handler)i.next();
                        final Type t = normalizeClassNames(h.type, env);
                        if (t.kind != Type.OBJECT ||
                            !t.subtypeOf(Type.throwable_type))
                            throw error("Handler type is not Throwable: "+t, env);
                        if (env.varDefined(h.varname)) {
                            throw error("Variable "+h.varname+" already defined", env);
                        }
                        final Environment env1 = env.defineVar(h.varname, t, h.flags, null);
                        final Var v = env1.getNewestVar();
                        final ClosedCode h1 = h.code.evalNoUse(env1);
                        final Instr handlercode = Instr.genStore(v.index, v.type).
                            append(h1.bytecode).
                            append(h1.type!=Type.bottom?
                                   Instr.genGoto(endlabel):
                                   Instr.genNil("Unreachable"));
                        handlercodes.append(handlercode);
                        c1.bytecode.addHandler(handlercode,
                                               v.type);
                        //result/*.addInfo(h1)*/;
                    }
                    result.bytecode.append(c1.bytecode).
                        append(c1.type!=Type.bottom?
                               Instr.genGoto(endlabel):
                               Instr.genNil("Unreachable")).
                        append(handlercodes).
                        append(endlabel);
                    return result;                      
                }
                public String toString() { 
                    String s = "try {\n";
                    s += c;
                    for (Iterator i = handlers.iterator(); i.hasNext(); ) {
                        Handler h = (Handler)i.next();
                        s += "} catch ("
                            +(h.flags!=0?Flags.toString(h.flags)+" ":"")
                            +h.type+" "+h.varname+") {\n";
                        s += h.code;
                    }
                    s += "}\n";
                    return s;
                }
            };
    }

    public static Code handlersWithFinally(final Code c,
                                           final MonoList handlers,
                                           final Code finalhandler) {
        return new Code() {
                public String getName() { return "handlersWithFinally"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    /* Each of these parts have a separate scope */
                    ClassEnvironment e1 = c.defineClasses(e.openScope());
                    e = e.closeScope(e1);
                    for (Iterator i = handlers.iterator(); i.hasNext(); ) {
                        Handler h = (Handler)i.next();
                        e1 = e.openScope();
                        e1 = e1.defineVar(h.varname, h.type, h.flags);
                        e1 = h.code.defineClasses(e1);
                        e = e.closeScope(e1);
                    }
                    e1 = e.openScope();
                    return e.closeScope(finalhandler.defineClasses(e1));
                }
                public Environment defineSupers(Environment e) {
                    /* Each of these parts have a separate scope */
                    e = c.defineSupers(e);
                    for (Iterator i = handlers.iterator(); i.hasNext(); ) {
                        Handler h = (Handler)i.next();
                        e = h.code.defineSupers(e);
                    }
                    return finalhandler.defineSupers(e);
                }
                public Environment defineMembers(Environment e) {
                    /* Each of these parts have a separate scope */
                    e = c.defineMembers(e);
                    for (Iterator i = handlers.iterator(); i.hasNext(); ) {
                        Handler h = (Handler)i.next();
                        e = h.code.defineMembers(e);
                    }
                    return finalhandler.defineMembers(e);
                }
                public ClosedCode eval (Environment env) {
                    final Instr endlabel = Instr.genNil("End of try statement");
                    final Instr handlercodes = Instr.genNil("Try statement handlers");
                    // Allocate var to hold exception for handlers
                    final Environment finallyenv = 
                        env.generateVar("$finallyexc",
                                        Type.throwable_type);
                    final Var fevar = finallyenv.getNewestVar();
                    // Allocate dummy var to ensure that a long/double
                    // return value can be stored
                    final Environment finallyenvtmp = 
                        finallyenv.generateVar("$dummy",
                                        Type.object_type);
                    // Allocate var to hold finally return address
                    final Environment finallyenv2 = 
                        finallyenvtmp.generateVar("$finally",
                                                  Type.object_type);
                    final Var retaddr = finallyenv2.getNewestVar();
                    final ClosedCode f1 = finalhandler.evalNoUse(finallyenv2);
                    final Instr finalcode = Instr.genStore(retaddr.index,
                                                           retaddr.type).
                        append(f1.bytecode).
                        addRet(retaddr.index);
                    final ClosedCode result = 
                        new ClosedCode(Instr.genNil("TryFinally"),
                                       env, Type.void_type);
                    final Instr finalhandlercode =
                        Instr.genStore(fevar.index, fevar.type).
                        addJsr(finalcode).
                        addLoad(fevar.index, fevar.type).
                        addThrow();
                    final Environment bodyenv = env.defineFinallyLabel(finalcode);
                    final ClosedCode c1 = c.evalNoUse(bodyenv);
                    //result/*.addInfo(c1)*/;
                    for (Iterator i = handlers.iterator(); i.hasNext(); ) {
                        final Handler h = (Handler)i.next();
                        final Type t = normalizeClassNames(h.type, env);
                        if (t.kind != Type.OBJECT ||
                            !t.subtypeOf(Type.throwable_type))
                            throw error("Handler type is not Throwable: "+t, env);
                        if (env.varDefined(h.varname)) {
                            throw error("Variable "+h.varname+" already defined", env);
                        }
                        final Environment env1 = bodyenv.defineVar(h.varname, t, h.flags, null);
                        final Var v = env1.getNewestVar();
                        final ClosedCode h1 = h.code.evalNoUse(env1);
                        final Instr handlercode = Instr.genStore(v.index, v.type).
                            append(h1.bytecode).
                            addJsr(finalcode).
                            addGoto(endlabel);
                        handlercodes.append(handlercode);
                        c1.bytecode.addHandler(handlercode,
                                               v.type);
                        //result/*.addInfo(h1)*/;
                    }
                    final Instr handled = c1.bytecode;
                    if (c1.type != Type.bottom) 
                        handled.addJsr(finalcode).
                            addGoto(endlabel);
                    handled.append(handlercodes);
                    handled.addHandler(finalhandlercode,
                                       Type.throwable_type);
                    result.bytecode.append(handled).
                        append(finalhandlercode).
                        append(finalcode).
                        append(endlabel);
                    return result;
                }
                public String toString() { 
                    String s = "try {\n";
                    s += c;
                    for (Iterator i = handlers.iterator(); i.hasNext(); ) {
                        Handler h = (Handler)i.next();
                        s += "} catch ("
                            +(h.flags!=0?Flags.toString(h.flags)+" ":"")
                            +h.type+" "+h.varname+") {\n";
                        s += h.code;
                    }
                    s += "} finally {\n";
                    s += finalhandler;
                    s += "}\n";
                    return s;
                }
            };
    }

    /** Constructor for a statement list.
     *  Note that the modified environment is passed out.
     *  This also has to handle lists of definitions.
     */
    public static Code statements(final MonoList codes) {
        return new Code() {
                public String getName() { return "statements"; }
                public ClassEnvironment defineClasses(ClassEnvironment env) {
                    for (Iterator i = codes.iterator(); i.hasNext(); ) {
                        final Code d = (Code)i.next();
                        env = d.defineClasses(env);
                    }
                    return env;
                }
                public Environment defineSupers(Environment env) {
                    for (Iterator i = codes.iterator(); i.hasNext(); ) {
                        final Code d = (Code)i.next();
                        env = d.defineSupers(env);
                    }
                    return env;
                }
                public Environment defineMembers(Environment env) {
                    for (Iterator i = codes.iterator(); i.hasNext(); ) {
                        final Code d = (Code)i.next();
                        env = d.defineMembers(env);
                    }
                    return env;
                }
                public ClosedCode eval (Environment env) {
                    final ClosedCode result = 
                        new ClosedCode(Instr.genNil("Statement start"),
                                       env, Type.void_type);
                    Type last_type = Type.void_type;
                    for (Iterator i = codes.iterator(); i.hasNext(); ) {
                        final Code c = (Code)i.next();
                        final ClosedCode c1 = c.evalNoUse(env);
                        env = c1.env;
                        result.bytecode.append(c1.bytecode);
                        //result/*.addInfo(c1)*/;
                        result.addWritable(c1);
                        last_type = c1.type;
                        if (last_type == Type.bottom &&
                            i.hasNext()) {
                            throw error("Unreachable code after "+c, env);
                        }
                    }
                    final ClosedCode result2 =
                        new ClosedCode(result.bytecode,
                                       env,
                                       (last_type == Type.bottom?
                                        Type.bottom:Type.void_type))
                        /*.addInfo(result)*/.addWritable(result);
                    return result2;
                }
                public String toString() {
                    String res = "";
                    for (Iterator i = codes.iterator(); i.hasNext(); ) {
                        Code s = (Code)i.next();
                        res += s;
                        res += ";\n";
                    }
                    return res;
                }
            };
    }

    /** Constructor for a single-element unscoped statement list.
     */
    public static Code statements(final Code code) {
        return new Code() {
                public String getName() { return "statements"; }
                public ClassEnvironment defineClasses(ClassEnvironment env) {
                    return code.defineClasses(env);
                }
                public Environment defineSupers(Environment env) {
                    return code.defineSupers(env);
                }
                public Environment defineMembers(Environment env) {
                    return code.defineMembers(env);
                }
                public ClosedCode eval (Environment env) {
                    final ClosedCode cc = code.evalNoUse(env);
                    return new ClosedCode(cc.bytecode,
                                          cc.env, 
                                          (cc.type == Type.bottom?
                                           Type.bottom:Type.void_type))
                        /*.addInfo(cc).addWritable(cc)*/;
                }
                public String toString() {
                    return code+";\n";
                }
            };
    }

    public static Code initializerBlock(final Code body) {
        return new Code() {
                public String getName() { return "initializerBlock"; }
                public ClassEnvironment defineClasses(ClassEnvironment env) {
                    return body.defineClasses(env);
                }
                public Environment defineSupers(Environment env) {
                    return body.defineSupers(env);
                }
                public Environment defineMembers(Environment env) {
                    env = body.defineMembers(env);
                    env.addInitializer(body, true);
                    return env;
                }
                public ClosedCode eval(Environment env) {
                    return new ClosedCode(Instr.genNil("Initializers"),
                                          env, Type.void_type);
                }
            };
    }

    public static Code staticInitializerBlock(final Code body) {
        return new Code() {
                public String getName() { return "staticInitializerBlock"; }
                public ClassEnvironment defineClasses(ClassEnvironment env) {
                    return body.defineClasses(env);
                }
                public Environment defineSupers(Environment env) {
                    return body.defineSupers(env);
                }
                public Environment defineMembers(Environment env) {
                    env = body.defineMembers(env);
                    env.addInitializer(body, false);
                    return env;
                }
                public ClosedCode eval(Environment env) {
                    return new ClosedCode(Instr.genNil("Static initializers"),
                                          env, Type.void_type);
                }
            };
    }

                    /** Constructor for a scoped statement list.
     *  The modified environment is discarded at the end.
     */
    public static Code scopedStatements(final MonoList codes) {
        return new Code() {
                public String getName() { return "scopedStatements"; }
                public ClassEnvironment defineClasses(ClassEnvironment env) {
                    ClassEnvironment env1 = env.openScope();
                    for (final Iterator i = codes.iterator(); i.hasNext(); ) {
                        final Code d = (Code)i.next();
                        env1 = d.defineClasses(env1);
                    }
                    return env.closeScope(env1);
                }
                public Environment defineSupers(Environment env) {
                    for (final Iterator i = codes.iterator(); i.hasNext(); ) {
                        final Code d = (Code)i.next();
                        env = d.defineSupers(env);
                    }
                    return env;
                }
                public Environment defineMembers(Environment env) {
                    for (final Iterator i = codes.iterator(); i.hasNext(); ) {
                        final Code d = (Code)i.next();
                        env = d.defineMembers(env);
                    }
                    return env;
                }
                public ClosedCode eval (Environment env) {
                    Environment env1 = env;
                    final ClosedCode result = 
                        new ClosedCode(Instr.genNil("Scoped statement start"),
                                       env, Type.void_type);
                    Type last_type = Type.void_type;
                    for (Iterator i = codes.iterator(); i.hasNext(); ) {
                        final Code c = (Code)i.next();
                        final ClosedCode c1 = c.evalNoUse(env1);
                        env1 = c1.env;
                        result.bytecode.append(c1.bytecode);
                        //result/*.addInfo(c1)*/;
                        result.addWritable(c1);
                        last_type = c1.type;
                        if (last_type == Type.bottom &&
                            i.hasNext()) {
                            throw error("Unreachable code after "+c, env);
                        }
                    }
                    if (last_type == Type.bottom)
                        return
                            new ClosedCode(result.bytecode, env,
                                           Type.bottom);
                    else
                        return result;
                }
                public String toString() {
                    String res = "";
                    for (Iterator i = codes.iterator(); i.hasNext(); ) {
                        Code s = (Code)i.next();
                        res += s;
                        res += ";\n";
                    }
                    return "{\n"+res+"}";
                }
            };
    }

    public static Code scopedStatements(final Code code) {
        return new Code() {
                public String getName() { return "scopedStatements"; }
                public ClassEnvironment defineClasses(ClassEnvironment env) {
                    ClassEnvironment env1 = env.openScope();
                    return env.closeScope(code.defineClasses(env1));
                }
                public Environment defineSupers(Environment env) {
                    return code.defineSupers(env);
                }
                public Environment defineMembers(Environment env) {
                    return code.defineMembers(env);
                }
                public ClosedCode eval (Environment env) {
                    final Environment envstore = env;
                    final ClosedCode cc = code.evalNoUse(env);
                    return new ClosedCode(cc.bytecode,
                                          env, 
                                          (cc.type == Type.bottom?
                                           Type.bottom:Type.void_type))
                        /*.addInfo(cc).addWritable(cc)*/;
                }
                public String toString() {
                    return "{\n"+code+";\n}";
                }
            };
    }

    /*
      (* Constructors that don't correspond to anything in Java *)

      val symcounter = ref 0;
      fun genname n = (symcounter := (!symcounter) + 1; 
      "#"^n^(Int.toString (!symcounter)));
    */

    /*

    (** Constructor: Generates a statement block with a return value 
    {@type Code list -> Code}
    *)
    fun blockexpr cs =
    fn env =>
    coderes (foldl (fn (c1, (c2, e, t2)) => 
    let val c1' = c1 e in
    (c2@(popType t2)@(codeof c1'), envof c1', typeof c1')
    end) ([], env, VOID) cs) handle e => raise e;
    */

    /*

    (** Constructor: Generates an expression with named arguments
    {@type Code list -> (Name list -> Code) -> Code}
    *)
    fun nameargs cs body =
    fn env =>
    let val ns = map (fn _ => genname "namedarg") cs
    val cs' = mapApply cs env
    val env' = foldl (fn ((n, c), env) => defineVar env n (typeof c) [ACC_FINAL])
    env (ListPair.zip (ns, cs'))
    in
    (body ns) env'
    end handle e => raise e;

    */

    /*

    (** Constructor: Generates an expression with a single named argument
    {@type Code -> (Name -> Code) -> Code}
    *)
    fun namearg c body =
    fn env =>
    let val n = genname "namedarg" 
    val c' = c env
    val env' = defineVar env n (typeof c') [ACC_FINAL]
    val assign = (putVariable n c) env'
    val body' = (body n) env'
    in
    coderes ((codeof assign)@(codeof body'), env, typeof body')
    end handle e => raise e;
    */

    /** Return statement.
     * @param c value to be returned
     */
    public static Code returnVal(final Code c) {
        return new Code() {
                public String getName() { return "returnVal"; }
                public ClassEnvironment defineClasses(ClassEnvironment env) {
                    return c.defineClasses(env);
                }
                public Environment defineSupers(Environment env) {
                    return c.defineSupers(env);
                }
                public Environment defineMembers(Environment env) {
                    return c.defineMembers(env);
                }
                public ClosedCode eval(Environment env) {
                    final ClosedCode c1 = c.eval(env);
                    final MethodInfo m = env.getCurrentMethod();
                    final MonoList finallys = env.getFinallyLabels();
                    final Type rettype = normalizeClassNames(m.getReturnType(), env);
                    if (rettype.kind == Type.VOID) {
                        if (c1.type.kind != Type.VOID)
                            throw error("Cannot return value of type "+c1.type+" from void method", env);
                    } else if (!c1.type.conformsAssign(rettype)) {
                        throw error("Method return type "+c1.type+" differs from declared type "+rettype, env);
                    }
                    if (finallys.size() > 0 && c1.type.kind != Type.VOID) {
                        Environment env1 = env.generateVar("finally", c1.type);
                        Var v = env1.getNewestVar();
                        return new ClosedCode(c1.bytecode.
                                              addStore(v.index, v.type).
                                              append(makeFinallyJumps(finallys)).
                                              addLoad(v.index, v.type).
                                              addReturn(c1.type),
                                              env, Type.bottom);
                    }
                    return new ClosedCode(c1.bytecode.
                                          append(makeFinallyJumps(finallys)).
                                          addReturn(c1.type),
                                          env, Type.bottom);
                }
                public String toString() {
                    return "return "+c;
                }
            };
    }

    /* *******************************************************************
     * Constructors that make definitions (methods, fields, classes)
     * ******************************************************************* */
    /* Special checks for constructors may be made here */
    protected static MethodType makeMethodType(Type returntype,
                                             MonoList args,
                                             Environment e) {
        final Type[] argtypes = new Type[args.size()];
        int index = 0;
        for (final Iterator i = args.iterator(); i.hasNext(); ) {
            argtypes[index++] = normalizeClassNames(((ArgDef)i.next()).type, e);
        }
        return new MethodType(normalizeClassNames(returntype, e), argtypes);
    }

    protected static ClosedCode makeConstructorBody(Code supercall,
                                                    MonoList inits,
                                                    Code body,
                                                    Environment env) {
        if (supercall.isThisConstructorCall()) 
            return statements(new ArrayMonoList().
                              add(supercall).
                              add(body)).evalNoUse(env);
        else {
            Code fullbody = statements(new ArrayMonoList().
                                       add(supercall).
                                       addAll(inits).
                                       add(body));
            return fullbody.evalNoUse(env);
        }
    }

    protected static boolean legalConstructorFlags(int flags) {
        return
            flagsLegal(Flags.PUBLIC|Flags.PROTECTED|Flags.PRIVATE, flags) &&
            flagsCompatible(Flags.PUBLIC, Flags.PROTECTED|Flags.PRIVATE, flags) &&
            flagsCompatible(Flags.PROTECTED, Flags.PRIVATE, flags);
    }

    protected static boolean legalMethodFlags(int flags) {
        return
            flagsLegal(Flags.PUBLIC|Flags.PROTECTED|Flags.PRIVATE|Flags.ABSTRACT|
                       Flags.STATIC|Flags.FINAL|Flags.SYNCHRONIZED|Flags.NATIVE|
                       Flags.STRICTFP, flags) &&
            flagsCompatible(Flags.PUBLIC, Flags.PROTECTED|Flags.PRIVATE, flags) &&
            flagsCompatible(Flags.PROTECTED, Flags.PRIVATE, flags) &&
            flagsCompatible(Flags.ABSTRACT,
                            Flags.PRIVATE|Flags.STATIC|Flags.FINAL|
                            Flags.SYNCHRONIZED|Flags.NATIVE|Flags.STRICTFP, flags) &&
            flagsCompatible(Flags.NATIVE, Flags.STRICTFP, flags);
    }

    /** Shared by defineMethod and defineConstructor: defineClasses() */
    protected static ClassEnvironment methodOrConstructorClasses
        (final MonoList args, final Code body, ClassEnvironment e) {
        ClassEnvironment e1 = e.openScope();
        for (Iterator i = args.iterator(); i.hasNext(); ) {
	    Object nextItem = i.next();
	    if(nextItem instanceof ArgDef){
		ArgDef ad = (ArgDef)nextItem;
		e1 = e1.defineVar(ad.name, ad.type, ad.flags);
	    }else{
		i.remove();
		args.addAll((MonoList)nextItem);
	    }
        }
        e1 = body.defineClasses(e1);
        return e.closeScope(e1);
    }

    /** Shared by defineMethod and defineConstructor: defineSupers() */
    protected static Environment methodOrConstructorSupers
        (final MonoList args, final Code body, Environment e) {
        return body.defineSupers(e);
    }

    /** Checks for compatibility with inherited methods */
    protected static void checkMethodOverrideLegal(MethodInfo m, Environment e) {
        ClassInfo superclass = null;
        try {
            superclass = e.get(m.getClassInfo().supername);
        } catch (ClassNotFoundException err) {
            return;
        }
        MethodInfo superm = null;
        try {
            superm = e.findMethod(superclass, m.getName(), m.getType());
        } catch (MethodNotFoundException err) {
            return;
        }
        if (!superm.getReturnType().equals(m.getReturnType()))
            throw error("Overridden method "+m.getName()+" has different return type", e);
        if (superm.flagsSet(Flags.PUBLIC) &&
            !m.flagsSet(Flags.PUBLIC)) 
            throw error("Overriding public method "+m.getName()+" with non-public.", e);
        if (superm.flagsSet(Flags.PROTECTED) &&
            m.flagsSet(Flags.PRIVATE))
            throw error("Overriding protected method "+m.getName()+" with private.", e);
    }

    public static Code defineMethod(final int flags,
                                    final Type returntype,
                                    final String name,
                                    final MonoList args,
                                    final MonoList throwsDecls,
                                    final Code body) {
        return new Code() {
                public String getName() { return "defineMethod"; }
                MethodInfo m;
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return methodOrConstructorClasses(args, body, e);
                }
                public Environment defineSupers(Environment e) {
                    return methodOrConstructorSupers(args, body, e);
                }
                public Environment defineMembers(Environment e) {
                    final ClassInfo c = e.getCurrentClass();
                    int realflags = flags;
                    if (c.isInterface()) {
                        if (!flagsLegal(Flags.ABSTRACT|Flags.PUBLIC, flags))
                            throw error("Illegal flags for interface method "+name+": "+flags, e);
                        realflags = Flags.ABSTRACT|Flags.PUBLIC;
                    } else {
                        if (!legalMethodFlags(flags))
                            throw error("Illegal flags for method "+name+": "+flags, e);
                    }
                    m = new MethodInfo(c, realflags, name, 
                                       makeMethodType(returntype, args, e));
                    c.addMethod(m);
                    final Environment e1 = e.startMethod(m, args);
                    body.defineMembers(e1);
                    return e;
                }
                public ClosedCode eval(Environment e) {
                    checkMethodOverrideLegal(m, e);

                    final Environment e1 = e.startMethod(m, args);
                    if (!m.flagsSet(Flags.ABSTRACT)) {
                        ClosedCode body1 = body.eval(e1);
                        final Instr bodycode = body1.bytecode;
                        if (returntype == Type.void_type) {
                            if (body1.type != Type.void_type)
                                bodycode.addPop(body1.type);
                            bodycode.addReturn(Type.void_type);
                        }
                        m.addCode(bodycode);
                    }
                    return new ClosedCode(Instr.genNil("Method def"),
                                          e, Type.void_type);
                }
                public String toString() {
                    return (flags!=0?Flags.toString(flags)+" ":"")+
                        returntype+" "+name+"("+toJoinedString(args, ", ")+")"+
                        (Flags.bitsSet(flags, Flags.ABSTRACT)?";":
                         "{\n"+body+";\n}");
                }
            };
    }

    /** The flags that should be set for autogenerated fields */
    public final static int genflags =
        Flags.PRIVATE|Flags.FINAL;

    public static Code defineConstructor(final int flags,
                                         final MonoList args,
                                         final MonoList throwsDecls,
                                         final Code supercall,
                                         final Code body) {
        return new Code() {
                public String getName() { return "defineConstructor"; }
                public boolean isConstructorMethod() { return true; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return methodOrConstructorClasses(args, body, e);
                }
                public Environment defineSupers(Environment e) {
                    return methodOrConstructorSupers(args, body, e);
                }
                public Environment defineMembers(Environment e) {
                    final ClassInfo c = e.getCurrentClass();
                    final String name = "<init>";
                    if (!legalConstructorFlags(flags))
                        throw error("Illegal flags for constructor: "+flags, e);
                    MonoList realargs = args.copy();
                    if (e.selfIsInner()) {
                        MethodInfo m = e.getOuter().getCurrentMethod();
                        if ((m != null && !m.isStatic()) ||
                            (m == null && !c.isStatic())) {
                            Type outertype = normalizeClassNames
                                (e.getOuter().getSelfType(), e);
                            String thisname = "this$"+e.getOuter().numEnclosingClasses();
                            realargs = realargs.add(new ArgDef(Flags.FINAL,
                                                               outertype,
                                                               thisname));
                            if (!c.hasField(thisname))
                                c.addField(new FieldInfo(c, genflags,
                                                         thisname,
                                                         outertype));
                        }
                    }
                    for (Local free = e.getFree(); free != null; 
                         free = free.next) {
                        Type freetype = normalizeClassNames(free.type, e);
                        String freename = "val$"+free.name;
                        realargs = realargs.add(new ArgDef(Flags.FINAL,
                                                           freetype,
                                                           freename));
                        if (!c.hasField(freename))
                            c.addField(new FieldInfo(c, genflags,
                                                     freename,
                                                     freetype));
                    }
                    MethodType mtype = makeMethodType(Type.void_type, realargs, e);
                    MethodInfo m = new MethodInfo(c, flags, "<init>", mtype);
                    c.addMethod(m);                             
                    final Environment e1 = e.startMethod(m, realargs);
                    body.defineMembers(e1);
                    return e;
                }
                public ClosedCode eval(Environment e) {
                    Code realsupercall;
                    final ClassInfo c = e.getCurrentClass();
                    if (supercall != null) {
                        realsupercall = supercall;
                    } else {
                        final ClassInfo s = c.getSuperClass();
                        if (s != null && s.isInnerClass()) {
                            // Should check for static here
                            MonoList superargs = new ArrayMonoList();
                            superargs = superargs.add(getVariable("this$"+(e.numEnclosingClasses()-1)));
                            realsupercall = invokeSuper("", "<init>", superargs);
                        } else {
                            realsupercall = invokeSuper("", "<init>", new ArrayMonoList());
                        }
                    }
                    MonoList realargs = args.copy();
                    if (e.selfIsInner()) {
                        MethodInfo m = e.getOuter().getCurrentMethod();
                        if ((m != null && !m.isStatic()) ||
                            (m == null && !c.isStatic())) {
                            Type outertype = normalizeClassNames
                                (e.getOuter().getSelfType(), e);
                            String thisname = "this$"+e.getOuter().numEnclosingClasses();
                            realargs = realargs.add(new ArgDef(Flags.FINAL,
                                                               outertype,
                                                               thisname));
                        }
                    }
                    for (Local free = e.getFree(); free != null; 
                         free = free.next) {
                        Type freetype = normalizeClassNames(free.type, e);
                        String freename = "val$"+free.name;
                        realargs = realargs.add(new ArgDef(Flags.FINAL,
                                                           freetype,
                                                           freename));
                    }
                    MethodType mtype = makeMethodType(Type.void_type, realargs, e);
                    MethodInfo m = c.getMethod("<init>", mtype);
                    if (m == null)
                        throw error("Can't find earlier created init method in "+c+" for "+mtype, e);
                    final Environment e1 = e.startMethod(m, realargs);
                    Local free = e.getFree();
                    if (e.selfIsInner()) {
                        // Since we can't have this constructor calls, we can
                        // tag the inits on the supercall where they belong.
                        MonoList freeinits = new ArrayMonoList(realsupercall);
                        MethodInfo outer = e.getOuter().getCurrentMethod();
                        if ((outer != null && !outer.isStatic()) ||
                            (outer == null && 
                             !e.getCurrentClass().isStatic())) {
                            freeinits = 
                                freeinits.add(putValue(getField(getThis(),
                                                                "this$"+e.getOuter().numEnclosingClasses()),
                                                       getQualifiedName("this$"+e.getOuter().numEnclosingClasses())));
                        }
                        for (; free != null; free = free.next) {
                            freeinits =
                                freeinits.add(putValue(getField(getThis(),
                                                                "val$"+free.name),
                                                       getQualifiedName("val$"+free.name)));
                        }
                        freeinits = freeinits.add(body);
                        realsupercall = statements(freeinits);
                    }
                    ClosedCode body1 = 
                        makeConstructorBody(realsupercall,
                                            e1.getInstanceInitializers(), 
                                            body, e1);
                    final Instr bodycode = body1.bytecode;
                    if (body1.type != Type.void_type)
                        bodycode.addPop(body1.type);
                    bodycode.addReturn(Type.void_type);
                    m.addCode(bodycode);

                    return new ClosedCode(Instr.genNil("Constructor def"),
                                          e, Type.void_type);
                }
                public String toString() {
                    return (flags!=0?Flags.toString(flags)+" ":"")+
                        "<init>("+toJoinedString(args, ", ")+")"+
                        "{\n"+(supercall!=null?supercall+";\n":"")+body+";\n}";
                }
            };
    }

    protected static Code defineStaticConstructor() {
        return new Code() {
                public String getName() { return "defineStaticConstructor"; }
                MethodInfo m;
                MonoList inits;
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    return e;
                }
                public Environment defineSupers(Environment e) {
                    return e;
                }
                public Environment defineMembers(Environment e) {
                    if (e.getCurrentClass() == null)
                        throw error("Static constructor outside any class", e);
                    inits = e.getStaticInitializers();
                    if (inits == null || inits.size() == 0) 
                        return e;
                    m = new MethodInfo(e.getCurrentClass(),
                                       Flags.STATIC, "<clinit>",
                                       new MethodType(Type.void_type));
                    e.getCurrentClass().addMethod(m);
                    return e;
                }
                public ClosedCode eval(Environment e) {
                    if (e.getCurrentClass() == null)
                        throw error("Static constructor outside any class", e);
                    if (inits == null || inits.size() == 0) 
                        return new ClosedCode(Instr.genNil("No <clinit> def"),
                                              e, Type.void_type);
                    final Environment e1 = e.startMethod(m, null);
                    /* Assemble the body (like Dr. Frankenstein:) */
                    final Code body = Constructor.statements(inits);
                    final ClosedCode body1 = body.evalNoUse(e1);
                    final Instr bodycode = body1.bytecode;
                    bodycode.addReturn(Type.void_type);
                    m.addCode(bodycode);
                    return new ClosedCode(Instr.genNil("<clinit> def"),
                                          e, Type.void_type);
                }
            };
    }

    protected static boolean legalFieldFlags(int flags) {
        return
            flagsLegal(Flags.PUBLIC|Flags.PROTECTED|Flags.PRIVATE|
                       Flags.STATIC|Flags.FINAL|Flags.TRANSIENT|
                       Flags.VOLATILE, flags) &&
            flagsCompatible(Flags.PUBLIC, Flags.PROTECTED|Flags.PRIVATE, flags) &&
            flagsCompatible(Flags.PROTECTED, Flags.PRIVATE, flags) &&
            flagsCompatible(Flags.FINAL, Flags.VOLATILE, flags);
    }


    public static Code defineField(final int flags,
                                   final Type type,
                                   final String name,
                                   final Code init) {
        return new Code() {
                public String getName() { return "defineField"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    if (init != null)
                        return init.defineClasses(e);
                    else
                        return e;
                }
                public Environment defineSupers(Environment e) {
                    if (init != null)
                        return init.defineSupers(e);
                    else
                        return e;
                }
                public Environment defineMembers(Environment e) {
                    final ClassInfo c = e.getCurrentClass();
                    int realflags = flags;
                    if (c.isInterface()) {
                        if (!flagsLegal(Flags.PUBLIC|Flags.STATIC|Flags.FINAL,
                                        flags))
                            throw error("Illegal flags for interface field "+name+": "+flags, e);
                        realflags = Flags.PUBLIC|Flags.STATIC|Flags.FINAL;
                    } else {
                        if (!legalFieldFlags(flags))
                            throw error("Illegal flags for field "+name+": "+flags, e);
                    }
                    final FieldInfo f = new FieldInfo(c, realflags, name, 
                                                      normalizeClassNames(type, e));
                    c.addField(f);
                    if (init != null) {
                        e = init.defineMembers(e);
                        e.addInitializer(putField(name, init),
                                         (realflags & Flags.STATIC) == 0);
                    }
                    return e;
                }
                public ClosedCode eval(Environment e) {
                    return new ClosedCode(Instr.genNil("Field def"),
                                          e, Type.void_type);
                }
                public String toString() {
                    return (flags!=0?Flags.toString(flags)+" ":"")+
                        type+" "+(init!=null?init.toString():name);
                }
            };
    }

    public static Code defineField(final int flags,
                                   final Type type,
                                   final String name) {
        return defineField(flags, type, name, null);
    }

    public static Code defineFieldOrVariable(final int flags,
                                             final Type type,
                                             final String name,
                                             final Code init) {
        return new Code() {
                public String getName() { return "defineFieldOrVariable"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    if (init != null)
                        return init.defineClasses(e);
                    else
                        return e;
                }
                public Environment defineSupers(Environment e) {
                    if (init != null)
                        return init.defineSupers(e);
                    else
                        return e;
                }
                public Environment defineMembers(Environment e) {
                    if (e.getCurrentMethod() == null) {
                        final ClassInfo c = e.getCurrentClass();
                        int realflags = flags;
                        if (c.isInterface()) {
                            if (!flagsLegal(Flags.PUBLIC|Flags.STATIC|Flags.FINAL,
                                            flags))
                                throw error("Illegal flags for interface field "+name+": "+flags, e);
                            realflags = Flags.PUBLIC|Flags.STATIC|Flags.FINAL;
                        } else {
                            if (!legalFieldFlags(flags))
                                throw error("Illegal flags for field "+name+": "+flags, e);
                        }
                        final FieldInfo f = new FieldInfo(c, realflags, name, 
                                                          normalizeClassNames(type, e));
                        c.addField(f);
                        if (init != null) {
                            e = init.defineMembers(e);
                            e.addInitializer(putField(name, init),
                                             (realflags & Flags.STATIC) == 0);
                        }
                        return e;
                    } else {
                        return e;
                    }
                }
                public ClosedCode eval(Environment e) {
                    if (e.getCurrentMethod() != null) {
                        return defineVariable(flags, 
                                              normalizeClassNames(type, e),
                                              name, init).evalNoUse(e);
                    } else {
                        return new ClosedCode(Instr.genNil("Field def"),
                                              e, Type.void_type);
                    }
                }
                public String toString() {
                    return (flags!=0?Flags.toString(flags)+" ":"")+
                        type+" "+(init!=null?name+" = "+init.toString():name);
                }
            };
    }

    public static Code defineFieldOrVariable(final int flags,
                                             final Type type,
                                             final String name) {
        return defineFieldOrVariable(flags, type, name, null);
    }

    protected static boolean legalClassFlags(int flags, boolean ismemberclass) {
        if (ismemberclass)
            return
                flagsLegal(Flags.PUBLIC|Flags.PROTECTED|Flags.PRIVATE|Flags.ABSTRACT|
                           Flags.STATIC|Flags.FINAL|Flags.STRICTFP, flags) &&
                flagsCompatible(Flags.PUBLIC, Flags.PROTECTED|Flags.PRIVATE, flags) &&
                flagsCompatible(Flags.PROTECTED, Flags.PRIVATE, flags) &&
                flagsCompatible(Flags.FINAL, Flags.ABSTRACT, flags);
        else
            return
                flagsLegal(Flags.PUBLIC|Flags.ABSTRACT|
                           Flags.FINAL|Flags.STRICTFP, flags) &&
                flagsCompatible(Flags.FINAL, Flags.ABSTRACT, flags);
    }

    protected static boolean containsInitMethod(MonoList members) {
        for (Iterator i = members.iterator(); i.hasNext(); ) {
            Code c = ((Code)i.next());
            if (c.isConstructorMethod()) return true;
        }
        return false;
    }

    public static Code defineClass(final int flags,
                                   final String name,
                                   final MonoList supername,
                                   final MonoList implementees,
                                   final MonoList members) {
	Iterator i = supername.iterator();
	String superClass = (String)i.next();
	while(i.hasNext())
	    superClass += "."+i.next();
	final String finalSuperClass = superClass;
	return defineClass(flags,name,finalSuperClass,implementees,members);
	
    }
    /** Constructor for a class (inner or otherwise).
     *  @param flags Flags for the class.
     *  @param name The name of the class (without package or outer class).
     *  @param supername The name of the superclass.
     *  @param implementees A list of interfaces implemented (strings).
     *  @param members A list of class members (fields, methods, innerclasses).
     */
    public static Code defineClass(final int flags,
                                   final String name,
                                   final String supername,
                                   final MonoList implementees,
                                   final MonoList members) {
        return new Code() {
                public String getName() { return "defineClass"; }
                ClassInfo c;
                Local free;
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    if (!legalClassFlags(flags, e.getCurrentClass() != null))
                        throw error("Illegal flags for class "+name+": "+flags, e);
                    String fullname = name;
                    if (e.getCurrentClass() != null) { /* Inner class */
                        if ((flags & Flags.PUBLIC) == 0) {
                            /* Inaccessible inner class -- give it a number */
                            fullname = e.genSym(e.getCurrentClass().getName())+"$"+name;
                            e = e.updateGenSym();
                        } else {
                            fullname = e.getCurrentClass().getName()+"/"+name;
                        }
                    } else {
                        fullname = e.fullClassName(name);
                    }
                    ClassInfo outerc = e.getCurrentClass();
                    c = new ClassInfo(flags|Flags.SUPER, fullname, null);
                    e.addClass(c);
                    if (outerc != null) {
                        outerc.addInnerClass(c);
                        c.setOuterClass(outerc);
                    }
                    ClassEnvironment e1 = e.startClass(c);

		    // [baris]
		    // if $Field<...>$ is used, what we get is an ArrayMonoList of Codes, 
		    // instead of a single Code. Therefore we need to extract all the Codes
		    // from that list, remove the list, and put the Codes into the members.
                    for (Iterator i = members.iterator(); i.hasNext(); ) {
			Object nextMember = i.next();
			if(nextMember instanceof Code){			    
			    Code d = (Code)nextMember;
			    e1 = d.defineClasses(e1);
			}else{
			    i.remove();
			    members.addAll((MonoList) nextMember);
			}			
                    }

                    // Unset class, but remember inner classes
                    if (e1.selfIsInner()) {
                        Local v = e1.free;
                        e1 = e1.resetFree();
                        for (; v != null; v = v.next) {
                            if (e1.outer.getVar(v.name) != null) {
                                this.free = new Var(v.name, 0, v.type,
                                                    v.flags, this.free);
                            } else {
                                /* Pass as free to outer outer class */
                                e1 = e1.addFree(v);
                            }
                        }
                        e.closeScope(e1);
                        return e;
                    } else {
                        if (e1.free != null) {
                            throw error("Internal:  Free variable in outermost class "+c, e);
                        }
                        e1.setCurrentClass(null);
                        return e1;
                    }
                }
                public Environment defineSupers(Environment e) {
                    Environment env = e.startClass(c, free);
                    if (supername != null && !(supername.equals(""))) {
                        final String norm_super = normalizeClassNames(supername, env);
                        c.setSuperClass(norm_super);
                        try {
                            final ClassInfo superc = env.get(norm_super);
                            if (superc.isInterface()) 
                                throw error("Super class of "+c+" is an interface: "+superc, e);
                        } catch (ClassNotFoundException err) {
                            throw error("Super class not found: "+err, e);
                        }
                    }
                    for (final Iterator i = implementees.iterator(); i.hasNext(); ) {
			Object nextItem = i.next();
			String className = "";
			if(nextItem instanceof MonoList){
			    Iterator iter=((MonoList)nextItem).iterator();
			    className += iter.next();
			    while(iter.hasNext()){
				className += "."+iter.next();
			    }
			}
			else
			    className += nextItem;
                        final String norm_interface =
                            normalizeClassNames(className, env);
                        c.implementsInterface(norm_interface);
                        try {
                            final ClassInfo iface = env.get(norm_interface);
                            if (!iface.isInterface()) 
                                throw error("Implemented class of "+c+" is not an interface: "+iface, e);
                        } catch (ClassNotFoundException err) {
                            throw error("Implemented interface not found: "+err, e);
                        }
                    }
                    for (final Iterator i = members.iterator(); i.hasNext(); ) {
			Code d = (Code)i.next();
			env = d.defineSupers(env);
                    }
                    return e;
                }
                Code statics = defineStaticConstructor();
                public Environment defineMembers(Environment e) {
                    if (c == null)
                        throw error("Internal: Couldn't refind class "+
                              e.fullClassName(name), e);
                    Environment env = e.startClass(c, free);
                    for (Iterator i = members.iterator(); i.hasNext(); ) {
			final Code d = (Code)i.next();
			env = d.defineMembers(env);
                    }
                    if (!containsInitMethod(members)) {
                        final Code d = defineConstructor(((c.flags & Flags.PUBLIC) != 0?
                                                          Flags.PUBLIC:0),
                                                         new ArrayMonoList(),
                                                         new ArrayMonoList(),
                                                         null,
                                                         noop());
                        // This is getting redefined below instead.
                        // members.add(d);
                        d.defineMembers(env);
                    }
                    statics.defineMembers(env);
                    return e;
                }
                public ClosedCode eval(Environment e) {
                    if (e.getCurrentMethod() != null)
                        e = e.defineLocal(name, new Type(c.getName()), flags);
                    Environment env = e.startClass(c, free);

                    statics.eval(env);

                    MonoList realmembers = members;
                    if (!containsInitMethod(members)) {
                        final Code d = defineConstructor(((c.flags & Flags.PUBLIC) != 0?
                                                          Flags.PUBLIC:0),
                                                         new ArrayMonoList(),
                                                         new ArrayMonoList(),
                                                         null,
                                                         noop());
                        realmembers = realmembers.add(d);
                    }
                    evalCodeList(realmembers, env);

                    try {
                        c.writeClassFile();
                    } catch (IOException ex) {
                    }

                    return new ClosedCode(Instr.genNil("Class def"),
                                          e, Type.void_type)
                        .addWritable(c);
                }
                public String toString() {
                    return (flags!=0?Flags.toString(flags)+" ":"")+
                        "class "+name+
                        (supername!=null&&supername.length()>0?
                         " extends "+supername:"")+
                        (implementees!=null && implementees.size() > 0?
                         " implements "+toJoinedString(implementees, ", "):"")+
                        " {\n"+toJoinedString(members, ";\n")+"\n}\n";
                }
            };
    }

    protected static boolean legalInterfaceFlags(int flags, boolean ismember) {
        if (ismember)
            return
                flagsLegal(Flags.PUBLIC|Flags.PROTECTED|Flags.PRIVATE|Flags.ABSTRACT|
                           Flags.STATIC|Flags.STRICTFP, flags) &&
                flagsCompatible(Flags.PUBLIC, Flags.PROTECTED|Flags.PRIVATE, flags) &&
                flagsCompatible(Flags.PROTECTED, Flags.PRIVATE, flags);
        else
            return
                flagsLegal(Flags.PUBLIC|Flags.ABSTRACT|Flags.FINAL|Flags.STRICTFP,
                           flags);
    }

    public static Code defineInterface(final int flags,
                                       final String name,
                                       final MonoList superinterfaces,
                                       final MonoList members) {
        return new Code() {
                public String getName() { return "defineInterface"; }
                InterfaceInfo i;
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    if (!legalInterfaceFlags(flags, e.getCurrentClass() != null))
                        throw error("Illegal flags for interface "+name+": "+flags, e);
                    final int realflags = flags | Flags.INTERFACE;
                    // Member interfaces are implicitly static
                    i = new InterfaceInfo(realflags,
                                          e.fullClassName(name),
                                                              null);
                    ClassInfo outerc = e.getCurrentClass();
                    e.addClass(i);
                    if (outerc != null) {
                        outerc.addInnerClass(i);
                        i.setOuterClass(outerc);
                    }
                    return e;
                }
                public Environment defineSupers(Environment e) {
                    Environment env = e.startClass(i, null);
                    for (final Iterator si = superinterfaces.iterator(); si.hasNext(); ) {
                        i.implementsInterface
                            (normalizeClassNames((String)si.next(), env));
                    }
                    for (final Iterator si = members.iterator(); si.hasNext(); ) {
                        Code d = (Code)si.next();
                        env = d.defineSupers(env);
                    }
                    return e;                    
                }
                public Environment defineMembers(Environment e) {
                    Environment env = e.startClass(i, null);
                    for (final Iterator si = members.iterator(); si.hasNext(); ) {
                        final Code d = (Code)si.next();
                        env = d.defineMembers(env);
                    }
                    return e;
                }
                public ClosedCode eval(Environment e) {
                    Environment env = e.startClass(i, null);
                    MonoList statics = env.getStaticInitializers();
                    if (statics != null)
                        for (final Iterator si = statics.iterator(); si.hasNext(); ) {
                            Code c = ((Code)si.next());
                            c.evalNoUse(env);
                        }
                    MonoList instances = env.getInstanceInitializers();
                    if (instances != null)
                        for (final Iterator si = instances.iterator(); si.hasNext(); ) {
                            Code c = ((Code)si.next());
                            c.evalNoUse(env);
                        }
                    evalCodeList(members, env);
                    try {
                        i.writeClassFile();
                    } catch (IOException ex) {
                    }
                    return new ClosedCode(Instr.genNil("Interface def"),
                                          e, Type.void_type)
                        .addWritable(i);
                }
                public String toString() {
                    return (flags!=0?Flags.toString(flags)+" ":"")+
                        "interface "+name+
                        (superinterfaces!=null && superinterfaces.size() != 0?" extends "+toJoinedString(superinterfaces, ", "):"")+" {\n"+
                        toJoinedString(members, ";\n")+"}\n";
                }

            };
    }

    public static Code importClass(final String classname) {
        return new Code() {
                public String getName() { return "importClass"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    e.addImport(classname.replace('.', '/'));
                    return e;
                }
                public Environment defineSupers(Environment e) {
                    e.addImport(classname.replace('.', '/'));
                    return e;
                }
                public Environment defineMembers(Environment e) {
                    e.addImport(classname.replace('.', '/'));
                    return e;
                }
                public ClosedCode eval(Environment e) {
                    e.addImport(classname.replace('.', '/'));
                    return new ClosedCode(Instr.genNil("Import def"),
                                          e, Type.void_type);
                }
                public String toString() {
                    return "import "+classname+";\n";
                }
            };
    }

    public static Code importPackage(final String packagename) {
        return new Code() {
                public String getName() { return "importPackage"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    e.addImportPackage(packagename.replace('.', '/'));
                    return e;
                }
                public Environment defineSupers(Environment e) {
                    e.addImportPackage(packagename.replace('.', '/'));
                    return e;
                }
                public Environment defineMembers(Environment e) {
                    e.addImportPackage(packagename.replace('.', '/'));
                    return e;
                }
                public ClosedCode eval(Environment e) {
                    e.addImportPackage(packagename.replace('.', '/'));
                    return new ClosedCode(Instr.genNil("Import def"),
                                          e, Type.void_type);
                }
                public String toString() {
                    return "import "+packagename+".*;\n";
                }
            };
    }

    public static Code definePackage(final MonoList packagenames) {
        return definePackage(toJoinedString(packagenames, "/"));
    }

    public static Code definePackage(final String packagename) {
        return new Code() {
                public String getName() { return "definePackage"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    e.setCurrentPackage(packagename.replace('.', '/'));
                    return e;
                }
                public Environment defineSupers(Environment e) {
                    e.setCurrentPackage(packagename.replace('.', '/'));
                    return e;
                }
                public Environment defineMembers(Environment e) {
                    e.setCurrentPackage(packagename.replace('.', '/'));
                    return e;
                }
                public ClosedCode eval(Environment e) {
                    e.setCurrentPackage(packagename.replace('.', '/'));
                    return new ClosedCode(Instr.genNil("Package def"),
                                          e, Type.void_type);
                }
                public String toString() {
                    return "package "+packagename+";\n";
                }
            };
    }

    public static Code defineCompilationUnit(final MonoList definitions) {
        return new Code() {
                public String getName() { return "defineCompilationUnit"; }
                public ClassEnvironment defineClasses(ClassEnvironment e) {
                    ClassEnvironment e1 = (ClassEnvironment)e.startCompilationUnit();
                    for (final Iterator i = definitions.iterator(); i.hasNext(); ) {
                        e1 = ((Code)i.next()).defineClasses(e1);
                    }
                    return e;
                }
                public Environment defineSupers(Environment e) {
                    Environment e1 = (Environment)e.startCompilationUnit();
                    for (final Iterator i = definitions.iterator(); i.hasNext(); ) {
                        final Code d = (Code)i.next();
                        e1 = d.defineSupers(e1);
                    }
                    return e;
                }
                public Environment defineMembers(Environment e) {
                    Environment e1 = (Environment)e.startCompilationUnit();
                    for (final Iterator i = definitions.iterator(); i.hasNext(); ) {
                        e1 = ((Code)i.next()).defineMembers(e1);
                    }
                    return e;
                }
                public ClosedCode eval(Environment e) {
                    Environment e1 = (Environment)e.startCompilationUnit();
                    final ClosedCode cc = new ClosedCode(Instr.genNil("Compilation unit"),
                                                         e1, Type.void_type);
                    for (final Iterator i = definitions.iterator(); i.hasNext(); ) {
                        final ClosedCode c = ((Code)i.next()).eval(e1);
                        cc.addWritable(c);
                    }
                    return cc;
                }
                public String toString() {
                    return toJoinedString(definitions, "\n");
                }
            };
    }
}

/* End of class Constructor */
